Type: Package
Package: logger.tester
Authors@R: c(
    person("Gergely", "Daroczi", , "daroczig@rapporter.net", role = c("aut", "cre")),
    person("System1", role = c("fnd"))
    )
Title: Dummy package providing testing functions for logger
Description: Dummy package providing testing functions for logger
Version: 0.1
Date: 2018-07-04
URL: https://github.com/daroczig/logger
RoxygenNote: 7.3.2
License: MIT + file LICENSE
Imports:
  logger

Type: Package
Package: logger
Title: A Lightweight, Modern and Flexible Logging Utility
Version: 0.4.2
Date: 2026-05-08
Authors@R: c(
    person("Gergely", "Daróczi", , "daroczig@rapporter.net", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-3149-8537")),
    person("Hadley", "Wickham", , "hadley@posit.co", role = "aut",
           comment = c(ORCID = "0000-0003-4757-117X")),
    person("Jonathan", "Carroll", , "rpkg@jcarroll.com.au", role = "ctb", 
           comment = c(ORCID = "0000-0002-1404-5264")),       
    person("Spare Cores", role = "fnd"),
    person("System1", role = "fnd")
  )
Description: Inspired by the the 'futile.logger' R package and 'logging'
    Python module, this utility provides a flexible and extensible way of
    formatting and delivering log messages with low overhead.
License: MIT + file LICENSE
URL: https://daroczig.github.io/logger/
BugReports: https://github.com/daroczig/logger/issues
Depends: R (>= 4.0.0)
Imports: utils
Suggests: botor, cli, covr, crayon, devtools, glue, jsonlite, knitr,
        mirai (>= 1.3.0), ntfy, pander, parallel, R.utils, rmarkdown,
        roxygen2, RPushbullet, rsyslog, shiny, slackr (>= 1.4.1),
        syslognet, telegram, testthat (>= 3.0.0), withr
Enhances: futile.logger, log4r, logging
VignetteBuilder: knitr
Config/testthat/edition: 3
Config/testthat/parallel: TRUE
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-08 20:57:17 UTC; daroczig
Author: Gergely Daróczi [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-3149-8537>),
  Hadley Wickham [aut] (ORCID: <https://orcid.org/0000-0003-4757-117X>),
  Jonathan Carroll [ctb] (ORCID: <https://orcid.org/0000-0002-1404-5264>),
  Spare Cores [fnd],
  System1 [fnd]
Maintainer: Gergely Daróczi <daroczig@rapporter.net>
Repository: CRAN
Date/Publication: 2026-05-10 05:10:27 UTC
Built: R 4.5.2; ; 2026-05-10 07:03:39 UTC; unix

Type: Package
Package: logger
Title: A Lightweight, Modern and Flexible Logging Utility
Version: 0.4.3
Date: 2026-08-23
Authors@R: c(
    person("Gergely", "Daróczi", , "daroczig@rapporter.net", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-3149-8537")),
    person("Hadley", "Wickham", , "hadley@posit.co", role = "aut",
           comment = c(ORCID = "0000-0003-4757-117X")),
    person("Jonathan", "Carroll", , "rpkg@jcarroll.com.au", role = "ctb", 
           comment = c(ORCID = "0000-0002-1404-5264")),       
    person("Spare Cores", role = "fnd"),
    person("System1", role = "fnd")
  )
Description: Inspired by the the 'futile.logger' R package and 'logging'
    Python module, this utility provides a flexible and extensible way of
    formatting and delivering log messages with low overhead.
License: MIT + file LICENSE
URL: https://daroczig.github.io/logger/
BugReports: https://github.com/daroczig/logger/issues
Depends: R (>= 4.0.0)
Imports: utils
Suggests: botor, cli, covr, crayon, devtools, glue, jsonlite, knitr,
        mirai (>= 1.3.0), ntfy, pander, parallel, R.utils, rmarkdown,
        roxygen2, RPushbullet, rsyslog, shiny, slackr (>= 1.4.1),
        syslognet, telegram, testthat (>= 3.0.0), withr
Enhances: futile.logger, log4r, logging
VignetteBuilder: knitr
Config/testthat/edition: 3
Config/testthat/parallel: TRUE
Encoding: UTF-8
Config/roxygen2/version: 8.1.0
NeedsCompilation: no
Packaged: 2026-08-24 12:00:28 UTC; daroczig
Author: Gergely Daróczi [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-3149-8537>),
  Hadley Wickham [aut] (ORCID: <https://orcid.org/0000-0003-4757-117X>),
  Jonathan Carroll [ctb] (ORCID: <https://orcid.org/0000-0002-1404-5264>),
  Spare Cores [fnd],
  System1 [fnd]
Maintainer: Gergely Daróczi <daroczig@rapporter.net>
Repository: CRAN
Date/Publication: 2026-08-24 16:20:18 UTC
Built: R 4.6.1; ; 2026-08-24 17:46:43 UTC; unix

Type: Package
Package: logger.tester
Authors@R: c(
    person("Gergely", "Daroczi", , "daroczig@rapporter.net", role = c("aut", "cre")),
    person("System1", role = c("fnd"))
    )
Title: Dummy package providing testing functions for logger
Description: Dummy package providing testing functions for logger
Version: 0.1
Date: 2018-07-04
URL: https://github.com/daroczig/logger
RoxygenNote: 6.1.0
License: AGPL-3
Imports:
  logger

# Generated by roxygen2: do not edit by hand

S3method(as.loglevel,character)
S3method(as.loglevel,default)
S3method(as.loglevel,integer)
S3method(as.loglevel,numeric)
S3method(print,loglevel)
export("%except%")
export(DEBUG)
export(ERROR)
export(FATAL)
export(INFO)
export(OFF)
export(SUCCESS)
export(TRACE)
export(WARN)
export(appender_async)
export(appender_console)
export(appender_file)
export(appender_kinesis)
export(appender_ntfy)
export(appender_pushbullet)
export(appender_slack)
export(appender_stderr)
export(appender_stdout)
export(appender_syslog)
export(appender_syslognet)
export(appender_tee)
export(appender_telegram)
export(appender_void)
export(as.loglevel)
export(colorize_by_log_level)
export(delete_logger_index)
export(deparse_to_one_line)
export(fail_on_missing_package)
export(formatter_cli)
export(formatter_glue)
export(formatter_glue_or_sprintf)
export(formatter_glue_safe)
export(formatter_json)
export(formatter_logging)
export(formatter_pander)
export(formatter_paste)
export(formatter_sprintf)
export(get_logger_meta_variables)
export(grayscale_by_log_level)
export(layout_blank)
export(layout_gha)
export(layout_glue)
export(layout_glue_colors)
export(layout_glue_generator)
export(layout_json)
export(layout_json_parser)
export(layout_logging)
export(layout_simple)
export(layout_syslognet)
export(log_appender)
export(log_chunk_time)
export(log_debug)
export(log_elapsed)
export(log_elapsed_start)
export(log_error)
export(log_errors)
export(log_eval)
export(log_failure)
export(log_fatal)
export(log_formatter)
export(log_indices)
export(log_info)
export(log_layout)
export(log_level)
export(log_messages)
export(log_namespaces)
export(log_separator)
export(log_shiny_input_changes)
export(log_success)
export(log_threshold)
export(log_tictoc)
export(log_trace)
export(log_warn)
export(log_warnings)
export(log_with_separator)
export(logger)
export(skip_formatter)
export(with_log_threshold)
importFrom(stats,setNames)
importFrom(utils,assignInMyNamespace)
importFrom(utils,assignInNamespace)
importFrom(utils,capture.output)
importFrom(utils,compareVersion)
importFrom(utils,getFromNamespace)
importFrom(utils,packageVersion)
importFrom(utils,str)

Most docstrings have been generated by LLMs and tools with minimal or no post-editing.
In alphabetical order:
    Anthropic/Claude: https://www.anthropic.com/, https://claude.ai/
    DeepSeek:         https://deep-seek.ai/
    Ollama:           https://ollama.com/
                      with models:
                          codegeex4
                          deepcoder
                          deepseek-coder-v2
                          granite-code

Code comments have been written by contributor(s).

<!-- README.md is generated from README.Rmd. Please edit that file -->

# logger <a href="https://daroczig.github.io/logger/"><img src="man/figures/logo.png" align="right" height="138" alt="logger website" /></a>

<!-- badges: start -->

[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![CRAN](https://www.r-pkg.org/badges/version/logger)](https://cran.r-project.org/package=logger)
[![Build
Status](https://github.com/daroczig/logger/workflows/R-CMD-check/badge.svg)](https://github.com/daroczig/logger/actions)
[![Code
Coverage](https://codecov.io/gh/daroczig/logger/branch/master/graph/badge.svg)](https://app.codecov.io/gh/daroczig/logger)
[![A Mikata
Project](https://mikata.dev/img/badge.svg)](https://mikata.dev)
<!-- badges: end -->

A lightweight, modern and flexible logging utility for R – heavily
inspired by the `futile.logger` R package and `logging` Python module.

## Installation

[![CRAN
version](https://www.r-pkg.org/badges/version-ago/logger)](https://cran.r-project.org/package=logger)

``` r
install.packages("logger")
```

The most recent, development version of `logger` can also be installed
from GitHub:

``` r
# install.packages("pak")
pak::pak("daroczig/logger")
```

## Quick example

Setting the log level threshold to something low and logging various
messages in ad-hoc and programmatic ways:

``` r
library(logger)
log_threshold(DEBUG)
log_info("Script starting up...")
#> INFO [2024-08-15 11:59:27] Script starting up...

pkgs <- available.packages()
log_info("There are {nrow(pkgs)} R packages hosted on CRAN!")
#> INFO [2024-08-15 11:59:28] There are 21131 R packages hosted on CRAN!

for (letter in letters) {
  lpkgs <- sum(grepl(letter, pkgs[, "Package"], ignore.case = TRUE))
  log_level(
    if (lpkgs < 5000) TRACE else DEBUG,
    "{lpkgs} R packages including the {shQuote(letter)} letter"
  )
}
#> DEBUG [2024-08-15 11:59:28] 10193 R packages including the 'a' letter
#> DEBUG [2024-08-15 11:59:28] 7016 R packages including the 'c' letter
#> DEBUG [2024-08-15 11:59:28] 5751 R packages including the 'd' letter
#> DEBUG [2024-08-15 11:59:28] 10907 R packages including the 'e' letter
#> DEBUG [2024-08-15 11:59:28] 8825 R packages including the 'i' letter
#> DEBUG [2024-08-15 11:59:28] 7059 R packages including the 'l' letter
#> DEBUG [2024-08-15 11:59:28] 7045 R packages including the 'm' letter
#> DEBUG [2024-08-15 11:59:28] 6665 R packages including the 'n' letter
#> DEBUG [2024-08-15 11:59:28] 7863 R packages including the 'o' letter
#> DEBUG [2024-08-15 11:59:28] 6581 R packages including the 'p' letter
#> DEBUG [2024-08-15 11:59:28] 11229 R packages including the 'r' letter
#> DEBUG [2024-08-15 11:59:28] 10296 R packages including the 's' letter
#> DEBUG [2024-08-15 11:59:28] 9531 R packages including the 't' letter

log_warn("There might be many, like {1:2} or more warnings!!!")
#> WARN [2024-08-15 11:59:28] There might be many, like 1 or more warnings!!!
#> WARN [2024-08-15 11:59:28] There might be many, like 2 or more warnings!!!
```

You can even use a custom log layout to render the log records with
colors, as you can see in `layout_glue_colors()`:

<img src="man/figures/colors.png" alt="colored log output">

But you could set up any custom colors and layout, eg using custom
colors only for the log levels, make it grayscale, include the calling
function or R package namespace with specific colors etc. For more
details, see `vignette("write_custom_extensions")`.

## Related work

There are many other logging packages available on CRAN:

- [`futile.logger`](https://cran.r-project.org/package=futile.logger):
  probably the most popular `log4j` variant (and I’m a big fan)
- [`logging`](https://cran.r-project.org/package=logging): just like
  Python’s `logging` package
- [`lgr`](https://cran.r-project.org/package=lgr): built on top of R6.
- [`loggit`](https://cran.r-project.org/package=loggit): capture
  `message`, `warning` and `stop` function messages in a JSON file
- [`log4r`](https://cran.r-project.org/package=log4r): `log4j`-based,
  object-oriented logger
- [`rsyslog`](https://cran.r-project.org/package=rsyslog): logging to
  `syslog` on ‘POSIX’-compatible operating systems
- [`lumberjack`](https://cran.r-project.org/package=lumberjack):
  provides a special operator to log changes in data

Why use logger? I decided to write the `n+1`th extensible `log4j` logger
that fits my liking — and hopefully yours as well — with the aim to:

- Keep it close to `log4j`.
- Respect the modern function/variable naming conventions and general R
  coding style.
- By default, rely on `glue()` when it comes to formatting / rendering
  log messages, but keep it flexible if others prefer `sprintf()`
  (e.g. for performance reasons) or other functions.
- Support vectorization (eg passing a vector to be logged on multiple
  lines).
- Make it easy to extend with new features (e.g. custom layouts, message
  formats and output).
- Prepare for writing to various services, streams etc
- Provide support for namespaces, preferably automatically finding and
  creating a custom namespace for all R packages writing log messages,
  each with optionally configurable log level threshold, message and
  output formats.
- Allow stacking loggers to implement logger hierarchy – even within a
  namespace, so that the very same `log` call can write all the `TRACE`
  log messages to the console, while only pushing `ERROR`s to DataDog
  and eg `INFO` messages to CloudWatch.
- Optionally colorize log message based on the log level.
- Make logging fun!

Welcome to the
[Bazaar](https://en.wikipedia.org/wiki/The_Cathedral_and_the_Bazaar)! If
you already use any of the above packages for logging, you might find
`vignette("migration")` useful.

## Interested in more details?

<div class=".pkgdown-hide">

Check out the main documentation site at
<https://daroczig.github.io/logger/> or the vignettes on the below
topics:

- [Introduction to
  logger](https://daroczig.github.io/logger/articles/Intro.html)
- [The Anatomy of a Log
  Request](https://daroczig.github.io/logger/articles/anatomy.html)
- [Customizing the Format and the Destination of a Log
  Record](https://daroczig.github.io/logger/articles/customize_logger.html)
- [Writing Custom Logger
  Extensions](https://daroczig.github.io/logger/articles/write_custom_extensions.html)
- [Migration Guide from other logging
  packages](https://daroczig.github.io/logger/articles/migration.html)
- [Logging from R
  Packages](https://daroczig.github.io/logger/articles/r_packages.html)
- [Simple Benchmarks on
  Performance](https://daroczig.github.io/logger/articles/performance.html)

</div>

If you prefer visual content, you can watch the video recording of the
“Getting things logged” talk at RStudio::conf(2020):

[![Gergely Daroczi presenting “Getting things logged” on using the
`logger` R package at the RStudio conference in
2020](https://img.youtube.com/vi/_rUuBbml9dU/0.jpg)](https://www.youtube.com/watch?v=_rUuBbml9dU)

# Logger

Logger is a simple but powerful logging utility to output messages in your Ruby program.

Logger has the following features:

 * Print messages to different levels such as `info` and `error`
 * Auto-rolling of log files
 * Setting the format of log messages
 * Specifying a program name in conjunction with the message

## Installation

Add this line to your application's Gemfile:

```ruby
gem 'logger'
```

And then execute:

    $ bundle

Or install it yourself as:

    $ gem install logger

## Usage

### Simple Example

```ruby
require 'logger'

# Create a Logger that prints to STDOUT
log = Logger.new(STDOUT)
log.debug("Created Logger")

log.info("Program finished")

# Create a Logger that prints to STDERR
error_log = Logger.new(STDERR)
error_log = error_log.error("fatal error")
```

## Development

After checking out the repo, run the following to install dependencies.

```
$ bin/setup
```

Then, run the tests as:

```
$ rake test
```

To install this gem onto your local machine, run

```
$ rake install
```

To release a new version, update the version number in `lib/logger/version.rb`, and then run

```
$ rake release
```

which will create a git tag for the version, push git commits and tags, and push the `.gem` file to [rubygems.org](https://rubygems.org).

## Advanced Development

### Run tests of a specific file

```
$ ruby test/logger/test_logger.rb
```

### Run tests filtering test methods by a name

`--name` option is available as:

```
$ ruby test/logger/test_logger.rb --name test_lshift
```

### Publish documents to GitHub Pages

```
$ rake gh-pages
```

Then, git commit and push the generated HTMLs onto `gh-pages` branch.

## Contributing

Bug reports and pull requests are welcome on GitHub at https://github.com/ruby/logger.

## License

The gem is available as open source under the terms of the [BSD-2-Clause](LICENSE.txt).

Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2023

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: README-LICENCE-utf8.txt

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/

## ----pkgchecks, echo = FALSE--------------------------------------------------
## check if other logger packages are available and exit if not
for (pkg in c("futile.logger", "logging", "log4r")) {
  if (!requireNamespace(pkg, quietly = TRUE)) {
    warning(paste(pkg, "package not available, so cannot build this vignette"))
    knitr::knit_exit()
  }
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(logger)
log_appender(appender_stdout)

## -----------------------------------------------------------------------------
library(logger)
log_info("Loading data")
data(mtcars)
log_info("The dataset includes {nrow(mtcars)} rows")
if (max(mtcars$hp) < 1000) {
  log_warn("Oh, no! There are no cars with more than 1K horsepower in the dataset :/")
  log_debug("The most powerful car is {rownames(mtcars)[which.max(mtcars$hp)]} with {max(mtcars$hp)} hp")
}

## -----------------------------------------------------------------------------
log_threshold()

## -----------------------------------------------------------------------------
log_threshold(TRACE)

## -----------------------------------------------------------------------------
log_info("Loading data")
data(mtcars)
log_info("The dataset includes {nrow(mtcars)} rows")
if (max(mtcars$hp) < 1000) {
  log_warn("Oh, no! There are no cars with more than 1K horsepower in the dataset :/")
  log_debug("The most powerful car is {rownames(mtcars)[which.max(mtcars$hp)]} with {max(mtcars$hp)} hp")
}

## -----------------------------------------------------------------------------
f <- sqrt
g <- mean
x <- 1:31
log_eval(y <- f(g(x)), level = INFO)
str(y)

## ----knitr-pander-setup, include = FALSE--------------------------------------
ppo1 <- pander::panderOptions("knitr.auto.asis")
ppo2 <- pander::panderOptions("table.style")
pander::panderOptions("knitr.auto.asis", FALSE)
pander::panderOptions("table.style", "simple")

## -----------------------------------------------------------------------------
log_formatter(formatter_pander)
log_info(head(iris))

## ----knitr-pander-revert, include = FALSE-------------------------------------
pander::panderOptions("knitr.auto.asis", ppo1)
pander::panderOptions("table.style", ppo2)

## ----cleanup, include = FALSE-------------------------------------------------
logger:::namespaces_reset()

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(logger)
log_appender(appender_stdout)

## ----loggerStructureImage, echo=FALSE, out.extra='style="width: 100%;" title="The structure of a logger and the flow of a log record request" alt="The structure of a logger and the flow of a log record request"'----
knitr::include_graphics("logger_structure.svg")

## -----------------------------------------------------------------------------
f <- function() get_logger_meta_variables(log_level = INFO)
f()

## -----------------------------------------------------------------------------
log_threshold()
ERROR <= INFO
log_error("Oops")

## -----------------------------------------------------------------------------
formatter <- function(...) paste(..., collapse = " ", sep = " ")
formatter(1:3, c("foo", "bar"))

## -----------------------------------------------------------------------------
appender <- function(line) cat(line, "\n")
appender("INFO [now] I am a log message")

## -----------------------------------------------------------------------------
log_threshold(INFO)
log_formatter(formatter_glue)
log_layout(layout_simple)
log_appender(appender_stdout)
log_debug("I am a low level log message that will not be printed with a high log level threshold")
log_warn("I am a higher level log message that is very likely to be printed")

## ----cleanup, include = FALSE-------------------------------------------------
logger:::namespaces_reset()

## ----pkgchecks, echo = FALSE--------------------------------------------------
## check if other logger packages are available and exit if not
for (pkg in c("devtools")) {
  if (!requireNamespace(pkg, quietly = TRUE)) {
    warning(paste(pkg, "package not available, so cannot build this vignette"))
    knitr::knit_exit()
  }
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## -----------------------------------------------------------------------------
library(logger)
devtools::load_all(system.file("demo-packages/logger-tester-package", package = "logger"))
logger_tester_function(INFO, "hi from tester package")

## ----cleanup, include = FALSE-------------------------------------------------
logger:::namespaces_reset()

#' Testing logging from package
#' @param level foo
#' @param msg bar
#' @export
#' @importFrom logger log_level
logger_tester_function <- function(level, msg) {
  set.seed(1014)
  x <- runif(1)
  log_level(level, "{msg} {x}")
}

#' Testing logging INFO from package
#' @param msg bar
#' @export
#' @importFrom logger log_info
logger_info_tester_function <- function(msg) {
  everything <- 42
  log_info("{msg} {everything}")
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(logger)
## backup settings
oldconf <- list(
    threshold = log_threshold(),
    layout = log_layout(),
    formatter = log_formatter(),
    appender = log_appender())
## knitr not picking up stderr
log_appender(appender_stdout)

## -----------------------------------------------------------------------------
library(logger)
log_info('Loading data')
data(mtcars)
log_info('The dataset includes {nrow(mtcars)} rows')
if (max(mtcars$hp) < 1000) {
    log_warn('Oh, no! There are no cars with more than 1K horsepower in the dataset :/')
    log_debug('The most powerful car is {rownames(mtcars)[which.max(mtcars$hp)]} with {max(mtcars$hp)} hp')
}

## -----------------------------------------------------------------------------
log_threshold()

## -----------------------------------------------------------------------------
log_threshold(TRACE)

## -----------------------------------------------------------------------------
log_info('Loading data')
data(mtcars)
log_info('The dataset includes {nrow(mtcars)} rows')
if (max(mtcars$hp) < 1000) {
    log_warn('Oh, no! There are no cars with more than 1K horsepower in the dataset :/')
    log_debug('The most powerful car is {rownames(mtcars)[which.max(mtcars$hp)]} with {max(mtcars$hp)} hp')
}

## -----------------------------------------------------------------------------
f <- sqrt
g <- mean
x <- 1:31
log_eval(y <- f(g(x)), level = INFO)
str(y)

## ----cleanup, include = FALSE-------------------------------------------------
## restore settings
log_threshold(oldconf$threshold)
log_layout(oldconf$layout)
log_formatter(oldconf$formatter)
log_appender(oldconf$appender)

log_levels_supported <- c("OFF", "FATAL", "ERROR", "WARN", "SUCCESS", "INFO", "DEBUG", "TRACE")

#' Log levels
#'
#' The standard Apache logj4 log levels plus a custom level for
#' `SUCCESS`. For the full list of these log levels and suggested
#' usage, check the below Details.
#'
#' List of supported log levels:
#'
#' *  `OFF` No events will be logged
#' *  `FATAL` Severe error that will prevent the application from continuing
#' *  `ERROR` An error in the application, possibly recoverable
#' *  `WARN` An event that might possible lead to an error
#' *  `SUCCESS` An explicit success event above the INFO level that you want to log
#' *  `INFO` An event for informational purposes
#' *  `DEBUG` A general debugging event
#' *  `TRACE` A fine-grained debug message, typically capturing the flow through the application.
#' @references <https://logging.apache.org/log4j/2.x/javadoc/log4j-api/org/apache/logging/log4j/Level.html>,
#'   <https://logging.apache.org/log4j/2.x/manual/customloglevels.html>
#' @name log_levels
NULL

#' @rdname log_levels
#' @export
#' @format NULL
OFF <- structure(0L, level = "OFF", class = c("loglevel", "integer"))
#' @export
#' @rdname log_levels
#' @format NULL
FATAL <- structure(100L, level = "FATAL", class = c("loglevel", "integer"))
#' @export
#' @rdname log_levels
#' @format NULL
ERROR <- structure(200L, level = "ERROR", class = c("loglevel", "integer"))
#' @export
#' @rdname log_levels
#' @format NULL
WARN <- structure(300L, level = "WARN", class = c("loglevel", "integer"))
#' @export
#' @rdname log_levels
#' @format NULL
SUCCESS <- structure(350L, level = "SUCCESS", class = c("loglevel", "integer"))
#' @export
#' @rdname log_levels
#' @format NULL
INFO <- structure(400L, level = "INFO", class = c("loglevel", "integer"))
#' @export
#' @rdname log_levels
#' @format NULL
DEBUG <- structure(500L, level = "DEBUG", class = c("loglevel", "integer"))
#' @export
#' @rdname log_levels
#' @format NULL
TRACE <- structure(600L, level = "TRACE", class = c("loglevel", "integer"))

#' @export
print.loglevel <- function(x, ...) {
  cat("Log level: ", attr(x, "level"), "\n", sep = "")
}


#' Convert R object into a logger log-level
#' @param x string or integer
#' @return pander log-level, e.g. `INFO`
#' @export
#' @examples
#' as.loglevel(INFO)
#' as.loglevel(400L)
#' as.loglevel(400)
as.loglevel <- function(x) { # nolint
  UseMethod("as.loglevel", x)
}


#' @export
as.loglevel.default <- function(x) {
  stop(paste(
    "Do not know how to convert",
    shQuote(class(x)[1]),
    "to a logger log-level."
  ))
}


#' @export
as.loglevel.character <- function(x) {
  stopifnot(
    length(x) == 1,
    x %in% log_levels_supported
  )
  getFromNamespace(x, "logger")
}


#' @export
as.loglevel.integer <- function(x) {
  loglevels <- mget(log_levels_supported, envir = asNamespace("logger"))
  stopifnot(
    length(x) == 1,
    x %in% as.integer(loglevels)
  )
  loglevels[[which(loglevels == x)]]
}


#' @export
as.loglevel.numeric <- as.loglevel.integer

## init storage for all logger settings
namespaces <- new.env(parent = emptyenv())

.onLoad <- function(libname, pkgname) {
  namespaces_set(namespaces_default())
}

namespaces_reset <- function() {
  rm(list = ls(namespaces), envir = namespaces)
  namespaces_set(namespaces_default())
}

namespaces_default <- function() {
  has_glue <- requireNamespace("glue", quietly = TRUE)
  is_running_gha <- isTRUE(Sys.getenv("GITHUB_ACTIONS") == "true") && !needs_stdout()

  list(
    global = list(
      default = list(
        threshold = as.loglevel(Sys.getenv("LOGGER_LOG_LEVEL", unset = "INFO")),
        layout    = if (is_running_gha) layout_gha else layout_simple,
        formatter = if (has_glue) formatter_glue else formatter_sprintf,
        appender  = if (needs_stdout()) appender_stdout else appender_console
      )
    ),
    .logger = list(
      default = list(
        threshold = ERROR,
        layout    = if (is_running_gha) layout_gha else layout_simple,
        formatter = formatter_sprintf,
        appender  = if (needs_stdout()) appender_stdout else appender_console
      )
    )
  )
}

namespaces_set <- function(new = namespaces_default()) {
  old <- as.list(namespaces)

  rm(list = ls(namespaces), envir = namespaces)
  list2env(new, namespaces)

  invisible(old)
}

.onAttach <- function(libname, pkgname) {
  ## warn user about using sprintf instead of glue due to missing dependency
  if (!requireNamespace("glue", quietly = TRUE)) {
    packageStartupMessage(
      paste(
        'logger: As the "glue" R package is not installed,',
        'using "sprintf" as the default log message formatter instead of "glue".'
      )
    )
  }
}

logger_meta_env <- function(log_level = NULL,
                            namespace = NA_character_,
                            .logcall = sys.call(),
                            .topcall = sys.call(-1),
                            .topenv = parent.frame(),
                            .timestamp = Sys.time(),
                            parent = emptyenv()) {
  env <- new.env(parent = parent)
  env$ns <- namespace
  env$ans <- fallback_namespace(namespace)

  force(.topcall)
  force(.topenv)
  delayedAssign("fn", deparse_to_one_line(.topcall[[1]]), assign.env = env)
  delayedAssign("call", deparse_to_one_line(.topcall), assign.env = env)
  delayedAssign("topenv", top_env_name(.topenv), assign.env = env)
  delayedAssign("location", log_call_location(.logcall), assign.env = env)

  env$time <- .timestamp
  env$levelr <- log_level
  env$level <- attr(log_level, "level")

  delayedAssign("pid", Sys.getpid(), assign.env = env)

  # R and ns package versions
  delayedAssign(
    "ns_pkg_version",
    tryCatch(as.character(packageVersion(namespace)), error = function(e) NA_character_),
    assign.env = env
  )
  delayedAssign("r_version", as.character(getRversion()), assign.env = env)

  # stuff from Sys.info
  delayedAssign(".sysinfo", Sys.info())
  delayedAssign("node", .sysinfo[["nodename"]], assign.env = env)
  delayedAssign("arch", .sysinfo[["machine"]], assign.env = env)
  delayedAssign("os_name", .sysinfo[["sysname"]], assign.env = env)
  delayedAssign("os_release", .sysinfo[["release"]], assign.env = env)
  delayedAssign("os_version", .sysinfo[["version"]], assign.env = env)
  delayedAssign("user", .sysinfo[["user"]], assign.env = env)

  env
}

#' Collect useful information about the logging environment to be used in log messages
#'
#' Available variables to be used in the log formatter functions, eg in [layout_glue_generator()]:
#'
#' * `levelr`: log level as an R object, eg [INFO()]
#' * `level`: log level as a string, eg [INFO()]
#' * `time`: current time as `POSIXct`
#' * `node`: name by which the machine is known on the network as reported by `Sys.info`
#' * `arch`: machine type, typically the CPU architecture
#' * `os_name`: Operating System's name
#' * `os_release`: Operating System's release
#' * `os_version`: Operating System's version
#' * `user`: name of the real user id as reported by `Sys.info`
#' * `pid`: the process identification number of the R session
#' * `node`: name by which the machine is known on the network as reported by `Sys.info`
#' * `r_version`: R's major and minor version as a string
#' * `ns`: namespace usually defaults to `global` or the name of the holding R package
#'   of the calling the logging function
#' * `ns_pkg_version`: the version of `ns` when it's a package
#' * `ans`: same as `ns` if there's a defined [logger()] for the namespace,
#'   otherwise a fallback namespace (eg usually `global`)
#' * `topenv`: the name of the top environment from which the parent call was called
#'   (eg R package name or `GlobalEnv`)
#' * `call`: parent call (if any) calling the logging function
#' * `location`: A list with element `path` and `line` giving the location of the
#'   log call
#' * `fn`: function's (if any) name calling the logging function
#'
#' @param log_level log level as per [log_levels()]
#' @inheritParams log_level
#' @return list
#' @export
#' @importFrom utils packageVersion
#' @seealso [layout_glue_generator()]
#' @family log_layouts
get_logger_meta_variables <- function(log_level = NULL,
                                      namespace = NA_character_,
                                      .logcall = sys.call(),
                                      .topcall = sys.call(-1),
                                      .topenv = parent.frame(),
                                      .timestamp = Sys.time()) {
  sysinfo <- Sys.info()

  list(
    ns = namespace,
    ans = fallback_namespace(namespace),
    topenv = top_env_name(.topenv),
    fn = deparse_to_one_line(.topcall[[1]]),
    call = deparse_to_one_line(.topcall),
    location = log_call_location(.logcall),
    time = .timestamp,
    levelr = log_level,
    level = attr(log_level, "level"),
    pid = Sys.getpid(),

    ## R and ns package versions
    r_version = paste0(R.Version()[c("major", "minor")], collapse = "."),
    ns_pkg_version = tryCatch(as.character(packageVersion(namespace)), error = function(e) NA_character_),

    ## stuff from Sys.info
    node = sysinfo[["nodename"]],
    arch = sysinfo[["machine"]],
    os_name = sysinfo[["sysname"]],
    os_release = sysinfo[["release"]],
    os_version = sysinfo[["version"]],
    user = sysinfo[["user"]]
    ## NOTE might be better to rely on the whoami pkg?

    ## TODO jenkins (or any) env vars => no need to get here, users can write custom layouts
    ## TODO seed
  )
}


#' Generate log layout function using common variables available via glue syntax
#'
#' `format` is passed to [glue::glue()] with access to the below variables:
#' \itemize{ \item msg: the actual log message \item further variables
#' set by [get_logger_meta_variables()] }
#' @param format [glue::glue()]-flavored layout of the message using the above
#'   variables
#' @return function taking `level` and `msg` arguments - keeping the
#'   original call creating the generator in the `generator` attribute
#'   that is returned when calling [log_layout()] for the currently
#'   used layout
#' @export
#' @examples
#' \dontshow{old <- logger:::namespaces_set()}
#' example_layout <- layout_glue_generator(
#'   format = "{node}/{pid}/{ns}/{ans}/{topenv}/{fn} {time} {level}: {msg}"
#' )
#' example_layout(INFO, "try {runif(1)}")
#'
#' log_layout(example_layout)
#' log_info("try {runif(1)}")
#' \dontshow{logger:::namespaces_set(old)}
#' @seealso See example calls from [layout_glue()] and [layout_glue_colors()].
#' @family log_layouts
layout_glue_generator <- function(format = '{level} [{format(time, "%Y-%m-%d %H:%M:%S")}] {msg}') {
  force(format)

  structure(function(level,
                     msg,
                     namespace = NA_character_,
                     .logcall = sys.call(),
                     .topcall = sys.call(-1),
                     .topenv = parent.frame(),
                     .timestamp = Sys.time()) {
    fail_on_missing_package("glue")
    if (!inherits(level, "loglevel")) {
      stop("Invalid log level, see ?log_levels")
    }

    meta <- logger_meta_env(
      log_level = level,
      namespace = namespace,
      .logcall = .logcall,
      .topcall = .topcall,
      .topenv = .topenv,
      .timestamp = .timestamp,
      parent = environment()
    )
    glue::glue(format, .envir = meta)
  }, generator = deparse(match.call()))
}


#' Format a log record by including the raw message without anything
#' added or modified
#' @inheritParams log_level
#' @param msg string message
#' @return character vector
#' @export
#' @family log_layouts
layout_blank <- function(level,
                         msg,
                         namespace = NA_character_,
                         .logcall = sys.call(),
                         .topcall = sys.call(-1),
                         .topenv = parent.frame(),
                         .timestamp = Sys.time()) {
  msg
}
attr(layout_blank, "generator") <- quote(layout_blank())

#' Format a log record by concatenating the log level, timestamp and
#' message
#' @inheritParams log_level
#' @param msg string message
#' @return character vector
#' @export
#' @family log_layouts
layout_simple <- function(level,
                          msg,
                          namespace = NA_character_,
                          .logcall = sys.call(),
                          .topcall = sys.call(-1),
                          .topenv = parent.frame(),
                          .timestamp = Sys.time()) {
  paste0(attr(level, "level"), " [", format(.timestamp, "%Y-%m-%d %H:%M:%S"), "] ", msg)
}
attr(layout_simple, "generator") <- quote(layout_simple())

#' Format a log record as the logging package does by default
#' @inheritParams layout_simple
#' @param msg string message
#' @return character vector
#' @export
#' @family log_layouts
#' @examples
#' \dontshow{old <- logger:::namespaces_set()}
#' log_layout(layout_logging)
#' log_info(42)
#' log_info(42, namespace = "everything")
#'
#' \dontrun{
#' devtools::load_all(system.file("demo-packages/logger-tester-package", package = "logger"))
#' logger_tester_function(INFO, 42)
#' }
#' \dontshow{logger:::namespaces_set(old)}
layout_logging <- function(level,
                           msg,
                           namespace = NA_character_,
                           .logcall = sys.call(),
                           .topcall = sys.call(-1),
                           .topenv = parent.frame(),
                           .timestamp = Sys.time()) {
  meta <- logger_meta_env(
    log_level = level,
    namespace = namespace,
    .logcall = .logcall,
    .topcall = .topcall,
    .topenv = .topenv,
    .timestamp = .timestamp
  )
  paste0(
    format(meta$time, "%Y-%m-%d %H:%M:%S"), " ",
    attr(level, "level"), ":",
    ifelse(meta$ns == "global", "", meta$ns), ":",
    msg
  )
}
attr(layout_logging, "generator") <- quote(layout_logging())

#' Format a log message with [glue::glue()]
#'
#' By default, this layout includes the log level of the log record as
#' per [log_levels()], the current timestamp and the actual log
#' message -- that you can override via calling
#' [layout_glue_generator()] directly. For colorized output, see
#' [layout_glue_colors()].
#' @inheritParams layout_simple
#' @return character vector
#' @export
#' @family log_layouts
layout_glue <- layout_glue_generator()
attr(layout_glue, "generator") <- quote(layout_glue())

#' Format a log message with [glue::glue()] and ANSI escape codes to add colors
#'
#' Colour log levels based on their severity. Log levels are coloured
#' with [colorize_by_log_level()] and the messages are coloured with
#' [grayscale_by_log_level()].
#'
#' @inheritParams layout_simple
#' @return character vector
#' @export
#' @family log_layouts
#' @note This functionality depends on the \pkg{crayon} package.
#' @examplesIf requireNamespace("crayon")
#' log_layout(layout_glue_colors)
#' log_threshold(TRACE)
#' log_info("Starting the script...")
#' log_debug("This is the second line")
#' log_trace("That is being placed right after the first one.")
#' log_warn("Some errors might come!")
#' log_error("This is a problem")
#' log_debug("Getting an error is usually bad")
#' log_error("This is another problem")
#' log_fatal("The last problem.")
layout_glue_colors <- layout_glue_generator(
  format = paste(
    "{crayon::bold(colorize_by_log_level(level, levelr))}",
    '[{crayon::italic(format(time, "%Y-%m-%d %H:%M:%S"))}]',
    "{grayscale_by_log_level(msg, levelr)}"
  )
)
attr(layout_glue_colors, "generator") <- quote(layout_glue_colors())

#' Generate log layout function rendering JSON
#' @param fields character vector of field names to be included in the
#'   JSON
#' @return character vector
#' @export
#' @note This functionality depends on the \pkg{jsonlite} package.
#' @family log_layouts
#' @examples
#' \dontshow{old <- logger:::namespaces_set()}
#' log_layout(layout_json())
#' log_info(42)
#' log_info("ok {1:3} + {1:3} = {2*(1:3)}")
#' \dontshow{logger:::namespaces_set(old)}
layout_json <- function(fields = default_fields()) {
  force(fields)

  if ("msg" %in% fields) {
    warning("'msg' is always automatically included")
    fields <- setdiff(fields, "msg")
  }

  structure(function(level,
                     msg,
                     namespace = NA_character_,
                     .logcall = sys.call(),
                     .topcall = sys.call(-1),
                     .topenv = parent.frame(),
                     .timestamp = Sys.time()) {
    fail_on_missing_package("jsonlite")

    meta <- logger_meta_env(
      log_level = level,
      namespace = namespace,
      .logcall = .logcall,
      .topcall = .topcall,
      .topenv = .topenv,
      .timestamp = .timestamp
    )
    json <- mget(fields, meta)
    sapply(msg, function(msg) jsonlite::toJSON(c(json, list(msg = msg)), auto_unbox = TRUE))
  }, generator = deparse(match.call()))
}


#' Generate log layout function rendering JSON after merging meta
#' fields with parsed list from JSON message
#' @param fields character vector of field names to be included in the
#'   JSON. If named, the names will be used as field names in the JSON.
#' @export
#' @note This functionality depends on the \pkg{jsonlite} package.
#' @family log_layouts
#' @examples
#' \dontshow{old <- logger:::namespaces_set()}
#' log_formatter(formatter_json)
#' log_info(everything = 42)
#'
#' log_layout(layout_json_parser())
#' log_info(everything = 42)
#'
#' log_layout(layout_json_parser(fields = c("time", "node")))
#' log_info(cars = row.names(mtcars), species = unique(iris$Species))
#'
#' log_layout(layout_json_parser(fields = c(timestamp = "time", "node")))
#' log_info(
#'   message = paste(
#'     "Compared to the previous example,
#'     the 'time' field is renamed to 'timestamp'"
#'   )
#' )
#' \dontshow{logger:::namespaces_set(old)}
#' @importFrom stats setNames
layout_json_parser <- function(fields = default_fields()) {
  force(fields)

  structure(function(level,
                     msg,
                     namespace = NA_character_,
                     .logcall = sys.call(),
                     .topcall = sys.call(-1),
                     .topenv = parent.frame(),
                     .timestamp = Sys.time()) {
    fail_on_missing_package("jsonlite")

    meta <- logger_meta_env(
      log_level = level,
      namespace = namespace,
      .logcall = .logcall,
      .topcall = .topcall,
      .topenv = .topenv,
      .timestamp = .timestamp
    )
    meta <- mget(fields, meta)
    field_names <- names(fields)
    if (!is.null(field_names)) {
      norename <- field_names == ""
      field_names[norename] <- fields[norename]
      meta <- setNames(meta, field_names)
    }
    msg <- jsonlite::fromJSON(msg)

    jsonlite::toJSON(c(meta, msg), auto_unbox = TRUE, null = "null")
  }, generator = deparse(match.call()))
}

default_fields <- function() {
  c(
    "time", "level", "ns", "ans", "topenv", "fn", "node", "arch",
    "os_name", "os_release", "os_version", "pid", "user"
  )
}

#' Format a log record for github actions
#'
#' GitHub Actions can recognise specially formatted output and make these
#' prominent in the output. The `layout_gha()` layout will ensure the correct
#' formatting when running in GitHub Actions.
#'
#' @note
#' GitHub Actions only recognise the log levels `error`, `warning`, `notice`,
#' and `debug`. Because of this, `FATAL` and `ERROR` are coerced to `error`,
#' `SUCCESS` and `INFO` are coerced to `notice`, and `DEBUG` and `TRACE` are
#' coerced to `debug` (`WARN` maps directly to `warning`).
#'
#' @inheritParams layout_simple
#' @return character vector
#' @export
#' @family log_layouts
#'
layout_gha <- structure(
  function(level,
           msg,
           namespace = NA_character_,
           .logcall = sys.call(),
           .topcall = sys.call(-1),
           .topenv = parent.frame(),
           .timestamp = Sys.time()) {
    level <- attr(level, "level")
    gha_level <- switch(level,
      FATAL = ,
      ERROR = "error",
      WARN = "warning",
      SUCCESS = ,
      INFO = "notice",
      DEBUG = ,
      TRACE = "debug",
      "notice" # Defaults to notice if for some reason another level gets used
    )
    title <- if (gha_level == "debug") "" else paste0(" title=", level)
    paste0("::", gha_level, title, "::", msg)
  },
  generator = quote(layout_gha())
)

# nocov start
#' Format a log record for syslognet
#'
#' Format a log record for syslognet.
#' This function converts the logger log level to a
#' log severity level according to RFC 5424 "The Syslog Protocol".
#'
#' @inheritParams layout_simple
#' @return A character vector with a severity attribute.
#' @export
layout_syslognet <- structure(
  function(level,
           msg,
           namespace = NA_character_,
           .logcall = sys.call(),
           .topcall = sys.call(-1),
           .topenv = parent.frame(),
           .timestamp = Sys.time()) {
    ret <- paste(attr(level, "level"), msg)
    attr(ret, "severity") <- switch(
      attr(level, "level", exact = TRUE),
      "FATAL" = "CRITICAL",
      "ERROR" = "ERR",
      "WARN" = "WARNING",
      "SUCCESS" = "NOTICE",
      "INFO" = "INFO",
      "DEBUG" = "DEBUG",
      "TRACE" = "DEBUG"
    )
    ret
  },
  generator = quote(layout_syslognet())
)
# nocov end

test_that("%except% logs errors and returns default value", {
  local_test_logger(layout = layout_glue_generator("{ns} / {ans} / {topenv} / {fn} / {call}\n{level} {msg}"))

  f <- function() {
    FunDoesNotExist(1:10) %except% 1
  }

  expect_snapshot(out <- f())
  expect_equal(out, 1)
})

test_that("%except% returns value when no error", {
  expect_equal(5 %except% 1, 5)
})

test_that("single line", {
  local_test_logger(layout = layout_glue_generator("{level} {msg}"))

  expect_output(log_eval(4, INFO), sprintf("INFO %s => %s", shQuote(4), shQuote(4)))
})

test_that("multi line", {
  local_test_logger(layout = layout_glue_generator("{level} {msg}"))

  expect_output(log_eval(4, INFO, multiline = TRUE), "Running expression")
  expect_output(log_eval(4, INFO, multiline = TRUE), "Results:")
  expect_output(log_eval(4, INFO, multiline = TRUE), "INFO 4")
})

test_that("invisible return", {
  local_test_logger(layout = layout_glue_generator("{level} {msg}"))
  expect_output(log_eval(require(logger), INFO), sprintf(
    "INFO %s => %s",
    shQuote("require\\(logger\\)"),
    shQuote(TRUE)
  ))
})

test_that("lower log level", {
  local_test_logger(TRACE, layout = layout_glue_generator("{level} {msg}"))
  expect_output(log_eval(4), sprintf("TRACE %s => %s", shQuote(4), shQuote(4)))
})

test_that("blank layout", {
  local_test_logger(layout = layout_blank)
  expect_output(log_info("foobar"), "foobar")
  expect_equal(capture.output(log_info("foobar")), "foobar")
})

test_that("colorized layout", {
  local_test_logger(layout = layout_glue_colors)
  expect_output(log_info("foobar"), "INFO")
  expect_output(log_info("foobar"), "foobar")
  expect_output(log_error("foobar"), "ERROR")
  expect_output(log_error("foobar"), "foobar")
})

test_that("metavars", {
  local_test_logger(layout = layout_glue_generator("{level} {ans} {fn}"))

  f_info <- function() log_info()
  expect_output(f_info(), "INFO global f_info()")

  f_warn <- function() log_warn()
  expect_output(f_warn(), "WARN global f_warn()")

  local_test_logger(layout = layout_glue_generator("{time}"))
  time <- as.POSIXct("2025-04-04 10:31:57 CEST")
  expect_output(log_info("test", .timestamp = time), "2025-04-04 10:31:57")
})

test_that("JSON layout", {
  local_test_logger(layout = layout_json(fields = "level"))

  out <- jsonlite::fromJSON(capture.output(log_info("foobar")))
  expect_equal(out, list(level = "INFO", msg = "foobar"))
})

test_that("JSON layout warns if you include msg", {
  expect_snapshot(layout <- layout_json(fields = "msg"))
  local_test_logger(layout = layout)
  out <- jsonlite::fromJSON(capture.output(log_info("foobar")))
  expect_equal(out, list(msg = "foobar"))
})

test_that("JSON parser layout", {
  local_test_logger(layout = layout_json_parser(fields = character()))
  expect_output(log_info(skip_formatter('{"x": 4}')), '{"x":4}', fixed = TRUE)
})

test_that("JSON parser layout can be renamed", {
  local_test_logger(layout = layout_json_parser(c(LEVEL = "level")))
  expect_output(log_info(skip_formatter('{"x": 4}')), '{"LEVEL":"INFO","x":4}', fixed = TRUE)
})

test_that("must throw errors", {
  skip_if_not(getRversion() >= "4.3") # error call changed

  expect_snapshot(error = TRUE, {
    layout_simple(FOOBAR)
    layout_simple(42)
    layout_simple(msg = "foobar")
  })

  expect_snapshot(error = TRUE, {
    layout_glue(FOOBAR)
    layout_glue(42)
    layout_glue(msg = "foobar")
    layout_glue(level = 53, msg = "foobar")
  })
})

test_that("logging layout", {
  local_test_logger(layout = layout_logging)
  expect_output(log_level(INFO, "foo", namespace = "bar"), "INFO:bar:foo")
  expect_output(log_info("foobar"), "INFO")
  expect_output(log_info("foo", namespace = "bar"), "foo")
  expect_output(log_info("foo", namespace = "bar"), "bar")
  expect_output(log_info("foo", namespace = "bar"), "INFO:bar:foo")
})

test_that("log_info() captures local info", {
  local_test_logger(
    layout = layout_glue_generator("{ns} / {ans} / {topenv} / {fn} / {call}")
  )
  f <- function() log_info("foobar")
  g <- function() f()

  expect_snapshot({
    log_info("foobar")
    f()
    g()
  })
})

test_that("log_info() captures package info", {
  devtools::load_all(
    system.file("demo-packages/logger-tester-package", package = "logger"),
    quiet = TRUE
  )
  withr::defer(devtools::unload("logger.tester"))

  local_test_logger(layout = layout_glue_generator("{ns} {level} {msg}"))
  expect_snapshot({
    logger_tester_function(INFO, "x = ")
    logger_info_tester_function("everything = ")
  })
})

#' Check if R package can be loaded and fails loudly otherwise
#' @param pkg string
#' @param min_version optional minimum version needed
#' @param call Call to include in error message.
#' @export
#' @importFrom utils packageVersion compareVersion
#' @examples
#' f <- function() fail_on_missing_package("foobar")
#' try(f())
#' g <- function() fail_on_missing_package("stats")
#' g()
fail_on_missing_package <- function(pkg, min_version, call = NULL) {
  pc <- call %||% sys.call(which = 1)
  if (!requireNamespace(pkg, quietly = TRUE)) {
    stop(
      sprintf(
        "Please install the '%s' package to use %s",
        pkg,
        deparse(pc[[1]])
      ),
      call. = FALSE
    )
  }
  if (!missing(min_version)) {
    if (compareVersion(min_version, as.character(packageVersion(pkg))) == 1) {
      stop(
        sprintf(
          "Please install min. %s version of %s to use %s",
          min_version,
          pkg,
          deparse(pc[[1]])
        ),
        call. = FALSE
      )
    }
  }
}


#' Returns the name of the top level environment from which the logger was called
#' @return string
#' @noRd
#' @param .topenv call environment
top_env_name <- function(.topenv = parent.frame()) {
  environmentName(topenv(.topenv))
}

#' Finds the location of the logger call (file and line)
#' @return list with path and line element
#' @noRd
#' @param .logcall The call that emitted the log
log_call_location <- function(.logcall) {
  call_string <- deparse(.logcall)
  loc <- list(
    path = "<console>",
    line = ""
  )
  for (trace in .traceback(0)) {
    if (identical(call_string, as.vector(trace))) {
      ref <- attr(trace, "srcref")
      loc$line <- ref[1L]
      file <- attr(ref, "srcfile")
      if (!is.null(file)) {
        loc$path <- normalizePath(file$filename, winslash = "/", mustWork = FALSE)
      }
      break
    }
  }
  loc
}

#' Deparse and join all lines into a single line
#'
#' Calling `deparse` and joining all the returned lines into a
#' single line, separated by whitespace, and then cleaning up all the
#' duplicated whitespace (except for excessive whitespace in strings
#' between single or double quotes).
#' @param x object to `deparse`
#' @return string
#' @export
deparse_to_one_line <- function(x) {
  gsub('\\s+(?=(?:[^\\\'"]*[\\\'"][^\\\'"]*[\\\'"])*[^\\\'"]*$)', " ",
    paste(deparse(x), collapse = " "),
    perl = TRUE
  )
}


#' Catch the log header
#' @return string
#' @param level see [log_levels()]
#' @param namespace string
#' @noRd
#' @examples
#' \dontshow{old <- logger:::namespaces_set()}
#' catch_base_log(INFO, NA_character_)
#' logger <- layout_glue_generator(format = "{node}/{pid}/{namespace}/{fn} {time} {level}: {msg}")
#' log_layout(logger)
#' catch_base_log(INFO, NA_character_)
#' fun <- function() catch_base_log(INFO, NA_character_)
#' fun()
#' catch_base_log(INFO, NA_character_, .topcall = call("funLONG"))
#' \dontshow{logger:::namespaces_set(old)}
catch_base_log <- function(level, namespace, .topcall = sys.call(-1), .topenv = parent.frame()) {
  namespace <- fallback_namespace(namespace)

  old <- log_appender(appender_console, namespace = namespace)
  on.exit(log_appender(old, namespace = namespace))

  # catch error, warning or message
  capture.output(
    log_level(
      level = level,
      "",
      namespace = namespace,
      .topcall = .topcall,
      .topenv = .topenv
    ),
    type = "message"
  )
}

`%||%` <- function(x, y) {
  if (is.null(x)) y else x
}

in_pkgdown <- function() {
  identical(Sys.getenv("IN_PKGDOWN"), "true")
}

is_testing <- function() {
  identical(Sys.getenv("TESTTHAT"), "true")
}

is_checking_logger <- function() {
  Sys.getenv("_R_CHECK_PACKAGE_NAME_", "") == "logger"
}

needs_stdout <- function() {
  in_pkgdown() || is_testing() || is_checking_logger()
}

# allow mocking
Sys.time <- NULL # nolint
proc.time <- NULL # nolint

test_that("fail_on_missing_package", {
  expect_no_error(fail_on_missing_package("logger"))

  expect_snapshot(error = TRUE, {
    fail_on_missing_package("logger", "9.9.9", call = quote(f()))
    fail_on_missing_package("an.R.package-that-doesNotExists", call = quote(f()))
  })
})

test_that("validate_log_level", {
  expect_equal(validate_log_level(ERROR), ERROR)
  expect_equal(validate_log_level("ERROR"), ERROR)
  expect_snapshot(validate_log_level("FOOBAR"), error = TRUE)
})

test_that("catch_base_log", {
  local_test_logger(layout = layout_simple)
  expect_true(nchar(catch_base_log(ERROR, NA_character_)) == 28)
  expect_true(nchar(catch_base_log(INFO, NA_character_)) == 27)
  local_test_logger(layout = layout_blank)
  expect_true(nchar(catch_base_log(INFO, NA_character_)) == 0)

  local_test_logger(layout = layout_glue_generator(format = "{namespace}/{fn}"))
  expect_equal(catch_base_log(INFO, "TEMP", .topcall = NA), "global/NA")
  expect_equal(catch_base_log(INFO, "TEMP", .topcall = quote(f())), "global/f")
})

#' Evaluate an expression and log results
#' @param expr R expression to be evaluated while logging the
#'   expression itself along with the result
#' @param level [log_levels()]
#' @param multiline setting to `FALSE` will print both the expression
#'   (enforced to be on one line by removing line-breaks if any) and
#'   its result on a single line separated by `=>`, while setting to
#'   `TRUE` will log the expression and the result in separate
#'   sections reserving line-breaks and rendering the printed results
#' @examples
#' \dontshow{old <- logger:::namespaces_set()}
#' log_eval(pi * 2, level = INFO)
#'
#' ## lowering the log level threshold so that we don't have to set a higher level in log_eval
#' log_threshold(TRACE)
#' log_eval(x <- 4)
#' log_eval(sqrt(x))
#'
#' ## log_eval can be called in-line as well as returning the return value of the expression
#' x <- log_eval(mean(runif(1e3)))
#' x
#'
#' ## https://twitter.com/krlmlr/status/1067864829547999232
#' f <- sqrt
#' g <- mean
#' x <- 1:31
#' log_eval(f(g(x)), level = INFO)
#' log_eval(y <- f(g(x)), level = INFO)
#'
#' ## returning a function
#' log_eval(f <- sqrt)
#' log_eval(f)
#'
#' ## evaluating something returning a wall of "text"
#' log_eval(f <- log_eval)
#' log_eval(f <- log_eval, multiline = TRUE)
#'
#' ## doing something computationally intensive
#' log_eval(system.time(for (i in 1:100) mad(runif(1000))), multiline = TRUE)
#' \dontshow{logger:::namespaces_set(old)}
#' @importFrom utils capture.output
#' @export
log_eval <- function(expr, level = TRACE, multiline = FALSE) {
  ## capture call
  expr <- substitute(expr)
  exprs <- gsub("\n", " ", deparse(expr), fixed = TRUE)

  ## evaluate call and store results
  timer <- Sys.time()
  res <- withVisible(eval.parent(expr))

  ## log expression and results
  if (multiline == FALSE) {
    log_level(level, skip_formatter(
      paste(
        shQuote(paste(exprs, collapse = " ")),
        "=>",
        shQuote(paste(gsub("\n", " ", deparse(res$value)), collapse = " "))
      )
    ))
  } else {
    log_level(level, "Running expression: ====================")
    log_level(level, skip_formatter(exprs))
    log_level(level, "Results: ===============================")
    log_level(level, skip_formatter(capture.output(res$value)))
    log_level(level, paste(
      "Elapsed time:",
      round(difftime(Sys.time(), timer, units = "secs"), 2),
      "sec"
    ))
  }

  ## return the results of the call
  if (res$visible == TRUE) {
    res$value
  } else {
    invisible(res$value)
  }
}


#' Logs a long line to stand out from the console
#' @inheritParams log_level
#' @param separator character to be used as a separator
#' @param width max width of message -- longer text will be wrapped into multiple lines
#' @export
#' @examples
#' \dontshow{old <- logger:::namespaces_set()}
#' log_separator()
#' log_separator(ERROR, separator = "!", width = 60)
#' log_separator(ERROR, separator = "!", width = 100)
#' logger <- layout_glue_generator(format = "{node}/{pid}/{namespace}/{fn} {time} {level}: {msg}")
#' log_layout(logger)
#' log_separator(ERROR, separator = "!", width = 100)
#' log_layout(layout_blank)
#' log_separator(ERROR, separator = "!", width = 80)
#' \dontshow{logger:::namespaces_set(old)}
#' @seealso [log_with_separator()]
log_separator <- function(level = INFO,
                          namespace = NA_character_,
                          separator = "=",
                          width = 80,
                          .logcall = sys.call(),
                          .topcall = sys.call(-1),
                          .topenv = parent.frame(),
                          .timestamp = Sys.time()) {
  stopifnot(length(separator) == 1, nchar(separator) == 1)

  base_info_chars <- nchar(catch_base_log(level, namespace, .topcall = .topcall, .topenv = .topenv))

  log_level(
    paste(rep(separator, max(0, width - base_info_chars)), collapse = ""),
    level = level,
    namespace = namespace,
    .logcall = .logcall,
    .topcall = .topcall,
    .topenv = .topenv,
    .timestamp = .timestamp
  )
}


#' Logs a message in a very visible way
#' @inheritParams log_level
#' @inheritParams log_separator
#' @export
#' @examples
#' \dontshow{old <- logger:::namespaces_set()}
#' log_with_separator("An important message")
#' log_with_separator("Some critical KPI down!!!", separator = "$")
#' log_with_separator("This message is worth a {1e3} words")
#' log_with_separator(paste(
#'   "A very important message with a bunch of extra words that will",
#'   "eventually wrap into a multi-line message for our quite nice demo :wow:"
#' ))
#' log_with_separator(
#'   paste(
#'     "A very important message with a bunch of extra words that will",
#'     "eventually wrap into a multi-line message for our quite nice demo :wow:"
#'   ),
#'   width = 60
#' )
#' log_with_separator("Boo!", level = FATAL)
#' log_layout(layout_blank)
#' log_with_separator("Boo!", level = FATAL)
#' logger <- layout_glue_generator(format = "{node}/{pid}/{namespace}/{fn} {time} {level}: {msg}")
#' log_layout(logger)
#' log_with_separator("Boo!", level = FATAL, width = 120)
#' \dontshow{logger:::namespaces_set(old)}
#' @seealso [log_separator()]
log_with_separator <- function(...,
                               level = INFO,
                               namespace = NA_character_,
                               separator = "=",
                               width = 80) {
  base_info_chars <- nchar(catch_base_log(level, namespace, .topcall = sys.call(-1)))

  log_separator(
    level = level,
    separator = separator,
    width = width,
    namespace = namespace,
    .logcall = sys.call(),
    .topcall = sys.call(-1),
    .topenv = parent.frame()
  )

  message <- do.call(eval(log_formatter()), list(...))
  message <- strwrap(message, max(0, width - base_info_chars - 4))
  message <- sapply(message, function(m) {
    paste0(
      separator, " ", m,
      paste(rep(" ", max(0, width - base_info_chars - 4 - nchar(m))), collapse = ""),
      " ", separator
    )
  })

  log_level(skip_formatter(message), level = level, namespace = namespace, .topenv = parent.frame())

  log_separator(
    level = level,
    separator = separator,
    width = width,
    namespace = namespace,
    .logcall = sys.call(),
    .topcall = sys.call(-1),
    .topenv = parent.frame()
  )
}


#' Tic-toc logging
#' @param ... passed to [log_level()]
#' @param level see [log_levels()]
#' @param namespace x
#' @export
#' @examples
#' log_tictoc("warming up")
#' Sys.sleep(0.1)
#' log_tictoc("running")
#' Sys.sleep(0.1)
#' log_tictoc("running")
#' Sys.sleep(runif(1))
#' log_tictoc("and running")
#' @author Thanks to Neal Fultz for the idea and original implementation!
log_tictoc <- function(..., level = INFO, namespace = NA_character_) {
  ns <- fallback_namespace(namespace)

  on.exit({
    assign(ns, toc, envir = tictocs)
  })

  tic <- get0(ns, envir = tictocs, ifnotfound = Sys.time())
  toc <- Sys.time()
  tictoc <- difftime(toc, tic)

  log_level(
    paste(
      ns, "timer",
      ifelse(round(tictoc, 2) == 0, "tic", "toc"),
      round(tictoc, 2), attr(tictoc, "units"), "-- "
    ),
    ...,
    level = level, namespace = namespace,
    .logcall = sys.call(),
    .topcall = sys.call(-1),
    .topenv = parent.frame()
  )
}
tictocs <- new.env(parent = emptyenv())

#' Log cumulative running time
#'
#' This function is working like [log_tictoc()] but differs in that it continues
#' to count up rather than resetting the timer at every call. You can set the
#' start time using `log_elapsed_start()`, but if that hasn't been called it
#' will show the time since the R session started.
#'
#' @inheritParams log_tictoc
#'
#' @export
#'
#' @examples
#' log_elapsed_start()
#' Sys.sleep(0.4)
#' log_elapsed("Tast 1")
#' Sys.sleep(0.2)
#' log_elapsed("Task 2")
#'
log_elapsed <- function(..., level = INFO, namespace = NA_character_) {
  ns <- fallback_namespace(namespace)

  start <- get0(ns, envir = elapsed, ifnotfound = 0)

  time_elapsed <- as.difftime(proc.time()["elapsed"] - start, units = "secs")

  log_level(
    paste(
      ns, "timer",
      round(time_elapsed, 2), attr(time_elapsed, "units"), "elapsed -- "
    ),
    ...,
    level = level, namespace = namespace,
    .logcall = sys.call(),
    .topcall = sys.call(-1),
    .topenv = parent.frame()
  )
}
#' @rdname log_elapsed
#' @param quiet Should starting the time emit a log message
#' @export
log_elapsed_start <- function(level = INFO, namespace = NA_character_, quiet = FALSE) {
  ns <- fallback_namespace(namespace)

  assign(ns, proc.time()["elapsed"], envir = elapsed)

  if (!quiet) {
    log_level(
      paste(
        "starting", ns, "timer"
      ),
      level = level, namespace = namespace,
      .logcall = sys.call(),
      .topcall = sys.call(-1),
      .topenv = parent.frame()
    )
  }
}
elapsed <- new.env(parent = emptyenv())

#' Logs the error message to console before failing
#' @param expression call
#' @export
#' @examples
#' log_failure("foobar")
#' try(log_failure(foobar))
log_failure <- function(expression) {
  withCallingHandlers(
    expression,
    error = function(e) {
      log_error(conditionMessage(e))
    }
  )
}

#' Automatically log execution time of knitr chunks
#'
#' Calling this function in the first chunk of a document will instruct knitr
#' to automatically log the execution time of each chunk. If using
#' [formatter_glue()] or [formatter_cli()] then the `options` variable will be
#' available, providing the chunk options such as chunk label etc.
#'
#' @inheritParams log_elapsed
#'
#' @export
#'
#' @examples
#' # To be put in the first chunk of a document
#' log_chunk_time("chunk {options$label}")
#'
log_chunk_time <- function(..., level = INFO, namespace = NA_character_) {
  if (!requireNamespace("knitr", quietly = TRUE)) {
    stop("knitr is required to use this functionality", call. = FALSE)
  }
  if (!isTRUE(getOption("knitr.in.progress"))) {
    return(invisible())
  }
  args <- list(...)
  args$level <- level
  args$namespace <- namespace
  knitr::knit_hooks$set(logger_timer = function(before, options) {
    if (before) {
      log_elapsed_start(namespace = namespace, quiet = TRUE)
    } else {
      do.call(log_elapsed, args)
    }
  })
  knitr::opts_chunk$set(logger_timer = TRUE)

  invisible()
}

test_that("separator", {
  local_test_logger(layout = layout_blank)
  expect_output(log_separator(), "={80,80}")

  local_test_logger()
  expect_output(log_separator(separator = "-"), "---")
  expect_output(log_separator(), "INFO")
  expect_output(log_separator(WARN), "WARN")
})

test_that("tictoc", {
  local_test_logger()
  local_mocked_bindings(Sys.time = function() as.POSIXct("2024-01-01 00:00:00"))
  expect_output(log_tictoc(), "timer tic 0 secs")
  ## simulate time passing
  local_mocked_bindings(Sys.time = function() as.POSIXct("2024-01-01 00:01:00"))
  expect_output(log_tictoc(), "timer toc 1 mins")
})

test_that("elapsed", {
  local_test_logger()

  proc_null <- proc.time() * 0

  with_mocked_bindings(
    expect_output(log_elapsed(), "timer 2 secs elapsed"),
    proc.time = function() proc_null + 2
  )

  with_mocked_bindings(
    expect_output(log_elapsed_start(), "starting global timer"),
    proc.time = function() proc_null + 1
  )

  with_mocked_bindings(
    expect_output(log_elapsed(), "timer 1 secs elapsed"),
    proc.time = function() proc_null + 2
  )

  with_mocked_bindings(
    expect_silent(log_elapsed_start(quiet = TRUE)),
    proc.time = function() proc_null
  )
})

test_that("log with separator", {
  local_test_logger()
  expect_output(log_with_separator(42), "===")
  expect_output(log_with_separator("Boo!", level = FATAL, width = 120), width = 120)
})

test_that("log failure", {
  skip_if_not(getRversion() >= "4.3") # error call changed

  local_test_logger()
  expect_output(log_failure("foobar"), NA)
  expect_output(try(log_failure(foobar), silent = TRUE), "ERROR.*foobar")
  expect_no_error(log_failure("foobar"))
  expect_snapshot(capture.output(log_failure(foobar)), error = TRUE)
})

test_that("log with separator", {
  local_test_logger(layout = layout_glue_generator("{level} {msg}"))

  expect_snapshot({
    log_with_separator(42)
    log_with_separator(42, separator = "|")
  })
})


test_that("single line", {
  local_test_logger(layout = layout_glue_generator("{level} {msg}"))

  expect_output(log_eval(4, INFO), sprintf("INFO %s => %s", shQuote(4), shQuote(4)))
})

test_that("multi line", {
  local_test_logger(layout = layout_glue_generator("{level} {msg}"))

  expect_output(log_eval(4, INFO, multiline = TRUE), "Running expression")
  expect_output(log_eval(4, INFO, multiline = TRUE), "Results:")
  expect_output(log_eval(4, INFO, multiline = TRUE), "INFO 4")
})

test_that("invisible return", {
  local_test_logger(layout = layout_glue_generator("{level} {msg}"))
  expect_output(log_eval(require(logger), INFO), sprintf(
    "INFO %s => %s",
    shQuote("require\\(logger\\)"),
    shQuote(TRUE)
  ))
})

test_that("lower log level", {
  local_test_logger(TRACE, layout = layout_glue_generator("{level} {msg}"))
  expect_output(log_eval(4), sprintf("TRACE %s => %s", shQuote(4), shQuote(4)))
})

test_that("knitr hook gets applied", {
  local_test_logger()
  mock_doc <- c(
    "```{r, include = FALSE}",
    "library(logger)",
    "log_chunk_time()",
    "```",
    "```{r}",
    "Sys.sleep(0.2)",
    "```"
  )
  expect_output(knitr::knit(text = mock_doc, quiet = TRUE), "global timer [0-9\\.]* secs elapsed")
})

#' Concatenate R objects into a character vector via `paste`
#' @param ... passed to `paste`
#' @inheritParams log_level
#' @return character vector
#' @export
#' @family log_formatters
formatter_paste <- function(...,
                            .logcall = sys.call(),
                            .topcall = sys.call(-1),
                            .topenv = parent.frame()) {
  paste(...)
}
attr(formatter_paste, "generator") <- quote(formatter_paste())

#' Apply [sprintf()] to convert R objects into a character vector
#' @param fmt passed to [sprintf()]
#' @param ... passed to [sprintf()]
#' @inheritParams log_level
#' @return character vector
#' @export
#' @family log_formatters
formatter_sprintf <- function(fmt,
                              ...,
                              .logcall = sys.call(),
                              .topcall = sys.call(-1),
                              .topenv = parent.frame()) {
  sprintf(fmt, ...)
}
attr(formatter_sprintf, "generator") <- quote(formatter_sprintf())

#' Apply [glue::glue()] to convert R objects into a character vector
#' @param ... passed to [glue::glue()] for the text interpolation
#' @inheritParams log_level
#' @return character vector
#' @export
#' @note Although this is the default log message formatter function,
#'     but when \pkg{glue} is not installed, [formatter_sprintf()]
#'     will be used as a fallback.
#' @family log_formatters
#' @importFrom utils str
formatter_glue <- function(...,
                           .logcall = sys.call(),
                           .topcall = sys.call(-1),
                           .topenv = parent.frame()) {
  fail_on_missing_package("glue")

  withCallingHandlers(
    glue::glue(..., .envir = .topenv),
    error = function(e) {
      args <- paste0(capture.output(str(...)), collapse = "\n")

      stop(paste0(
        "`glue` failed in `formatter_glue` on:\n\n",
        args,
        "\n\nRaw error message:\n\n",
        conditionMessage(e),
        "\n\nPlease consider using another `log_formatter` or ",
        "`skip_formatter` on strings with curly braces."
      ))
    }
  )
}
attr(formatter_glue, "generator") <- quote(formatter_glue())


#' Apply [glue::glue_safe()] to convert R objects into a character vector
#' @param ... passed to [glue::glue_safe()] for the text interpolation
#' @inheritParams log_level
#' @return character vector
#' @export
#' @family log_formatters
#' @importFrom utils str
formatter_glue_safe <- function(...,
                                .logcall = sys.call(),
                                .topcall = sys.call(-1),
                                .topenv = parent.frame()) {
  fail_on_missing_package("glue")
  as.character(
    tryCatch(
      glue::glue_safe(..., .envir = .topenv),
      error = function(e) {
        stop(paste(
          "`glue_safe` failed in `formatter_glue_safe` on:\n\n",
          capture.output(str(...)),
          "\n\nRaw error message:\n\n",
          e$message,
          "\n\nPlease consider using another `log_formatter` or",
          "`skip_formatter` on strings with curly braces."
        ))
      }
    )
  )
}
attr(formatter_glue_safe, "generator") <- quote(formatter_glue_safe())


#' Apply [glue::glue()] and [sprintf()]
#'
#' The best of both words: using both formatter functions in your log
#' messages, which can be useful eg if you are migrating from
#' [sprintf()] formatted log messages to [glue::glue()] or similar.
#'
#' Note that this function tries to be smart when passing arguments to
#' [glue::glue()] and [sprintf()], but might fail with some edge cases, and
#' returns an unformatted string.
#' @param msg passed to [sprintf()] as `fmt` or handled as part of `...`
#'     in [glue::glue()]
#' @param ... passed to [glue::glue()] for the text interpolation
#' @inheritParams log_level
#' @return character vector
#' @family log_formatters
#' @export
#' @examples
#' formatter_glue_or_sprintf("{a} + {b} = %s", a = 2, b = 3, 5)
#' formatter_glue_or_sprintf("{pi} * {2} = %s", pi * 2)
#' formatter_glue_or_sprintf("{pi} * {2} = {pi*2}")
#'
#' formatter_glue_or_sprintf("Hi ", "{c('foo', 'bar')}, did you know that 2*4={2*4}")
#' formatter_glue_or_sprintf("Hi {c('foo', 'bar')}, did you know that 2*4={2*4}")
#' formatter_glue_or_sprintf("Hi {c('foo', 'bar')}, did you know that 2*4=%s", 2 * 4)
#' formatter_glue_or_sprintf("Hi %s, did you know that 2*4={2*4}", c("foo", "bar"))
#' formatter_glue_or_sprintf("Hi %s, did you know that 2*4=%s", c("foo", "bar"), 2 * 4)
formatter_glue_or_sprintf <- function(msg,
                                      ...,
                                      .logcall = sys.call(),
                                      .topcall = sys.call(-1),
                                      .topenv = parent.frame()) {
  params <- list(...)

  ## params without a name are potential sprintf params
  sprintfparams <- which(names(params) == "")
  if (length(params) > 0 && length(sprintfparams) == 0) {
    sprintfparams <- seq_along(params)
  }
  if (is.null(msg) || length(msg) == 0) {
    msg <- ""
  }

  ## early return
  if (is.null(msg) || length(msg) == 0) {
    return("")
  }

  ## but some unnamed params might belong to glue actually, so
  ## let's look for the max number of first unnamed params sprintf expects
  sprintftags <- regmatches(msg, gregexpr("%[0-9.+0]*[aAdifeEgGosxX]", msg))[[1]]
  sprintfparams <- sprintfparams[seq_len(min(length(sprintftags), length(sprintfparams)))]

  ## get the actual params instead of indexes
  glueparams <- params[setdiff(seq_along(params), sprintfparams)]
  sprintfparams <- params[sprintfparams]

  ## first try to apply sprintf
  if (length(sprintfparams) > 0) {
    sprintfparams[vapply(sprintfparams, is.null, logical(1))] <- "NULL"
    msg <- tryCatch(
      do.call(sprintf, c(msg, sprintfparams), envir = .topenv),
      error = function(e) msg
    )
  }

  ## then try to glue
  fail_on_missing_package("glue")
  msg <- tryCatch(
    as.character(sapply(msg, function(msg) {
      do.call(glue::glue, c(msg, glueparams), envir = .topenv)
    }, USE.NAMES = FALSE)),
    error = function(e) msg
  )

  ## return
  msg
}
attr(formatter_glue_or_sprintf, "generator") <- quote(formatter_glue_or_sprintf())

#' Apply [cli::cli_text()] to format string with cli syntax
#' @param ... passed to [cli::cli_text()] for the text interpolation
#' @inheritParams log_level
#' @return character vector
#' @export
#' @family log_formatters
#' @importFrom utils str
formatter_cli <- function(...,
                          .logcall = sys.call(),
                          .topcall = sys.call(-1),
                          .topenv = parent.frame()) {
  fail_on_missing_package("cli")

  withCallingHandlers(
    cli::cli_fmt(cli::cli_text(..., .envir = .topenv)),
    error = function(e) {
      args <- paste0(capture.output(str(...)), collapse = "\n")

      stop(paste0(
        "`cli` failed in `formatter_cli` on:\n\n",
        args,
        "\n\nRaw error message:\n\n",
        conditionMessage(e),
        "\n\nPlease consider using another `log_formatter` or ",
        "`skip_formatter` on strings with curly braces."
      ))
    }
  )
}
attr(formatter_cli, "generator") <- quote(formatter_cli())

#' Transforms all passed R objects into a JSON list
#' @param ... passed to `toJSON` wrapped into a `list`
#' @inheritParams log_level
#' @return character vector
#' @export
#' @note This functionality depends on the \pkg{jsonlite} package.
#' @family log_formatters
#' @examples
#' \dontshow{old <- logger:::namespaces_set()}
#' log_formatter(formatter_json)
#' log_layout(layout_json_parser())
#' log_info(everything = 42)
#' log_info(mtcars = mtcars, species = iris$Species)
#' \dontshow{logger:::namespaces_set(old)}
formatter_json <- function(...,
                           .logcall = sys.call(),
                           .topcall = sys.call(-1),
                           .topenv = parent.frame()) {
  fail_on_missing_package("jsonlite")
  eval(as.character(jsonlite::toJSON(list(...), auto_unbox = TRUE)), envir = .topenv)
}
attr(formatter_json, "generator") <- quote(formatter_json())

#' Skip the formatter function
#'
#' Adds the `skip_formatter` attribute to an object so that logger
#' will skip calling the formatter function(s). This is useful if you
#' want to preprocess the log message with a custom function instead
#' of the active formatter function(s). Note that the `message` should
#' be a string, and `skip_formatter` should be the only input for the
#' logging function to make this work.
#' @param message character vector directly passed to the appender
#'     function in [logger()]
#' @param ... should be never set
#' @return character vector with `skip_formatter` attribute set to
#'     `TRUE`
#' @export
skip_formatter <- function(message, ...) {
  if (!inherits(message, "character")) {
    stop("Cannot skip the formatter function if the log message is not already formatter to a character vector")
  }
  if (length(list(...)) > 0) {
    stop("Cannot skip the formatter function if further arguments are passed besides the actual log message(s)")
  }
  structure(message, skip_formatter = TRUE)
}


#' Mimic the default formatter used in the \pkg{logging} package
#'
#' The \pkg{logging} package uses a formatter that behaves differently
#' when the input is a string or other R object. If the first argument
#' is a string, then [sprintf()] is being called -- otherwise it does
#' something like [log_eval()] and logs the R expression(s) and the
#' result(s) as well.
#' @param ... string and further params passed to [sprintf()] or R
#'     expressions to be evaluated
#' @inheritParams log_level
#' @return character vector
#' @export
#' @family log_formatters
#' @examples
#' \dontshow{old <- logger:::namespaces_set()}
#' log_formatter(formatter_logging)
#' log_info("42")
#' log_info(42)
#' log_info(4 + 2)
#' log_info("foo %s", "bar")
#' log_info("vector %s", 1:3)
#' log_info(12, 1 + 1, 2 * 2)
#' \dontshow{logger:::namespaces_set(old)}
formatter_logging <- function(...,
                              .logcall = sys.call(),
                              .topcall = sys.call(-1),
                              .topenv = parent.frame()) {
  params <- list(...)
  .logcall <- substitute(.logcall)

  if (is.character(params[[1]])) {
    return(do.call(sprintf, params, envir = .topenv))
  }

  sapply(seq_along(params), function(i) {
    paste(deparse(as.list(.logcall)[-1][[i]]), params[[i]], sep = ": ")
  })
}
attr(formatter_logging, "generator") <- quote(formatter_logging())


#' Formats R objects with pander
#' @param x object to be logged
#' @param ... optional parameters passed to `pander`
#' @inheritParams log_level
#' @return character vector
#' @note This functionality depends on the \pkg{pander} package.
#' @export
#' @family log_formatters
#' @examples
#' \dontshow{old <- logger:::namespaces_set()}
#' log_formatter(formatter_pander)
#' log_info("42")
#' log_info(42)
#' log_info(4 + 2)
#' log_info(head(iris))
#' log_info(head(iris), style = "simple")
#' log_info(lm(hp ~ wt, mtcars))
#' \dontshow{logger:::namespaces_set(old)}
formatter_pander <- function(x,
                             ...,
                             .logcall = sys.call(),
                             .topcall = sys.call(-1),
                             .topenv = parent.frame()) {
  fail_on_missing_package("pander")
  pander::pander_return(x, ...)
}
attr(formatter_pander, "generator") <- quote(formatter_pander())

test_that("glue works", {
  local_test_logger(formatter = formatter_glue)
  a <- 43

  expect_equal(formatter_glue("Hi"), "Hi")
  expect_equal(formatter_glue("   Hi"), "   Hi")
  expect_equal(formatter_glue("1 + {1}"), "1 + 1")
  expect_equal(formatter_glue("{1:2}"), as.character(1:2))
  expect_equal(formatter_glue("pi is {round(pi, 2)}"), "pi is 3.14")
  expect_equal(formatter_glue("Hi {42}"), "Hi 42")
  expect_equal(formatter_glue("Hi {a}", a = 42), "Hi 42")
  expect_equal(formatter_glue("Hi {1:2}"), paste("Hi", 1:2))

  expect_output(do.call(logger, namespaces$global[[1]])(INFO, 42), "42")
  expect_output(do.call(logger, namespaces$global[[1]])(INFO, "Hi {a}"), "43")

  expect_equal(formatter_glue("Hi {a}"), "Hi 43")
  expect_output(log_info("Hi {a}"), "43")
  expect_output(log_warn("Hi {a}"), "43")
  f <- function() log_info("Hi {a}")
  expect_output(f(), "43")

  local_test_logger(
    formatter = formatter_glue,
    appender = appender_void,
  )
  expect_snapshot(formatter_glue("malformed {"), error = TRUE)
  expect_no_error(formatter_glue("malformed {{"))

  ## nolint start
  ## disabled for https://github.com/atalv/azlogr/issues/35
  ## expect_warning(formatter_glue(NULL))
  ## expect_warning(log_info(NULL))
  ## expect_warning(log_info(a = 42, b = "foobar"))
  ## nolint end
})

test_that("glue gives informative error if message contains curlies", {
  local_test_logger(formatter = formatter_glue)
  expect_snapshot(log_info("hi{"), error = TRUE)
})

test_that("glue_safe works", {
  local_test_logger(formatter = formatter_glue_safe)

  expect_equal(formatter_glue_safe("Hi"), "Hi")
  expect_equal(formatter_glue_safe("   Hi"), "   Hi")
  expect_equal(formatter_glue_safe("Hi {a}", a = 42), "Hi 42")

  a <- 43
  expect_equal(formatter_glue_safe("Hi {a}"), "Hi 43")
  expect_output(log_info("Hi {a}"), "43")
  expect_output(log_warn("Hi {a}"), "43")
  f <- function() log_info("Hi {a}")
  expect_output(f(), "43")

  expect_snapshot(error = TRUE, {
    formatter_glue_safe("Hi {42}")
    formatter_glue_safe("malformed {")
  })
  expect_no_error(formatter_glue_safe("malformed {{"))
})

test_that("sprintf works", {
  local_test_logger(formatter = formatter_sprintf)

  expect_equal(formatter_sprintf("Hi"), "Hi")
  expect_equal(formatter_sprintf("Hi %s", 42), "Hi 42")
  expect_equal(formatter_sprintf("Hi %s", 1:2), paste("Hi", 1:2))
  expect_equal(formatter_sprintf("1 + %s", 1), "1 + 1")
  expect_equal(formatter_sprintf("=>%2i", 2), "=> 2")
  expect_equal(formatter_sprintf("%s", 1:2), as.character(1:2))
  expect_equal(formatter_sprintf("pi is %s", round(pi, 2)), "pi is 3.14")
  expect_equal(formatter_sprintf("pi is %1.2f", pi), "pi is 3.14")

  expect_snapshot(formatter_sprintf("%s and %i", 1), error = TRUE)
  expect_equal(formatter_sprintf("%s and %i", 1, 2), "1 and 2")

  a <- 43
  expect_output(log_info("Hi %s", a), "43")
  expect_equal(formatter_sprintf("Hi %s", a), "Hi 43")
  f <- function() log_info("Hi %s", a)
  expect_output(f(), "43")
})


test_that("glue+sprintf works", {
  result <- c(
    "Hi foo, did you know that 2*4=8?",
    "Hi bar, did you know that 2*4=8?"
  )

  expect_equal(formatter_glue_or_sprintf("Hi ", "{c('foo', 'bar')}, did you know that 2*4={2*4}?"), result)
  expect_equal(formatter_glue_or_sprintf("Hi {c('foo', 'bar')}, did you know that 2*4={2*4}?"), result)
  expect_equal(formatter_glue_or_sprintf("Hi {c('foo', 'bar')}, did you know that 2*4=%s?", 2 * 4), result)
  expect_equal(formatter_glue_or_sprintf("Hi %s, did you know that 2*4={2*4}?", c("foo", "bar")), result)
  expect_equal(formatter_glue_or_sprintf("Hi %s, did you know that 2*4=%s?", c("foo", "bar"), 2 * 4), result)

  expect_equal(formatter_glue_or_sprintf("%s and %i"), "%s and %i")
  expect_equal(formatter_glue_or_sprintf("%s and %i", 1), "%s and %i")
  expect_equal(formatter_glue_or_sprintf("fun{fun}"), "fun{fun}")

  for (fn in c(formatter_sprintf, formatter_glue_or_sprintf)) {
    local_test_logger(formatter = fn, appender = appender_void)
    expect_no_error(log_info(character(0)))

    local_test_logger(formatter = fn)
    expect_output(log_info(character(0)), "INFO")
  }
})

test_that("cli works", {
  local_test_logger(formatter = formatter_cli)

  a <- 43

  expect_equal(formatter_cli("Hi"), "Hi")
  expect_equal(formatter_cli("{.arg Hi}"), "`Hi`")
  expect_equal(formatter_cli("1 + {1}"), "1 + 1")
  expect_equal(formatter_cli("{1:2}"), "1 and 2")
  expect_equal(formatter_cli("pi is {round(pi, 2)}"), "pi is 3.14")
  expect_equal(formatter_cli("Hi {42}"), "Hi 42")
  expect_equal(formatter_cli("Hi {1:2}"), paste("Hi 1 and 2"))

  expect_output(do.call(logger, namespaces$global[[1]])(INFO, 42), "42")
  expect_output(do.call(logger, namespaces$global[[1]])(INFO, "Hi {a}"), "43")

  expect_equal(formatter_cli("Hi {a}"), "Hi 43")
  expect_output(log_info("Hi {a}"), "43")
  expect_output(log_warn("Hi {a}"), "43")
  f <- function() log_info("Hi {a}")
  expect_output(f(), "43")
})

test_that("formatter_logging works", {
  local_test_logger(formatter = formatter_logging)

  expect_output(log_info("42"), "42")
  expect_output(log_info(42), "42")
  expect_output(log_info(4 + 2), "4 \\+ 2")
  expect_output(log_info(4 + 2), "6")
  expect_output(log_info("foo %s", "bar"), "foo bar")
  expect_output(log_info(12, 100 + 100, 2 * 2), "12")
  expect_output(log_info(12, 100 + 100, 2 * 2), "100 \\+ 100")
  expect_output(log_info(12, 100 + 100, 2 * 2), "200")
  expect_output(log_info(12, 100 + 100, 2 * 2), "2 \\* 2")
  expect_output(log_info(12, 100 + 100, 2 * 2), "4")
})

test_that("special chars in the text work", {
  array <- "[1, 2, 3, 4]"
  object <- '{"x": 1, "y": 2}'
  expect_equal(formatter_glue("JSON: {array}"), paste0("JSON: ", array))
  expect_equal(formatter_glue("JSON: {object}"), paste0("JSON: ", object))

  local_test_logger()
  expect_output(log_info("JSON: {array}"), paste0("JSON: ", array), fixed = TRUE)
  expect_output(log_info("JSON: {object}"), paste0("JSON: ", object), fixed = TRUE)
})

test_that("pander formatter", {
  local_test_logger(formatter = formatter_pander)
  # pander partially matches coef to coefficient
  withr::local_options(warnPartialMatchDollar = FALSE)

  expect_output(log_info(42), "_42_")
  expect_output(log_info("42"), "42")
  expect_output(log_info(head(iris)), "Sepal.Length")
  expect_output(log_info(lm(hp ~ wt, mtcars)), "Fitting linear model")
})

test_that("paste formatter in actual logs", {
  local_test_logger(formatter = formatter_paste)
  expect_output(log_info("hi", 5), "hi 5")
})

test_that("skip formatter", {
  local_test_logger(formatter = formatter_glue)
  expect_output(log_info(skip_formatter("hi {pi}")), "hi \\{pi\\}")
  expect_snapshot(log_info(skip_formatter("hi {x}", x = 4)), error = TRUE)
})

test_that("skip formatter", {
  local_test_logger(formatter = formatter_json)
  expect_output(log_info(skip_formatter("hi {pi}")), "hi \\{pi\\}")
  expect_output(log_info(x = 1), '\\{"x":1\\}')
})

#' Color string by the related log level
#'
#' Color log messages according to their severity with either a rainbow
#' or grayscale color scheme. The greyscale theme assumes a dark background on
#' the terminal.
#'
#' @param msg String to color.
#' @param level see [log_levels()]
#' @return A string with ANSI escape codes.
#' @export
#' @examplesIf requireNamespace("crayon")
#' cat(colorize_by_log_level("foobar", FATAL), "\n")
#' cat(colorize_by_log_level("foobar", ERROR), "\n")
#' cat(colorize_by_log_level("foobar", WARN), "\n")
#' cat(colorize_by_log_level("foobar", SUCCESS), "\n")
#' cat(colorize_by_log_level("foobar", INFO), "\n")
#' cat(colorize_by_log_level("foobar", DEBUG), "\n")
#' cat(colorize_by_log_level("foobar", TRACE), "\n")
#'
#' cat(grayscale_by_log_level("foobar", FATAL), "\n")
#' cat(grayscale_by_log_level("foobar", ERROR), "\n")
#' cat(grayscale_by_log_level("foobar", WARN), "\n")
#' cat(grayscale_by_log_level("foobar", SUCCESS), "\n")
#' cat(grayscale_by_log_level("foobar", INFO), "\n")
#' cat(grayscale_by_log_level("foobar", DEBUG), "\n")
#' cat(grayscale_by_log_level("foobar", TRACE), "\n")
colorize_by_log_level <- function(msg, level) {
  fail_on_missing_package("crayon")

  color <- switch(attr(level, "level"),
    "FATAL"   = crayon::combine_styles(crayon::bold, crayon::make_style("red1")),
    "ERROR"   = crayon::make_style("red4"),
    "WARN"    = crayon::make_style("darkorange"),
    "SUCCESS" = crayon::combine_styles(crayon::bold, crayon::make_style("green4")),
    "INFO"    = crayon::reset,
    "DEBUG"   = crayon::make_style("deepskyblue4"),
    "TRACE"   = crayon::make_style("dodgerblue4"),
    stop("Unknown log level")
  )

  paste0(color(msg), crayon::reset(""))
}

#' @export
#' @rdname colorize_by_log_level
grayscale_by_log_level <- function(msg, level) {
  fail_on_missing_package("crayon")

  color <- switch(attr(level, "level"),
    "FATAL"   = crayon::make_style("gray100"),
    "ERROR"   = crayon::make_style("gray90"),
    "WARN"    = crayon::make_style("gray80"),
    "SUCCESS" = crayon::make_style("gray70"),
    "INFO"    = crayon::make_style("gray60"),
    "DEBUG"   = crayon::make_style("gray50"),
    "TRACE"   = crayon::make_style("gray40"),
    stop("Unknown log level")
  )

  paste0(color(msg), crayon::reset(""))
}

#' Generate logging utility
#'
#' A logger consists of a log level `threshold`, a log message
#' `formatter` function, a log record `layout` formatting function and
#' the `appender` function deciding on the destination of the log
#' record. For more details, see the package `README.md`.
#'
#' By default, a general logger definition is created when loading the `logger` package, that uses
#'
#' * [INFO()] (or as per the `LOGGER_LOG_LEVEL` environment variable override) as the log level threshold
#' * [layout_simple()] as the layout function showing the log level, timestamp and log message
#' * [formatter_glue()] (or [formatter_sprintf()] if \pkg{glue} is not installed) as the
#'   default formatter function transforming the R objects to be logged to a character vector
#' * [appender_console()] as the default log record destination
#'
#' @param threshold omit log messages below this [log_levels()]
#' @param formatter function pre-processing the message of the log
#'   record when it's not wrapped in a [skip_formatter()] call
#' @param layout function rendering the layout of the actual log
#'   record
#' @param appender function writing the log record
#' @return A function taking the log `level` to compare with the set
#'   threshold, all the `...` arguments passed to the formatter
#'   function, besides the standard `namespace`, `.logcall`,
#'   `.topcall` and `.topenv` arguments (see [log_level()] for more
#'   details). The function invisibly returns a list including the
#'   original `level`, `namespace`, all `...` transformed to a list as
#'   `params`, the log `message` (after calling the `formatter`
#'   function) and the log `record` (after calling the `layout`
#'   function), and a list of `handlers` with the `formatter`,
#'   `layout` and `appender` functions.
#' @export
#' @references For more details, see the Anatomy of a Log Request
#'   vignette at
#'   <https://daroczig.github.io/logger/articles/anatomy.html>.
#' @note It's quite unlikely that you need to call this function
#'   directly, but instead set the logger parameters and functions at
#'   [log_threshold()], [log_formatter()], [log_layout()] and
#'   [log_appender()] and then call [log_levels()] and its
#'   derivatives, such as [log_info()] directly.
#' @examples \dontrun{
#' do.call(logger, logger:::namespaces$global[[1]])(INFO, 42)
#' do.call(logger, logger:::namespaces$global[[1]])(INFO, "{pi}")
#' x <- 42
#' do.call(logger, logger:::namespaces$global[[1]])(INFO, "{x}^2 = {x^2}")
#' }
logger <- function(threshold, formatter, layout, appender) {
  force(threshold)
  threshold <- validate_log_level(threshold)
  force(layout)
  force(appender)

  function(level, ..., namespace = NA_character_,
           .logcall = sys.call(), .topcall = sys.call(-1),
           .topenv = parent.frame(), .timestamp = Sys.time()) {
    res <- list(
      level = level,
      namespace = namespace,
      params = list(...),
      handlers = list(
        formatter = formatter,
        layout = layout,
        appender = appender
      ),
      message = NULL,
      record = NULL
    )

    if (level > threshold) {
      return(invisible(res))
    }

    ## workaround to be able to avoid any formatter function, eg when passing in a string
    if (length(res$params) == 1 && isTRUE(attr(res$params[[1]], "skip_formatter", exact = TRUE))) {
      res$message <- res$params[[1]]
    } else {
      res$message <- do.call(formatter, c(res$params, list(
        .logcall = substitute(.logcall),
        .topcall = substitute(.topcall),
        .topenv = .topenv
      )))
    }

    ## .timestamp arg was added in 0.4.1 and external layout fns might not support it yet
    if (".timestamp" %in% names(formals(layout))) {
      res$record <- layout(
        level, res$message,
        namespace = namespace,
        .logcall = substitute(.logcall),
        .topcall = substitute(.topcall),
        .topenv = .topenv,
        .timestamp = .timestamp
      )
    } else {
      res$record <- layout(
        level, res$message,
        namespace = namespace,
        .logcall = substitute(.logcall),
        .topcall = substitute(.topcall),
        .topenv = .topenv
      )
    }

    appender(res$record)
    invisible(res)
  }
}


#' Checks if provided namespace exists and falls back to global if not
#' @param namespace string
#' @return string
#' @noRd
fallback_namespace <- function(namespace) {
  if (!exists(namespace, envir = namespaces, inherits = FALSE)) {
    namespace <- "global"
  }
  namespace
}

#' @param namespace logger namespace
#' @param index index of the logger within the namespace
#' @return If `value` is `NULL`, will return the currently set value.
#'   If `value` is not `NULL`, will return the previously set value.
#' @noRd
log_config_setter <- function(name, value, namespace = "global", index = 1) {
  if (length(namespace) > 1) {
    for (ns in namespace) {
      log_config_setter(name, value, ns, index)
    }
    return(invisible())
  }

  configs <- get(fallback_namespace(namespace), envir = namespaces)
  config <- configs[[min(index, length(configs))]]
  old <- config[[name]]

  if (name == "threshold") {
    if (is.null(value)) {
      return(config[[name]])
    }
    config[[name]] <- validate_log_level(value)
  } else {
    if (is.null(value)) {
      res <- config[[name]]
      if (!is.null(attr(res, "generator"))) {
        res <- parse(text = attr(res, "generator"))[[1]]
      }
      return(res)
    }

    config[[name]] <- value
  }

  configs[[min(index, length(config) + 1)]] <- config
  assign(namespace, configs, envir = namespaces)
  invisible(old)
}


#' Delete an index from a logger namespace
#' @inheritParams log_threshold
#' @export
delete_logger_index <- function(namespace = "global", index) {
  configs <- get(fallback_namespace(namespace), envir = namespaces)
  if (index > length(configs)) {
    stop(sprintf("%s namespace has only %i indexes", namespace, length(configs)))
  }
  configs[index] <- NULL
  assign(namespace, configs, envir = namespaces)
}


#' Get or set log level threshold
#' @param level see [log_levels()]
#' @param namespace logger namespace
#' @param index index of the logger within the namespace
#' @return currently set log level threshold
#' @export
#' @examples
#' \dontshow{old <- logger:::namespaces_set()}
#' ## check the currently set log level threshold
#' log_threshold()
#'
#' ## change the log level threshold to WARN
#' log_threshold(WARN)
#' log_info(1)
#' log_warn(2)
#'
#' ## add another logger with a lower log level threshold and check the number of logged messages
#' log_threshold(INFO, index = 2)
#' log_info(1)
#' log_warn(2)
#'
#' ## set the log level threshold in all namespaces to ERROR
#' log_threshold(ERROR, namespace = log_namespaces())
#' \dontshow{logger:::namespaces_set(old)}
#' @family log configutation functions
log_threshold <- function(level = NULL, namespace = "global", index = 1) {
  log_config_setter("threshold", level, namespace = namespace, index = index)
}


#' Get or set log record layout
#' @param layout function defining the structure of a log record, eg
#'   [layout_simple()], [layout_glue()] or [layout_glue_colors()],
#'   [layout_json()], or generator functions such as
#'   [layout_glue_generator()], default NULL
#' @inheritParams log_threshold
#' @export
#' @examples
#' \dontshow{old <- logger:::namespaces_set()}
#' log_layout(layout_json())
#' log_info(42)
#' \dontshow{logger:::namespaces_set(old)}
#' @family log configutation functions
log_layout <- function(layout = NULL, namespace = "global", index = 1) {
  if (!is.null(layout) && !is.function(layout)) {
    stop("`layout` must be a function")
  }
  log_config_setter("layout", layout, namespace = namespace, index = index)
}


#' Get or set log message formatter
#' @param formatter function defining how R objects are converted into
#'   a single string, eg [formatter_paste()], [formatter_sprintf()],
#'   [formatter_glue()], [formatter_glue_or_sprintf()],
#'   [formatter_logging()], default NULL
#' @inheritParams log_threshold
#' @export
#' @family log configutation functions
log_formatter <- function(formatter = NULL, namespace = "global", index = 1) {
  if (!is.null(formatter) && !is.function(formatter)) {
    stop("`formatter` must be a function")
  }
  log_config_setter("formatter", formatter, namespace = namespace, index = index)
}


#' Get or set log record appender function
#' @param appender function delivering a log record to the
#'   destination, eg [appender_console()], [appender_file()] or
#'   [appender_tee()], default NULL
#' @inheritParams log_threshold
#' @export
#' @examples
#' \dontshow{old <- logger:::namespaces_set()}
#' ## change appender to "tee" that writes to the console and a file as well
#' t <- tempfile()
#' log_appender(appender_tee(t))
#' log_info(42)
#' log_info(43)
#' log_info(44)
#' readLines(t)
#'
#' ## poor man's tee by stacking loggers in the namespace
#' t <- tempfile()
#' log_appender(appender_stdout)
#' log_appender(appender_file(t), index = 2)
#' log_info(42)
#' readLines(t)
#' \dontshow{logger:::namespaces_set(old)}
#' @family log configutation functions
log_appender <- function(appender = NULL, namespace = "global", index = 1) {
  if (!is.null(appender) && !is.function(appender)) {
    stop("`appender` must be a function")
  }
  log_config_setter("appender", appender, namespace = namespace, index = index)
}


#' Find the logger definition(s) specified for the current namespace
#' with a fallback to the global namespace
#' @return list of function(s)
#' @noRd
#' @importFrom utils getFromNamespace
#' @param namespace override the default / auto-picked namespace with
#'   a custom string
get_logger_definitions <- function(namespace = NA_character_, .topenv = parent.frame()) {
  namespace <- ifelse(is.na(namespace), top_env_name(.topenv), namespace)
  if (!exists(namespace, envir = namespaces, inherits = FALSE)) {
    namespace <- "global"
  }
  get(namespace, envir = getFromNamespace("namespaces", "logger"))
}


#' Looks up logger namespaces
#' @return character vector of namespace names
#' @export
log_namespaces <- function() {
  ls(envir = namespaces)
}


#' Returns number of currently active indices
#' @param namespace override the default / auto-picked namespace with
#'   a custom string
#' @return number of indices
#' @export
log_indices <- function(namespace = "global") {
  length(get(fallback_namespace(namespace), envir = namespaces))
}


#' Log a message with given log level
#' @param level log level, see [log_levels()] for more details
#' @param ... R objects that can be converted to a character vector
#'   via the active message formatter function
#' @param namespace string referring to the `logger` environment /
#'   config to be used to override the target of the message record to
#'   be used instead of the default namespace, which is defined by the
#'   R package name from which the logger was called, and falls back
#'   to a common, global namespace.
#' @param .logcall the logging call being evaluated (useful in
#'   formatters and layouts when you want to have access to the raw,
#'   unevaluated R expression)
#' @param .topcall R expression from which the logging function was
#'   called (useful in formatters and layouts to extract the calling
#'   function's name or arguments)
#' @param .topenv original frame of the `.topcall` calling function
#'   where the formatter function will be evaluated and that is used
#'   to look up the `namespace` as well via `logger:::top_env_name`
#' @param .timestamp The time the logging occured. Defaults to the current time
#'   but may be overwritten if the logging is delayed from the time it happend
#' @export
#' @examples
#' \dontshow{old <- logger:::namespaces_set()}
#' log_level(INFO, "hi there")
#' log_info("hi there")
#'
#' ## output omitted
#' log_debug("hi there")
#'
#' ## lower threshold and retry
#' log_threshold(TRACE)
#' log_debug("hi there")
#'
#' ## multiple lines
#' log_info("ok {1:3} + {1:3} = {2*(1:3)}")
#'
#' ## use json layout
#' log_layout(layout_json(c("time", "level")))
#' log_info("ok {1:3} + {1:3} = {2*(1:3)}")
#' \dontshow{logger:::namespaces_set(old)}
#' @return Invisible `list` of `logger` objects. See [logger()] for more details on the format.
log_level <- function(level, ..., namespace = NA_character_,
                      .logcall = sys.call(), .topcall = sys.call(-1),
                      .topenv = parent.frame(), .timestamp = Sys.time()) {
  ## guess namespace
  if (is.na(namespace)) {
    topenv <- top_env_name(.topenv)
    namespace <- ifelse(topenv == "R_GlobalEnv", "global", topenv)
  }

  definitions <- get_logger_definitions(namespace, .topenv = .topenv)
  level <- validate_log_level(level)

  ## super early return (even before evaluating passed parameters)
  if (length(definitions) == 1 && level > definitions[[1]]$threshold) {
    return(invisible(NULL))
  }

  log_arg <- list(...)
  log_arg$level <- level
  log_arg$.logcall <- .logcall
  log_arg$.topcall <- if (!is.null(.topcall)) {
    .topcall
  } else {
    ## cannot pass NULL
    NA
  }
  log_arg$.topenv <- .topenv
  log_arg$.timestamp <- .timestamp
  log_arg$namespace <- namespace

  invisible(lapply(definitions, function(definition) {
    if (level > definition$threshold) {
      return(NULL)
    }

    log_fun <- do.call(logger, definition)
    structure(do.call(log_fun, log_arg), class = "logger")
  }))
}


#' Assure valid log level
#' @param level [log_levels()] object or string representation
#' @return [log_levels()] object
#' @noRd
validate_log_level <- function(level) {
  if (inherits(level, "loglevel")) {
    return(level)
  }
  if (is.character(level) && level %in% log_levels_supported) {
    return(get(level))
  }
  stop("Invalid log level")
}


#' @export
#' @rdname log_level
log_fatal <- function(..., namespace = NA_character_,
                      .logcall = sys.call(), .topcall = sys.call(-1),
                      .topenv = parent.frame(), .timestamp = Sys.time()) {
  log_level(FATAL, ..., namespace = namespace, .logcall = .logcall,
            .topcall = .topcall, .topenv = .topenv, .timestamp = .timestamp)
}
#' @export
#' @rdname log_level
log_error <- function(..., namespace = NA_character_,
                      .logcall = sys.call(), .topcall = sys.call(-1),
                      .topenv = parent.frame(), .timestamp = Sys.time()) {
  log_level(ERROR, ..., namespace = namespace, .logcall = .logcall,
            .topcall = .topcall, .topenv = .topenv, .timestamp = .timestamp)
}
#' @export
#' @rdname log_level
log_warn <- function(..., namespace = NA_character_,
                     .logcall = sys.call(), .topcall = sys.call(-1),
                     .topenv = parent.frame(), .timestamp = Sys.time()) {
  log_level(WARN, ..., namespace = namespace, .logcall = .logcall,
            .topcall = .topcall, .topenv = .topenv, .timestamp = .timestamp)
}
#' @export
#' @rdname log_level
log_success <- function(..., namespace = NA_character_,
                        .logcall = sys.call(), .topcall = sys.call(-1),
                        .topenv = parent.frame(), .timestamp = Sys.time()) {
  log_level(SUCCESS, ..., namespace = namespace, .logcall = .logcall,
            .topcall = .topcall, .topenv = .topenv, .timestamp = .timestamp)
}
#' @export
#' @rdname log_level
log_info <- function(..., namespace = NA_character_,
                     .logcall = sys.call(), .topcall = sys.call(-1),
                     .topenv = parent.frame(), .timestamp = Sys.time()) {
  log_level(INFO, ..., namespace = namespace, .logcall = .logcall,
            .topcall = .topcall, .topenv = .topenv, .timestamp = .timestamp)
}
#' @export
#' @rdname log_level
log_debug <- function(..., namespace = NA_character_,
                      .logcall = sys.call(), .topcall = sys.call(-1),
                      .topenv = parent.frame(), .timestamp = Sys.time()) {
  log_level(DEBUG, ..., namespace = namespace, .logcall = .logcall,
            .topcall = .topcall, .topenv = .topenv, .timestamp = .timestamp)
}
#' @export
#' @rdname log_level
log_trace <- function(..., namespace = NA_character_,
                      .logcall = sys.call(), .topcall = sys.call(-1),
                      .topenv = parent.frame(), .timestamp = Sys.time()) {
  log_level(TRACE, ..., namespace = namespace, .logcall = .logcall,
            .topcall = .topcall, .topenv = .topenv, .timestamp = .timestamp)
}


#' Evaluate R expression with a temporarily updated log level threshold
#' @param expression R command
#' @param threshold [log_levels()]
#' @inheritParams log_threshold
#' @export
#' @examples
#' \dontshow{old <- logger:::namespaces_set()}
#' log_threshold(TRACE)
#' log_trace("Logging everything!")
#' x <- with_log_threshold(
#'   {
#'     log_info("Now we are temporarily suppressing eg INFO messages")
#'     log_warn("WARN")
#'     log_debug("Debug messages are suppressed as well")
#'     log_error("ERROR")
#'     invisible(42)
#'   },
#'   threshold = WARN
#' )
#' x
#' log_trace("DONE")
#' \dontshow{logger:::namespaces_set(old)}
with_log_threshold <- function(expression, threshold = ERROR, namespace = "global", index = 1) {
  old <- log_threshold(threshold, namespace = namespace, index = index)
  on.exit(log_threshold(old, namespace = namespace, index = index))
  expression
}

library(logger)
library(testthat)
library(jsonlite)

## save current settings so that we can reset later
layout <- log_layout()
appender <- log_appender()

log_appender(appender_stdout)

context('layouts')

log_layout(layout_blank)
test_that('blank layout', {
    expect_output(log_info('foobar'), 'foobar')
    expect_equal(capture.output(log_info('foobar')), 'foobar')
})

log_layout(layout_glue_colors)
test_that('colorized layout', {
    expect_output(log_info('foobar'), 'INFO')
    expect_output(log_info('foobar'), 'foobar')
    expect_output(log_error('foobar'), 'ERROR')
    expect_output(log_error('foobar'), 'foobar')
})

log_layout(layout_json())
test_that('JSON layout', {
    expect_equal(fromJSON(capture.output(log_info('foobar')))$level, 'INFO')
    expect_equal(fromJSON(capture.output(log_info('foobar')))$msg, 'foobar')
})

log_layout(layout_json_parser(fields = c()))
test_that('JSON parser layout', {
    expect_output(log_info(skip_formatter('{"x": 4}')), '\\{"x":4\\}')
    expect_equal(capture.output(log_info(skip_formatter('{"x": 4}'))), '{"x":4}')
})

test_that('must throw errors', {

    expect_error(layout_simple(FOOBAR))
    expect_error(layout_simple(42))
    expect_error(layout_simple(msg = 'foobar'))

    expect_error(layout_glue(FOOBAR))
    expect_error(layout_glue(42))
    expect_error(layout_glue(msg = 'foobar'))
    expect_error(layout_glue(level = 53, msg = 'foobar'))

})

log_layout(layout_logging)
test_that('logging layout', {
    expect_output(log_info('foobar'), 'INFO')
    expect_output(log_info('foo', namespace = 'bar'), 'foo')
    expect_output(log_info('foo', namespace = 'bar'), 'bar')
    expect_output(log_info('foo', namespace = 'bar'), 'INFO:bar:foo')
})

## reset settings
log_layout(layout)
log_appender(appender)

test_that("log levels", {
  local_test_logger(WARN)

  expect_output(log_fatal("foo"), "FATAL.*foo")
  expect_output(log_error("foo"), "ERROR.*foo")
  expect_output(log_warn("foo"), "WARN.*foo")
  expect_output(log_success("foo"), NA)
  expect_output(log_info("foo"), NA)
  expect_output(log_debug("foo"), NA)
  expect_output(log_trace("foo"), NA)
  expect_output(log_level("ERROR", "foo"), "ERROR.*foo")
  expect_output(log_level(ERROR, "foo"), "ERROR.*foo")
  expect_output(log_level(as.loglevel(ERROR), "foo"), "ERROR.*foo")
  expect_output(log_level(as.loglevel("ERROR"), "foo"), "ERROR.*foo")
  expect_output(log_level(as.loglevel(200L), "foo"), "ERROR.*foo")
  expect_output(log_level("TRACE", "foo"), NA)
  expect_output(log_level(TRACE, "foo"), NA)
  expect_output(log_level(as.loglevel(TRACE), "foo"), NA)
  expect_output(log_level(as.loglevel("TRACE"), "foo"), NA)
  expect_output(log_level(as.loglevel(600L), "foo"), NA)
})

test_that("log levels - OFF", {
  local_test_logger(OFF)
  expect_output(log_fatal("foo"), NA)
  expect_output(log_error("foo"), NA)
  expect_output(log_warn("foo"), NA)
  expect_output(log_success("foo"), NA)
  expect_output(log_info("foo"), NA)
  expect_output(log_debug("foo"), NA)
  expect_output(log_trace("foo"), NA)
})

test_that("log thresholds", {
  local_test_logger(TRACE)
  expect_output(log_fatal("foo"), "FATAL.*foo")
  expect_output(log_error("foo"), "ERROR.*foo")
  expect_output(log_warn("foo"), "WARN.*foo")
  expect_output(log_success("foo"), "SUCCESS.*foo")
  expect_output(log_info("foo"), "INFO.*foo")
  expect_output(log_debug("foo"), "DEBUG.*foo")
  expect_output(log_trace("foo"), "TRACE.*foo")
})

test_that("with log thresholds", {
  local_test_logger(WARN)
  expect_output(with_log_threshold(log_fatal("foo"), threshold = TRACE), "FATAL.*foo")
  expect_output(with_log_threshold(log_error("foo"), threshold = TRACE), "ERROR.*foo")
  expect_output(with_log_threshold(log_error("foo"), threshold = FATAL), NA)
  expect_output(with_log_threshold(log_error("foo"), threshold = INFO), "ERROR.*foo")
  expect_output(with_log_threshold(log_debug("foo"), threshold = INFO), NA)
})

test_that("simple glue layout with no threshold", {
  local_test_logger(TRACE, layout = layout_glue_generator("{level} {msg}"))

  expect_equal(capture.output(log_fatal("foobar")), "FATAL foobar")
  expect_equal(capture.output(log_error("foobar")), "ERROR foobar")
  expect_equal(capture.output(log_warn("foobar")), "WARN foobar")
  expect_equal(capture.output(log_info("foobar")), "INFO foobar")
  expect_equal(capture.output(log_debug("foobar")), "DEBUG foobar")
  expect_equal(capture.output(log_trace("foobar")), "TRACE foobar")
})

test_that("simple glue layout with threshold", {
  local_test_logger(INFO, layout = layout_glue_generator("{level} {msg}"))
  expect_equal(capture.output(log_fatal("foobar")), "FATAL foobar")
  expect_equal(capture.output(log_error("foobar")), "ERROR foobar")
  expect_equal(capture.output(log_warn("foobar")), "WARN foobar")
  expect_equal(capture.output(log_info("foobar")), "INFO foobar")
  expect_equal(capture.output(log_debug("foobar")), character())
  expect_equal(capture.output(log_trace("foobar")), character())
})

test_that("namespaces", {
  local_test_logger(ERROR, namespace = "custom", layout = layout_glue_generator("{level} {msg}"))
  expect_output(log_fatal("foobar", namespace = "custom"), "FATAL foobar")
  expect_output(log_error("foobar", namespace = "custom"), "ERROR foobar")
  expect_output(log_info("foobar", namespace = "custom"), NA)
  expect_output(log_debug("foobar", namespace = "custom"), NA)

  local_test_logger(INFO, namespace = "custom", layout = layout_glue_generator("{level} {msg}"))
  expect_output(log_info("foobar", namespace = "custom"), "INFO foobar")
  expect_output(log_debug("foobar", namespace = "custom"), NA)

  log_threshold(TRACE, namespace = log_namespaces())
  expect_output(log_debug("foobar", namespace = "custom"), "DEBUG foobar")
})

test_that("simple glue layout with threshold directly calling log", {
  local_test_logger(layout = layout_glue_generator("{level} {msg}"))
  expect_equal(capture.output(log_level(FATAL, "foobar")), "FATAL foobar")
  expect_equal(capture.output(log_level(ERROR, "foobar")), "ERROR foobar")
  expect_equal(capture.output(log_level(WARN, "foobar")), "WARN foobar")
  expect_equal(capture.output(log_level(INFO, "foobar")), "INFO foobar")
  expect_equal(capture.output(log_level(DEBUG, "foobar")), character())
  expect_equal(capture.output(log_level(TRACE, "foobar")), character())
})

test_that("built in variables: pid", {
  local_test_logger(layout = layout_glue_generator("{pid}"))
  expect_equal(capture.output(log_info("foobar")), as.character(Sys.getpid()))
})

test_that("built in variables: fn and call", {
  local_test_logger(layout = layout_glue_generator("{fn} / {call}"))
  f <- function() log_info("foobar")
  expect_output(f(), "f / f()")
  g <- function() f()
  expect_output(g(), "f / f()")
  g <- f
  expect_output(g(), "g / g()")
})

test_that("built in variables: namespace", {
  local_test_logger(layout = layout_glue_generator("{ns}"))
  expect_output(log_info("bar", namespace = "foo"), "foo")

  local_test_logger(layout = layout_glue_generator("{ans}"))
  expect_output(log_info("bar", namespace = "foo"), "global")
})

test_that("print.level", {
  expect_equal(capture.output(print(INFO)), "Log level: INFO")
})

test_that("config setter called from do.call", {
  local_test_logger()

  t <- withr::local_tempfile()
  expect_no_error(do.call(log_appender, list(appender_file(t))))
  log_info(42)
  expect_length(readLines(t), 1)
  expect_no_error(do.call(log_threshold, list(ERROR)))
  log_info(42)
  expect_length(readLines(t), 1)
  expect_no_error(do.call(log_threshold, list(INFO)))
  log_info(42)
  expect_length(readLines(t), 2)
  expect_no_error(do.call(log_layout, list(formatter_paste)))
  log_info(42)
  expect_length(readLines(t), 3)
})

test_that("providing log_level() args to wrappers diretly is OK", {
  local_test_logger(WARN)
  expect_silent(log_info("{Sepal.Length}", .topenv = list2env(iris)))
})

test_that("setters check inputs", {
  expect_snapshot(error = TRUE, {
    log_appender(1)
    log_formatter(1)
    log_layout(1)
    log_threshold("x")
  })
})

glue_or_sprintf_result <- c(
  "Hi foo, did you know that 2*4=8?",
  "Hi bar, did you know that 2*4=8?"
)

test_that("return value is formatted string", {
  local_test_logger(appender = appender_file(withr::local_tempfile()))

  log_formatter(formatter_glue)
  expect_equal(log_info("pi is {round(pi, 2)}")[[1]]$message, "pi is 3.14")
  expect_match(log_info("pi is {round(pi, 2)}")[[1]]$record, "INFO [[0-9: -]*] pi is 3.14")
  log_formatter(formatter_paste, index = 2)
  expect_equal(log_info("pi is {round(pi, 2)}")[[1]]$message, "pi is 3.14")
  expect_equal(log_info("pi is {round(pi, 2)}")[[2]]$message, "pi is {round(pi, 2)}")
})

#' Dummy appender not delivering the log record to anywhere
#' @param lines character vector
#' @export
appender_void <- function(lines) {}
attr(appender_void, "generator") <- quote(appender_void())

#' Append log record to stderr
#' @param lines character vector
#' @export
#' @family log_appenders
appender_console <- function(lines) {
  cat(lines, file = stderr(), sep = "\n")
}
attr(appender_console, "generator") <- quote(appender_console())


#' @export
#' @rdname appender_console
appender_stderr <- appender_console
attr(appender_stderr, "generator") <- quote(appender_stderr())


#' Append log record to stdout
#' @param lines character vector
#' @export
#' @family log_appenders
appender_stdout <- function(lines) {
  cat(lines, sep = "\n")
}
attr(appender_stdout, "generator") <- quote(appender_stdout())

#' Append log messages to a file
#'
#' Log messages are written to a file with basic log rotation: when
#' max number of lines or bytes is defined to be other than `Inf`,
#' then the log file is renamed with a `.1` suffix and a new log file
#' is created. The renaming happens recursively (eg `logfile.1`
#' renamed to `logfile.2`) until the specified `max_files`, then the
#' oldest file (\code{logfile.{max_files-1}}) is deleted.
#' @param file path
#' @param append boolean passed to `cat` defining if the file should
#'     be overwritten with the most recent log message instead of
#'     appending
#' @param max_lines numeric specifying the maximum number of lines
#'     allowed in a file before rotating
#' @param max_bytes numeric specifying the maximum number of bytes
#'     allowed in a file before rotating
#' @param max_files integer specifying the maximum number of files to
#'     be used in rotation
#' @export
#' @return function taking `lines` argument
#' @family log_appenders
#' @examples
#' \dontshow{old <- logger:::namespaces_set()}
#' ## ##########################################################################
#' ## simple example logging to a file
#' t <- tempfile()
#' log_appender(appender_file(t))
#' for (i in 1:25) log_info(i)
#' readLines(t)
#'
#' ## ##########################################################################
#' ## more complex example of logging to file
#' ## rotated after every 3rd line up to max 5 files
#'
#' ## create a folder storing the log files
#' t <- tempfile()
#' dir.create(t)
#' f <- file.path(t, "log")
#'
#' ## define the file logger with log rotation enabled
#' log_appender(appender_file(f, max_lines = 3, max_files = 5L))
#'
#' ## enable internal logging to see what's actually happening in the logrotate steps
#' log_threshold(TRACE, namespace = ".logger")
#' ## log 25 messages
#' for (i in 1:25) log_info(i)
#'
#' ## see what was logged
#' lapply(list.files(t, full.names = TRUE), function(t) {
#'   cat("\n##", t, "\n")
#'   cat(readLines(t), sep = "\n")
#' })
#'
#' \dontshow{logger:::namespaces_set(old)}
appender_file <- function(file, append = TRUE, max_lines = Inf, max_bytes = Inf, max_files = 1L) { # nolint
  force(file)
  force(append)
  force(max_lines)
  force(max_bytes)
  force(max_files)

  if (!is.integer(max_files) || max_files < 1) {
    stop("max_files must be a positive integer")
  }

  structure(
    function(lines) {
      if (is.finite(max_lines) | is.finite(max_bytes)) {
        fail_on_missing_package("R.utils")

        n_lines <- tryCatch(
          suppressWarnings(R.utils::countLines(file)),
          error = function(e) 0
        )
        n_bytes <- ifelse(file.exists(file), file.info(file)$size, 0)

        if (n_lines >= max_lines || n_bytes >= max_bytes) {
          log_trace(
            "lines: %s, max_lines: %s, bytes: %s, max_bytes: %s",
            n_lines, max_lines, n_bytes, max_bytes,
            namespace = ".logger"
          )
          log_trace(
            "lines >= max_lines || bytes >= max_bytes: %s",
            n_lines >= max_lines || n_bytes >= max_bytes,
            namespace = ".logger"
          )
          for (i in max_files:1) {
            ## just kill the old file
            if (i == 1) {
              log_trace("killing the main file: %s", file, namespace = ".logger")
              unlink(file)
            } else {
              ## rotate the old file
              new <- paste(file, i - 1, sep = ".")
              if (i == 2) {
                old <- file
              } else {
                old <- paste(file, i - 2, sep = ".")
              }

              if (file.exists(old)) {
                log_trace("renaming %s to %s", old, new, namespace = ".logger")
                file.rename(old, new)
              }

              ## kill the rotated, but not needed file
              if (i > max_files) {
                log_trace("killing the file with too many rotations: %s", new, namespace = ".logger")
                unlink(new)
              }
            }
          }
        }
      }
      log_trace("logging %s to %s", shQuote(lines), file, namespace = ".logger")
      cat(lines, sep = "\n", file = file, append = append)
    },
    generator = deparse(match.call())
  )
}


#' Append log messages to a file and stdout as well
#'
#' This appends log messages to both console and a file. The same
#' rotation options are available as in [appender_file()].
#' @inheritParams appender_file
#' @export
#' @return function taking `lines` argument
#' @family log_appenders
appender_tee <- function(file, append = TRUE, max_lines = Inf, max_bytes = Inf, max_files = 1L) {
  force(file)
  force(append)
  force(max_lines)
  force(max_bytes)
  force(max_files)
  structure(
    function(lines) {
      if (needs_stdout()) appender_stdout(lines) else appender_console(lines)
      appender_file(file, append, max_lines, max_bytes, max_files)(lines)
    },
    generator = deparse(match.call())
  )
}


#' Send log messages to a Slack channel
#' @param channel Slack channel name with a hashtag prefix for public
#'     channel and no prefix for private channels
#' @param username Slack (bot) username
#' @param icon_emoji optional override for the bot icon
#' @param api_token Slack API token
#' @param preformatted use code tags around the message?
#' @return function taking `lines` argument
#' @export
#' @note This functionality depends on the \pkg{slackr} package.
#' @family log_appenders
appender_slack <- function(channel = Sys.getenv("SLACK_CHANNEL"),
                           username = Sys.getenv("SLACK_USERNAME"),
                           icon_emoji = Sys.getenv("SLACK_ICON_EMOJI"),
                           api_token = Sys.getenv("SLACK_API_TOKEN"),
                           preformatted = TRUE) {
  fail_on_missing_package("slackr", "1.4.1")
  force(channel)
  force(username)
  force(icon_emoji)
  force(api_token)
  force(preformatted)

  structure(
    function(lines) {
      slackr::slackr_msg(
        text = lines, channel = channel, username = username,
        icon_emoji = icon_emoji, token = api_token, preformatted = preformatted
      )
    },
    generator = deparse(match.call())
  )
}


#' Send log messages to Pushbullet
#' @param ... parameters passed to [RPushbullet::pbPost], such as `recipients`
#'     or `apikey`, although it's probably much better to set all these
#'     in the `~/.rpushbullet.json` as per package docs at
#'     <http://dirk.eddelbuettel.com/code/rpushbullet.html>
#' @export
#' @note This functionality depends on the \pkg{RPushbullet} package.
#' @family log_appenders
#' @export
appender_pushbullet <- function(...) {
  fail_on_missing_package("RPushbullet")

  structure(
    function(lines) {
      RPushbullet::pbPost(type = "note", body = paste(lines, sep = "\n"), ...)
    },
    generator = deparse(match.call())
  )
}


#' Send log messages to ntfy
#' @param title notification title
#' @param tags emoji (or general tag) for notification. See [ntfy::emoji]
#' @param ... extra parameters passed to [ntfy::ntfy_send] such as
#'     `priority`, `topic`, etc.
#' @details Configure server and topic via environment variables. See
#' [ntfy::ntfy_topic()] for details
#' @export
#' @note This functionality depends on the \pkg{ntfy} package.
#' @family log_appenders
#' @export
appender_ntfy <- function(title = "{logger}",
                          tags = c("memo"),
                          ...) {
  fail_on_missing_package("ntfy")
  force(title)
  force(tags)

  structure(
    function(lines) {
      ntfy::ntfy_send(title = title, tags = tags, message = paste(lines, sep = "\n"), ...)
    },
    generator = deparse(match.call())
  )
}



#' Send log messages to a Telegram chat
#' @param chat_id Unique identifier for the target chat or username of
#'     the target channel (in the format @channelusername)
#' @param bot_token Telegram Authorization token
#' @param parse_mode Message parse mode. Allowed values: Markdown or
#'     HTML
#' @return function taking `lines` argument
#' @export
#' @note This functionality depends on the \pkg{telegram} package.
#' @family log_appenders
appender_telegram <- function(chat_id = Sys.getenv("TELEGRAM_CHAT_ID"),
                              bot_token = Sys.getenv("TELEGRAM_BOT_TOKEN"),
                              parse_mode = NULL) {
  fail_on_missing_package("telegram")
  force(chat_id)
  force(bot_token)
  force(parse_mode)

  tb <- telegram::TGBot$new(token = bot_token)
  structure(
    function(lines) {
      tb$sendMessage(text = lines, parse_mode = parse_mode, chat_id = chat_id)
    },
    generator = deparse(match.call())
  )
}


#' Send log messages to the POSIX system log
#' @param identifier A string identifying the process.
#' @param ... Further arguments passed on to [rsyslog::open_syslog()].
#' @return function taking `lines` argument
#' @export
#' @note This functionality depends on the \pkg{rsyslog} package.
#' @family log_appenders
#' @examples \dontrun{
#' if (requireNamespace("rsyslog", quietly = TRUE)) {
#'   log_appender(appender_syslog("test"))
#'   log_info("Test message.")
#' }
#' }
appender_syslog <- function(identifier, ...) {
  fail_on_missing_package("rsyslog")
  rsyslog::open_syslog(identifier = identifier, ...)
  structure(
    function(lines) {
      for (line in lines) {
        rsyslog::syslog(line)
      }
    },
    generator = deparse(match.call())
  )
}


# nocov start
#' Send log messages to a network syslog server
#' @param identifier program/function identification (string).
#' @param server machine where syslog daemon runs (string).
#' @param port port where syslog daemon listens (integer).
#'
#' @return A function taking a `lines` argument.
#' @export
#' @note This functionality depends on the \pkg{syslognet} package.
#' @examples \dontrun{
#' if (requireNamespace("syslognet", quietly = TRUE)) {
#'   log_appender(appender_syslognet("test_app", "remoteserver"))
#'   log_info("Test message.")
#' }
#' }
appender_syslognet <- function(identifier, server, port = 601L) {
  fail_on_missing_package("syslognet")
  force(identifier)
  force(server)
  force(port)
  structure(
    function(lines) {
      sev <- attr(lines, "severity", exact = TRUE)
      for (line in lines) {
        syslognet::syslog(line, sev, app_name = identifier, server = server, port = port)
      }
    },
    generator = deparse(match.call())
  )
}
# nocov end


#' Send log messages to a Amazon Kinesis stream
#' @param stream name of the Kinesis stream
#' @return function taking `lines` and optional `partition_key`
#'     argument
#' @export
#' @note This functionality depends on the \pkg{botor} package.
#' @family log_appenders
appender_kinesis <- function(stream) {
  fail_on_missing_package("botor")
  force(stream)
  structure(
    function(lines, partition_key = NA_character_) {
      for (line in lines) {
        botor::kinesis()$put_record(StreamName = stream, Data = line, PartitionKey = partition_key)
      }
    },
    generator = deparse(match.call())
  )
}


#' Delays executing the actual appender function to the future in a
#' background process to avoid blocking the main R session
#' @param appender a [log_appender()] function with a `generator`
#'     attribute (TODO note not required, all fn will be passed if
#'     not)
#' @param namespace `logger` namespace to use for logging messages on
#'     starting up the background process
#' @param init optional function to run in the background process that
#'     is useful to set up the environment required for logging, eg if
#'     the `appender` function requires some extra packages to be
#'     loaded or some environment variables to be set etc
#' @return function taking `lines` argument
#' @export
#' @note This functionality depends on the \pkg{mirai} package.
#' @family log_appenders
#' @examples \dontrun{
#' appender_file_slow <- function(file) {
#'   force(file)
#'   function(lines) {
#'     Sys.sleep(1)
#'     cat(lines, sep = "\n", file = file, append = TRUE)
#'   }
#' }
#'
#' ## log what's happening in the background
#' log_threshold(TRACE, namespace = "async_logger")
#' log_appender(appender_console, namespace = "async_logger")
#'
#' ## start async appender
#' t <- tempfile()
#' log_info("Logging in the background to {t}")
#'
#' ## use async appender
#' log_appender(appender_async(appender_file_slow(file = t)))
#' log_info("Was this slow?")
#' system.time(for (i in 1:25) log_info(i))
#'
#' readLines(t)
#' Sys.sleep(10)
#' readLines(t)
#'
#' }
appender_async <- function(appender,
                           namespace = "async_logger",
                           init = function() log_info("Background process started")) {
  fail_on_missing_package("mirai")
  force(appender)

  # Start one non-dispatcher background process if not already started
  if (is.null(mirai::nextget("n", .compute = namespace))) {
    mirai::daemons(1L, dispatcher = FALSE, cleanup = FALSE, .compute = namespace)
  }
  mirai::everywhere(
    {
      library(logger)
      init()
    },
    appender = appender, # remains in .GlobalEnv on daemon
    .args = list(init = init),
    .compute = namespace
  )

  structure(
    function(lines) {
      mirai::mirai(
        for (line in lines) {
          appender(line)
        },
        .args = list(lines = lines),
        .compute = namespace
      )
    },
    generator = deparse(match.call())
  )
}

## TODO other appenders: graylog, datadog, cloudwatch, email via sendmailR, ES etc

test_that("append to file", {
  t <- withr::local_tempfile()
  local_test_logger(
    appender = appender_file(t),
    layout = layout_glue_generator("{level} {msg}"),
    threshold = TRACE
  )
  log_info("foobar")
  log_info("{1:2}")
  expect_equal(length(readLines(t)), 3)
  expect_equal(readLines(t)[1], "INFO foobar")
  expect_equal(readLines(t)[3], "INFO 2")
})

test_that("overwrite file", {
  t <- withr::local_tempfile()
  local_test_logger(
    appender = appender_file(t, append = FALSE),
    layout = layout_glue_generator("{level} {msg}"),
    threshold = TRACE
  )

  log_info("foobar")
  log_info("{1:2}")
  expect_equal(length(readLines(t)), 2)
  expect_equal(readLines(t), c("INFO 1", "INFO 2"))

  log_info("42")
  expect_equal(length(readLines(t)), 1)
  expect_equal(readLines(t), "INFO 42")
})

test_that("append to file + print to console", {
  t <- withr::local_tempfile()
  local_test_logger(
    appender = appender_tee(t),
    layout = layout_glue_generator("{level} {msg}"),
  )

  expect_output(log_info("foobar"), "INFO foobar")
  capture.output(log_info("{1:2}"))
  expect_equal(length(readLines(t)), 3)
  expect_equal(readLines(t)[1], "INFO foobar")
})

test_that("logrotate", {
  t <- withr::local_tempdir()
  f <- file.path(t, "log")
  local_test_logger(
    appender = appender_file(f, max_lines = 2, max_files = 5L),
    layout = layout_glue_generator("{msg}"),
    threshold = TRACE
  )

  for (i in 1:24) log_info(i)
  expect_equal(length(readLines(f)), 2)
  expect_equal(length(list.files(t)), 5)
  expect_equal(readLines(f), c("23", "24"))
  log_info("42")
  expect_equal(length(readLines(f)), 1)
  expect_equal(readLines(f), "42")
})

test_that("async logging", {
  skip_on_cran()

  t <- withr::local_tempfile()
  local_test_logger(
    layout = layout_blank,
    appender = appender_async(appender_file(file = t))
  )

  for (i in 1:5) log_info(i)
  Sys.sleep(1)

  expect_equal(readLines(t)[1], "1")
  expect_equal(length(readLines(t)), 5)
})

# Do not move this to another line as the location of this piece of code is tested for
test_info <- function() {
  log_info("TEST")
}

local_test_logger <- function(threshold = INFO,
                              formatter = formatter_glue,
                              layout = layout_simple,
                              appender = appender_stdout,
                              namespace = "global",
                              frame = parent.frame()) {
  old <- namespaces[[namespace]]

  namespaces[[namespace]] <- list(
    default = list(
      threshold = as.loglevel(threshold),
      layout    = layout,
      formatter = formatter,
      appender  = appender
    )
  )

  withr::defer(namespaces[[namespace]] <- old, frame)
  invisible()
}

eval_outside <- function(...) {
  input <- normalizePath(withr::local_tempfile(lines = character()), winslash = "/")
  output <- normalizePath(withr::local_tempfile(lines = character()), winslash = "/")
  writeLines(con = input, c(
    "library(logger)",
    "log_layout(layout_glue_generator('{level} {msg}'))",
    paste0("log_appender(appender_file('", output, "'))"),
    ...
  ))
  path <- file.path(R.home("bin"), "Rscript")
  if (Sys.info()[["sysname"]] == "Windows") {
    path <- paste0(path, ".exe")
  }
  suppressWarnings(system2(path, input, stdout = TRUE, stderr = TRUE))
  readLines(output)
}

# This function is needed to test traceback logging
function_that_fails <- function() {
  stop("I'm failing")
}

test_that("captures call/environment info", {
  f <- function(...) logger_meta_env()
  env <- f(x = 1)

  # values are computed lazily
  expect_type(substitute(fn, env), "language")
  expect_type(substitute(call, env), "language")

  # and give correct values
  expect_equal(env$fn, "f")
  expect_equal(env$call, "f(x = 1)")
  expect_equal(env$topenv, "logger")
})

test_that("captures namespace info", {
  env <- logger_meta_env(namespace = "testthat")
  expect_equal(env$ns, "testthat")
  expect_equal(env$ans, "global")
  expect_equal(env$ns_pkg_version, as.character(packageVersion("testthat")))
})

test_that("captures other environmental metadata", {
  env <- logger_meta_env()
  expect_equal(env$pid, Sys.getpid())
  expect_equal(env$r_version, as.character(getRversion()))

  sysinfo <- as.list(Sys.info())
  expect_equal(env$node, sysinfo$nodename)
  expect_equal(env$arch, sysinfo$machine)
  expect_equal(env$os_name, sysinfo$sysname)
  expect_equal(env$os_release, sysinfo$release)
  expect_equal(env$os_version, sysinfo$version)
  expect_equal(env$user, sysinfo$user)

  local_test_logger(layout = layout_glue_generator("{location$path}#{location$line}: {msg}"))
  expect_output(test_info(), file.path(getwd(), "helper.R#3: TEST"))
})

#' Try to evaluate an expressions and evaluate another expression on
#' exception
#' @param try R expression
#' @param except fallback R expression to be evaluated if `try` fails
#' @export
#' @note Suppress log messages in the `except` namespace if you don't
#'   want to throw a `WARN` log message on the exception branch.
#' @examples
#' everything %except% 42
#' everything <- "640kb"
#' everything %except% 42
#'
#' FunDoesNotExist(1:10) %except% sum(1:10) / length(1:10)
#' FunDoesNotExist(1:10) %except% (sum(1:10) / length(1:10))
#' FunDoesNotExist(1:10) %except% MEAN(1:10) %except% mean(1:10)
#' FunDoesNotExist(1:10) %except% (MEAN(1:10) %except% mean(1:10))
`%except%` <- function(try, except) {
  ## Need to capture these in the evaluation frame of `%except%` but only want
  ## to do the work if there's an error
  delayedAssign("call", sys.call(-1))
  delayedAssign("env", parent.frame())
  delayedAssign("except_text", deparse(substitute(except)))
  delayedAssign("try_text", deparse(substitute(try)))

  tryCatch(try,
    error = function(e) {
      log_level(
        WARN,
        paste0("Running '", except_text, "' as '", try_text, "' failed: '", e$message, "'"),
        namespace = "except",
        .topcall = call,
        .topenv = env
      )
      except
    }
  )
}

#' Injects a logger call to standard messages
#'
#' This function sets a hook to trigger [log_info()] when `message` is called
#' to log the informative messages with the global `logger` layout and appender.
#' @export
#' @examples \dontrun{
#' log_messages()
#' message("hi there")
#' }
log_messages <- function() {
  if (any(sapply(
    globalCallingHandlers()[names(globalCallingHandlers()) == "message"],
    attr,
    which = "implements"
  ) == "log_messages")) {
    warning("Ignoring this call to log_messages as it was registered previously.")
  } else {
    globalCallingHandlers(
      message = structure(function(m) {
        logger::log_level(logger::INFO, m$message, .topcall = m$call)
      }, implements = "log_messages")
    )
  }
}


#' Injects a logger call to standard warnings
#'
#' This function sets a hook to trigger [log_warn()] when `warning` is called
#' to log the warning messages with the global `logger` layout and appender.
#' @param muffle if TRUE, the warning is not shown after being logged
#' @export
#' @examples \dontrun{
#' log_warnings()
#' for (i in 1:5) {
#'   Sys.sleep(runif(1))
#'   warning(i)
#' }
#' }
log_warnings <- function(muffle = getOption("logger_muffle_warnings", FALSE)) {
  if (any(sapply(
    globalCallingHandlers()[names(globalCallingHandlers()) == "warning"],
    attr,
    which = "implements"
  ) == "log_warnings")) {
    warning("Ignoring this call to log_warnings as it was registered previously.")
  } else {
    globalCallingHandlers(
      warning = structure(function(m) {
        logger::log_level(logger::WARN, m$message, .topcall = m$call)
        if (isTRUE(muffle)) {
          invokeRestart("muffleWarning")
        }
      }, implements = "log_warnings")
    )
  }
}

#' Injects a logger call to standard errors
#'
#' This function sets a hook to trigger [log_error()] when [stop()] is called
#' to log the error messages with the global `logger` layout and appender.
#' @param muffle if TRUE, the error is not thrown after being logged
#' @param traceback if TRUE the error traceback is logged along with the error
#' message
#' @export
#' @examples \dontrun{
#' log_errors()
#' stop("foobar")
#' }
log_errors <- function(muffle = getOption("logger_muffle_errors", FALSE), traceback = FALSE) {
  if (any(sapply(
    globalCallingHandlers()[names(globalCallingHandlers()) == "error"],
    attr,
    which = "implements"
  ) == "log_errors")) {
    warning("Ignoring this call to log_errors as it was registered previously.")
  } else {
    globalCallingHandlers(
      error = structure(function(m) {
        logger::log_level(logger::ERROR, m$message, .topcall = m$call)
        if (traceback) {
          bt <- .traceback(3L)
          logger::log_level(logger::ERROR, "Traceback:", .topcall = m$call)
          for (i in seq_along(bt)) {
            msg <- paste0(length(bt) - i + 1L, ": ", bt[[i]])
            ref <- attr(bt[[i]], "srcref")
            file <- attr(ref, "srcfile")
            if (inherits(file, "srcfile")) {
              file <- basename(file$filename)
              line <- paste(unique(c(ref[1L], ref[3L])), collapse = "-")
              msg <- paste0(msg, " at ", file, " #", line)
            }
            logger::log_level(logger::ERROR, skip_formatter(msg), .topcall = m$call)
          }
        }
        if (isTRUE(muffle)) {
          invokeRestart("abort")
        }
      }, implements = "log_errors")
    )
  }
}


#' Auto logging input changes in Shiny app
#'
#' This is to be called in the `server` section of the Shiny app.
#' @export
#' @param input passed from Shiny's `server`
#' @param level log level
#' @param excluded_inputs character vector of input names to exclude from logging
#' @param namespace the name of the namespace
#' @importFrom utils assignInMyNamespace assignInNamespace
#' @examples \dontrun{
#' library(shiny)
#'
#' ui <- bootstrapPage(
#'   numericInput("mean", "mean", 0),
#'   numericInput("sd", "sd", 1),
#'   textInput("title", "title", "title"),
#'   textInput("foo", "This is not used at all, still gets logged", "foo"),
#'   passwordInput("password", "Password not to be logged", "secret"),
#'   plotOutput("plot")
#' )
#'
#' server <- function(input, output) {
#'   logger::log_shiny_input_changes(input, excluded_inputs = "password")
#'
#'   output$plot <- renderPlot({
#'     hist(rnorm(1e3, input$mean, input$sd), main = input$title)
#'   })
#' }
#'
#' shinyApp(ui = ui, server = server)
#' }
log_shiny_input_changes <- function(input,
                                    level = INFO,
                                    namespace = NA_character_,
                                    excluded_inputs = character()) {
  fail_on_missing_package("shiny")
  fail_on_missing_package("jsonlite")

  session <- shiny::getDefaultReactiveDomain()
  ns <- ifelse(!is.null(session), session$ns(character(0)), "")

  if (!(shiny::isRunning() || inherits(session, "MockShinySession") || inherits(session, "session_proxy"))) {
    stop("No Shiny app running, it makes no sense to call this function outside of a Shiny app")
  }

  input_values <- shiny::isolate(shiny::reactiveValuesToList(input))
  assignInMyNamespace("shiny_input_values", input_values)
  log_level(level, skip_formatter(trimws(paste(
    ns,
    "Default Shiny inputs initialized:",
    as.character(jsonlite::toJSON(input_values, auto_unbox = TRUE))
  ))), namespace = namespace)

  shiny::observe({
    old_input_values <- shiny_input_values
    new_input_values <- shiny::reactiveValuesToList(input)
    names <- unique(c(names(old_input_values), names(new_input_values)))
    names <- setdiff(names, excluded_inputs)
    for (name in names) {
      old <- old_input_values[name]
      new <- new_input_values[name]
      if (!identical(old, new)) {
        message <- trimws("{ns} Shiny input change detected in {name}: {old} -> {new}")
        log_level(level, message, namespace = namespace)
      }
    }
    assignInNamespace("shiny_input_values", new_input_values, ns = "logger")
  })
}
shiny_input_values <- NULL

test_that("log_messages", {
  expect_snapshot({
    writeLines(eval_outside("log_messages()", "message(42)"))
  })
})

test_that("log_warnings", {
  expect_snapshot({
    writeLines(eval_outside("log_warnings(TRUE)", "warning(42)", "log(-1)"))
  })
})

test_that("log_errors", {
  expect_snapshot({
    writeLines(eval_outside("log_errors()", "stop(42)"))
    writeLines(eval_outside("log_errors()", "foobar"))
    writeLines(eval_outside("log_errors()", 'f<-function(x) {42 * "foobar"}; f()'))
    writeLines(eval_outside("log_errors(traceback = TRUE)",
                            'source("helper.R", keep.source = TRUE)',
                            "function_that_fails()"))
  })
})

test_that("shiny input initialization is detected", {
  obs <- eval_outside("
    .globals <- shiny:::.globals
    .globals$appState <- new.env(parent = emptyenv())
    server <- function(input, output, session) {
        logger::log_shiny_input_changes(input)
    }
    shiny::testServer(server, {})
  ")
  expect_snapshot(writeLines(obs))
})

test_that("shiny input initialization is detected with different log-level", {
  obs <- eval_outside("
    .globals <- shiny:::.globals
    .globals$appState <- new.env(parent = emptyenv())
    server <- function(input, output, session) {
        logger::log_shiny_input_changes(input, level = logger::ERROR)
    }
    shiny::testServer(server, {})
  ")
  expect_snapshot(writeLines(obs))
})

test_that("shiny input change is detected", {
  obs <- eval_outside("
    .globals <- shiny:::.globals
    .globals$appState <- new.env(parent = emptyenv())
    server <- function(input, output, session) {
        logger::log_shiny_input_changes(input)
        x <- shiny::reactive(input$a)
    }
    shiny::testServer(server, {
        session$setInputs(a = 2)
    })
  ")
  expect_snapshot(writeLines(obs))
})

test_that("shiny input change is logged with different level", {
  obs <- eval_outside("
    .globals <- shiny:::.globals
    .globals$appState <- new.env(parent = emptyenv())
    server <- function(input, output, session) {
        logger::log_shiny_input_changes(input, level = logger::ERROR)
        x <- shiny::reactive(input$a)
    }
    shiny::testServer(server, {
        session$setInputs(a = 2)
    })
  ")
  expect_snapshot(writeLines(obs))
})

library(logger)
library(testthat)

context('hooks')

eval_outside <- function(expr) {
    t <- tempfile()
    on.exit(unlink(t))
    cat('library(logger); log_messages(); log_warnings(); log_errors();', file = t)
    cat(expr, file = t, append = TRUE, sep = '\n')
    paste(
        suppressWarnings(system(paste('$R_HOME/bin/Rscript', t, '2>&1'), intern = TRUE)),
        collapse = '\n')
}

test_that('log_messages', {
    expect_match(eval_outside('message(42)'), 'INFO')
    if (R.Version()$os == 'linux-gnu') {
        expect_match(eval_outside('system("echo 42", invisible = TRUE)'), 'INFO')
    }
})

test_that('log_warnings', {
    expect_match(eval_outside('warning(42)'), 'WARN')
    if (R.Version()$major >= 4) {
        expect_match(eval_outside('log(-1)'), 'WARN')
    }
})

test_that('log_errors', {
    expect_match(eval_outside('stop(42)'), 'ERROR')
    if (R.Version()$major >= 4) {
        expect_match(eval_outside('foobar'), 'ERROR')
        expect_match(eval_outside('f<-function(x) {42 * "foobar"}; f()'), 'ERROR')
    }
})

---
title: "Introduction to logger"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to logger}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(logger)
log_appender(appender_stdout)
```

If you are not only using R in the interactive console for ad-hoc data analysis, but running eg batch jobs (for ETL, reporting, modeling, forecasting etc) as well, then logging the status(changes) of your script is a must so that later on you can review / debug what have happened.

For most cases, it's enough to load the package and use the functions with the `log` prefix to log important and not so important messages, for example:

```{r}
library(logger)
log_info("Loading data")
data(mtcars)
log_info("The dataset includes {nrow(mtcars)} rows")
if (max(mtcars$hp) < 1000) {
  log_warn("Oh, no! There are no cars with more than 1K horsepower in the dataset :/")
  log_debug("The most powerful car is {rownames(mtcars)[which.max(mtcars$hp)]} with {max(mtcars$hp)} hp")
}
```

Interestingly, the most powerful car was not being logged -- because by default the `logger` prints messages with at least the `INFO` log level:

```{r}
log_threshold()
```

To change that, specify the new log level threshold, eg `TRACE` to log everything:

```{r}
log_threshold(TRACE)
```

The rerunning the above code chunk:

```{r}
log_info("Loading data")
data(mtcars)
log_info("The dataset includes {nrow(mtcars)} rows")
if (max(mtcars$hp) < 1000) {
  log_warn("Oh, no! There are no cars with more than 1K horsepower in the dataset :/")
  log_debug("The most powerful car is {rownames(mtcars)[which.max(mtcars$hp)]} with {max(mtcars$hp)} hp")
}
```

You may also find the `?log_eval` function useful to log both an R expression and its result in the same log record:

```{r}
f <- sqrt
g <- mean
x <- 1:31
log_eval(y <- f(g(x)), level = INFO)
str(y)
```


Sometimes, it may be reasonable to log R objects as markdown, e.g. a smallish `data.frame` or `data.table`, e.g. `mtcars` or `iris`. Calling the formatter using `pander` instead of `glue` can help:

```{r knitr-pander-setup, include = FALSE}
ppo1 <- pander::panderOptions("knitr.auto.asis")
ppo2 <- pander::panderOptions("table.style")
pander::panderOptions("knitr.auto.asis", FALSE)
pander::panderOptions("table.style", "simple")
```

```{r}
log_formatter(formatter_pander)
log_info(head(iris))
```

```{r knitr-pander-revert, include = FALSE}
pander::panderOptions("knitr.auto.asis", ppo1)
pander::panderOptions("table.style", ppo2)
```

For more details, check the [function reference in the manual](https://daroczig.github.io/logger/reference/index.html), or start with the [The Anatomy of a Log Request](https://daroczig.github.io/logger/articles/anatomy.html) and [Customizing the Format and the Destination of a Log Record](https://daroczig.github.io/logger/articles/customize_logger.html) vignettes.

```{r cleanup, include = FALSE}
logger:::namespaces_reset()
```

---
title: "The Anatomy of a Log Request"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{The Anatomy of a Log Request}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
resource_files:
  - logger_structure.svg
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(logger)
log_appender(appender_stdout)
```

```{r loggerStructureImage, echo=FALSE, out.extra='style="width: 100%;" title="The structure of a logger and the flow of a log record request" alt="The structure of a logger and the flow of a log record request"'}
knitr::include_graphics("logger_structure.svg")
```

To make a successful log record, `logger` requires the below components:

- a **log request**, eg

    ```r
    log_error('Oops')
    ```

    - including the log level (importance) of the record, which will be later used to decide if the log record is to be delivered or not: `ERROR` in this case
    - R objects to be logged: a simple string in this case, although it could be a character vector or any R object(s) that can be converted into a character vector by the `formatter` function

- the **environment** and meta-information of the log request, eg actual timestamp, hostname of the computer, the name of the user running the R script, the pid of the R process, calling function and the actual call etc.


    ```{r}
f <- function() get_logger_meta_variables(log_level = INFO)
f()
    ```

- a **logger definition** to process the log request, including

    - log level `threshold`, eg `INFO`, which defines the minimum log level required for actual logging -- all log requests with lower log level will be thrown away

        ```{r}
log_threshold()
ERROR <= INFO
log_error("Oops")
        ```

    - `formatter` function, which takes R objects and converts those into actual log message(s) to be then passed to the `layout` function for the log record rendering -- such as `paste`, `sprintf`, `glue` or eg the below custom example:


        ```{r}
formatter <- function(...) paste(..., collapse = " ", sep = " ")
formatter(1:3, c("foo", "bar"))
        ```

    - `layout` function, which takes log message(s) and further information on the log request (such as timestamp, hostname, username, calling function etc) to render the actual log records eg human-readable text, JSON etc


        ```r
        library(jsonlite)
        layout <- function(level, msg) toJSON(level = level, timestamp = time, hostname = node, message = msg)
        layout(INFO, 'Happy Thursday!')
        #> {'level': 'INFO', 'timestamp': '1970-01-01 00:00:00', 'hostname': 'foobar', 'message': 'Happy Thursday!'}
        ```

    - `appender` function, which takes fully-rendered log record(s) and delivers to somewhere, eg `stdout`, a file or a streaming service, eg

        ```{r}
appender <- function(line) cat(line, "\n")
appender("INFO [now] I am a log message")
        ```

Putting all these together (by explicitly setting the default config in the `global` namespace):

```{r}
log_threshold(INFO)
log_formatter(formatter_glue)
log_layout(layout_simple)
log_appender(appender_stdout)
log_debug("I am a low level log message that will not be printed with a high log level threshold")
log_warn("I am a higher level log message that is very likely to be printed")
```

Note, that all `logger` definitions and requests are tied to a logging namespace, and one log request might trigger multiple `logger` definitions as well (stacking). Find more information on these in the [Customizing the format and destination of log records](https://daroczig.github.io/logger/articles/customize_logger.html) vignette.

```{r cleanup, include = FALSE}
logger:::namespaces_reset()
```

---
title: "Customizing the Format and the Destination of a Log Record"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Customizing the Format and the Destination of a Log Record}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r pkgchecks, echo = FALSE}
## check if other logger packages are available and exit if not
for (pkg in c("devtools", "parallel")) {
  if (!requireNamespace(pkg, quietly = TRUE)) {
    warning(paste(pkg, "package not available, so cannot build this vignette"))
    knitr::knit_exit()
  }
}
```

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(logger)
log_appender(appender_stdout)
```

In this vignette I suppose that you are already familiar with [The Anatomy of a Log Request](https://daroczig.github.io/logger/articles/anatomy.html) vignette.

## What gets logged?

`logger` mostly relies on and uses the default `log4j` log levels and supports suppressing log messages with a lower log level compared to the currently set threshold in the logging namespace:

```{r}
log_info("Hi, there!")
log_debug("How are you doing today?")
log_threshold()
log_threshold(TRACE)
log_debug("How are you doing today?")
```

So the `?log_threshold` function can both get and set the log level threshold for all future log requests.

For the full list of all supported log levels and so thus the possible log level thresholds, see `?log_levels`.

If you want to define the log level in a programmatic way, check out `?log_level`, eg

```{r}
log_level(INFO, "Hi, there!")
```

To temporarily update the log level threshold, you may also find the `?with_log_threshold` function useful:

```{r}
log_threshold(INFO)
log_debug("pst, can you hear me?")
log_info("no")

with_log_threshold(log_debug("pst, can you hear me?"), threshold = TRACE)
log_info("yes")

with_log_threshold(
  {
    log_debug("pst, can you hear me?")
    log_info("yes")
  },
  threshold = TRACE
)
```

You can also define your own log level(s) if needed, for example introducing an extra level between `DEBUG` and `INFO`:

```{r}
FYI <- structure(450L, level = "FYI", class = c("loglevel", "integer"))
log_threshold(FYI)
log_debug("ping")
log_level(FYI, "ping")
log_info("pong")
```

## Log namespaces

By default, all log messages will be processed by the global `logger` definition, but you may also use custom namespaces (eg to deliver specific log records to a special destination or to apply a custom log level threshold) and even multiple loggers as well within the very same namespace (eg to deliver all `INFO` and above log levels in the console and everything below that to a trace log file).

If you specify an unknown `namespace` in a log request, it will fall back to the global settings:

```{r}
log_threshold(INFO)
log_trace("Hi, there!", namespace = "kitchensink")
log_info("Hi, there!", namespace = "kitchensink")
```

But once you start customizing that namespace, it gets forked from the global settings and live on its own without modifying the original namespace:


```{r}
log_threshold(TRACE, namespace = "kitchensink")
log_trace("Hi, there!", namespace = "kitchensink")
log_info("Hi, there!", namespace = "kitchensink")
log_trace("Hi, there!")
```

## Log message formatter functions

In the above example, we logged strings without any dynamic parameter, so the task of the logger was quite easy. But in most cases you want to log a parameterized string and the formatter function's task to transform that to a regular character vector.

By default, `logger` uses `glue` in the background:

```{r}
log_formatter(formatter_glue)
log_info("There are {nrow(mtcars)} cars in the mtcars dataset")
log_info("2 + 2 = {2+2}")
```

If you don't like this syntax, or want to save a dependency, you can use other formatter functions as well, such as `?formatter_sprintf` (being the default in eg the [`logging` and `futile.logger` packages](https://daroczig.github.io/logger/articles/migration.html)) or `?formatter_paste`, or [write your own formatter function](https://daroczig.github.io/logger/articles/write_custom_extensions.html) converting R objects into string.

## Log message layouts

By default, `?log_level` and its derivative functions (eg `?log_info`) will simply record the log-level, the current timestamp and the message after being processed by `glue`:

```{r}
log_info(42)
log_info("The answer is {42}")
log_info("The answers are {1:5}")
```

In the above example, first, `42` was converted to a string by the `?formatter_glue` message formatter, then the message was passed to the `?layout_simple` layout function to generate the actual log record.

An example of another layout function writing the same log messages in JSON:

```{r}
log_layout(layout_json())
log_info(42)
log_info("The answer is {42}")
log_info("The answers are {1:5}")
```

If you need colorized logs highlighting the important log messages, check out `?layout_glue_colors`, and for other formatter and layout functions, see the manual of the above mentioned functions that have references to all the other functions and generator functions bundled with the package.

## Custom log record layout

To define a custom format on how the log messages should be rendered, you may write your own `formatter` and `layout` function(s) or rely on the function generator functions bundled with the `logger` package, such as `?layout_glue_generator`.

This function returns a `layout` function that you can define by `glue`-ing together variables describing the log request via `?get_logger_meta_variables`, so having easy access to (package) namespace, calling function's name, hostname, user running the R process etc.

A quick example:

* define custom logger:

    ```{r}
logger <- layout_glue_generator(format = "{node}/{pid}/{namespace}/{fn} {time} {level}: {msg}")
log_layout(logger)
    ```

* check what's being logged when called from the global environment:

    ```{r}
log_info("foo")
    ```

* check what's being logged when called from a custom function:

    ```{r}
f <- function() log_info("foo")
f()
    ```

* check what's being logged when called from a package:

    ```{r}
devtools::load_all(system.file("demo-packages/logger-tester-package", package = "logger"))
logger_tester_function(INFO, "hi from tester package")
    ```

* suppress messages in a namespace:

    ```{r}
log_threshold(namespace = "logger.tester")
log_threshold(WARN, namespace = "logger.tester")
logger_tester_function(INFO, "hi from tester package")
logger_tester_function(WARN, "hi from tester package")
log_info("I am still working in the global namespace")
    ```

Another example of making use of the generator function is to update the layout to include the Process ID that might be very useful eg when forking, see for example the below code chunk still using the above defined log layout:

```r
f <- function(x) {
    log_info('received {length(x)} values')
    log_success('with the mean of {mean(x)}')
    mean(x)
}
library(parallel)
mclapply(split(runif(100), 1:10), f, mc.cores = 5)
#> nevermind/26448/R_GlobalEnv/FUN 2018-12-02 21:54:11 INFO: received 10 values
#> nevermind/26448/R_GlobalEnv/FUN 2018-12-02 21:54:11 SUCCESS: with the mean of 0.403173440974206
#> nevermind/26449/R_GlobalEnv/FUN 2018-12-02 21:54:11 INFO: received 10 values
#> nevermind/26448/R_GlobalEnv/FUN 2018-12-02 21:54:11 INFO: received 10 values
#> nevermind/26449/R_GlobalEnv/FUN 2018-12-02 21:54:11 SUCCESS: with the mean of 0.538581100990996
#> nevermind/26448/R_GlobalEnv/FUN 2018-12-02 21:54:11 SUCCESS: with the mean of 0.485734378430061
#> nevermind/26450/R_GlobalEnv/FUN 2018-12-02 21:54:11 INFO: received 10 values
#> nevermind/26449/R_GlobalEnv/FUN 2018-12-02 21:54:11 INFO: received 10 values
#> nevermind/26450/R_GlobalEnv/FUN 2018-12-02 21:54:11 SUCCESS: with the mean of 0.580483326432295
#> nevermind/26452/R_GlobalEnv/FUN 2018-12-02 21:54:11 INFO: received 10 values
#> nevermind/26449/R_GlobalEnv/FUN 2018-12-02 21:54:11 SUCCESS: with the mean of 0.461282140854746
#> nevermind/26450/R_GlobalEnv/FUN 2018-12-02 21:54:11 INFO: received 10 values
#> nevermind/26451/R_GlobalEnv/FUN 2018-12-02 21:54:11 INFO: received 10 values
#> nevermind/26450/R_GlobalEnv/FUN 2018-12-02 21:54:11 SUCCESS: with the mean of 0.465152264293283
#> nevermind/26452/R_GlobalEnv/FUN 2018-12-02 21:54:11 SUCCESS: with the mean of 0.618332817289047
#> nevermind/26451/R_GlobalEnv/FUN 2018-12-02 21:54:11 SUCCESS: with the mean of 0.493527933699079
#> nevermind/26452/R_GlobalEnv/FUN 2018-12-02 21:54:11 INFO: received 10 values
#> nevermind/26452/R_GlobalEnv/FUN 2018-12-02 21:54:11 SUCCESS: with the mean of 0.606248055002652
#> nevermind/26451/R_GlobalEnv/FUN 2018-12-02 21:54:11 INFO: received 10 values
#> nevermind/26451/R_GlobalEnv/FUN 2018-12-02 21:54:11 SUCCESS: with the mean of 0.537314630229957
```

*Note that the `layout_glue_generator` functions also adds a special attribute to the resulting formatting function so that when printing the layout function to the console, the user can easily interpret what's being used instead of just showing the actual functions's body:*

```{r}
log_layout()
```

For more details on this, see the [Writing custom logger extensions](https://daroczig.github.io/logger/articles/write_custom_extensions.html) vignette.

```{r}
## reset layout
log_layout(layout_simple)
```

## Delivering log records to their destination

By default, `logger` will write to the `stderr` via the `?appender_console` function:

```{r}
log_appender()
```

To write to a logfile instead, use the `?appender_file` generator function, that returns a function that can be used in any namespace:

```{r}
t <- tempfile()
log_appender(appender_file(t))
log_info("where is this message going?")
log_appender()
readLines(t)
unlink(t)
```

There's a similar generator function that returns an appender function delivering log messages to Slack channels:

```r
## load Slack configuration, API token etc from a (hopefully encrypted) yaml file or similar
slack_config <- config::config(...)
## redirect log messages to Slack
log_appender(appender_slack(
    channel   = '#gergely-test',
    username  = 'logger',
    api_token = slack_config$token
), namespace = 'slack')
log_info('Excited about sending logs to Slack!')
#> INFO [2018-11-28 00:21:13] Excited about sending logs to Slack!
log_info('Hi there from logger@R!', namespace = 'slack')
```

You may find `?appender_tee` also useful, that writes the log messages to both `stdout` and a file.

```{r}
## reset appender
log_appender(appender_stdout)
```

And the are many other appender functions bundled with `logger` as well, eg some writing to Syslog, Telegram, Pushbullet, a database table or an Amazon Kinesis stream -- even doing that asynchronously via `appender_async` -- see [Simple Benchmarks on Performance](https://daroczig.github.io/logger/articles/performance.html) for more details.

## Stacking loggers

Note that the `?appender_tee` functionality can be implemented by stacking loggers as well, eg setting two loggers for the global namespace: `?appender_console` and `?appender_file`. The advantage of this approach is that you can set different log level thresholds for each logger, for example:

```{r}
log_threshold()

## create a new logger with index 2
log_threshold(TRACE, index = 2)

## note that the original logger still have the same log level threshold
log_threshold()
log_threshold(index = 2)

## update the appender of the new logger
t <- tempfile()
log_appender(appender_file(t), index = 2)

## test both loggers
log_info("info msg")
log_debug("info msg")

readLines(t)
unlink(t)
```

```{r cleanup, include = FALSE}
logger:::namespaces_reset()
```

---
title: "Migration Guide"
vignette: >
  %\VignetteIndexEntry{Migration Guide}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{css, echo=FALSE}
div.comparison {
    width: 49%;
    display: inline-block;
    vertical-align: top;
}
div.comparison p package {
    font-family: monospace;
    color: brown;
}
div#tocnav ul.nav li ul {
    padding-left: 10px;
}
```

```{r pkgchecks, echo = FALSE}
## check if other logger packages are available and exit if not
for (pkg in c("futile.logger", "logging", "log4r")) {
  if (!requireNamespace(pkg, quietly = TRUE)) {
    warning(paste(pkg, "package not available, so cannot build this vignette"))
    knitr::knit_exit()
  }
}
```

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## load the main package first
library(logger)
log_appender(appender_stdout)
```

In this vignette I suppose that you are already familiar with at least one of the [similar logging R packages](https://daroczig.github.io/logger/index.html#why-another-logging-r-package) and you are looking for suggestions on how to switch to `logger`. Before moving forward, please make sure that you have read the [Introduction to logger](https://daroczig.github.io/logger/articles/Intro.html), [The Anatomy of a Log Request](https://daroczig.github.io/logger/articles/anatomy.html) and [Customizing the Format and the Destination of a Log Record](https://daroczig.github.io/logger/articles/customize_logger.html) vignettes for a decent background on `logger`, and use this vignette as a quick-reference sheet to help you migrate from another package.

## futile.logger

The `logger` package has been very heavily inspired by [`futile.logger`](https://cran.r-project.org/package=futile.logger) and have been using it for many years, also opened multiple pull requests to extend `futile.logger` before I decided to revamp my ideas into a new R package -- but there are still many common things between `futile.logger` and `logger`.

### Initialize

Both packages comes with a default log engine / config, so it's enough to load the packages and those are ready to be used right away:

<div class="comparison"><package>futile.logger</package>
```{r}
library(futile.logger)
```
</div>
<div class="comparison"><package>logger</package>
```{r}
library(logger)
```
</div>

### Logging functions

The most important change is that function names are by snake_case in `logger`, while `futile.logger` uses dot.separated expressions, and `futile.logger` prefixes function names by `flog` while `logger` uses `log` for that:

<div class="comparison"><package>futile.logger</package>
```{r}
flog.info("hi there")
flog.warn("watch out")
```
</div>
<div class="comparison"><package>logger</package>
```{r}
log_info("hi there")
log_warn("watch out")
```
</div>

As you can see above, the default layout of the messages is exactly the same.

### Log levels

Regarding log levels, `futile.logger` bundles the default `log4j` levels (`TRACE`, `DEBUG`, `INFO`, `WARN`, `ERROR` and `FATAL`) that is extended by `SUCCESS` in `logger` as sometimes it's worth logging with a higher than `INFO` level that something succeeded.

### Log record layout

Changing layouts is easy in both package, as you simply pass a layout function:

<div class="comparison"><package>futile.logger</package>
```{r}
flog.layout(layout.json)
flog.info("hi again")
```
</div>
<div class="comparison"><package>logger</package>
```{r}
log_layout(layout_json())

log_info("hi again")
```
</div>

As you can see, `logger` provided a bit more information about the log request compared to `futile.logger`, but it's easy to change the list of fields to be used in the JSON -- see `?get_logger_meta_variables` for a complete list of variable names to be passed to `?layout_json`. `logger` also ships a lot more layouts, eg `?layout_glue_colors` or roll out your own via the `?layout_glue_generator` factory function.

```{r echo=FALSE, results='hide'}
flog.layout(layout.simple)
log_layout(layout_simple)
```

### Log message formatting

By default, `futile.logger` uses an `sprintf` formatter, while `logger` passes the objects to be logged to `glue`:

<div class="comparison"><package>futile.logger</package>
```{r}
flog.info("hi")
flog.info("hi %s", 84 / 2)
flog.info(paste("hi", 84 / 2))
flog.info(glue::glue("hi {84/2}"))
```
</div>
<div class="comparison"><package>logger</package>
```{r}
log_info("hi")
log_info("hi {84/2}")
log_formatter(formatter_sprintf)
log_info("hi %s", 84 / 2)
log_formatter(formatter_paste)
log_info("hi", 84 / 2)
```
</div>


```{r echo=FALSE, results='hide'}
log_formatter(formatter_glue)
```

It's easy to change this default formatter in both packages: use `flog.layout` handles this as well in `futile.logger`, while the formatter is separated from the layout function in `logger`, so check `?log_formatter` instead. `logger` ships with a bit more formatter functions, eg the default `?formatter_glue` and `?formatter_glue_or_sprintf` that tries to combine the best from both words.

### Log record destination

Setting the destination of the log records works similarly in both packages, although he `logger` packages bundles a lot more options:

<div class="comparison"><package>logging</package>
```{r}
t <- tempfile()
flog.appender(appender.file(t))
flog.appender(appender.tee(t))
```
</div>
<div class="comparison"><package>logger</package>
```{r}
t <- tempfile()
log_appender(appender_file(t))
log_appender(appender_tee(t))
```
</div>


```{r echo=FALSE, results='hide'}
flog.appender(appender.console)
log_appender(appender_stdout)
```

### Hierarchical logging and performance

Both packages support using different logging namespaces and stacking loggers within the same namespace. Performance-wise, there's `logger` seems to be faster than `futile.logger`, but for more details, check the [Simple Benchmarks on Performance](https://daroczig.github.io/logger/articles/performance.html) vignette.

### Using `logger` as a drop-in-replacement of `futile.logger`

`logger` has no hard requirements, so it's a very lightweight alternative of `futile.logger`. Although the function names are a bit different, and the message formatter also differs, but with some simple tweaks, `logger` can become an almost perfect drop-in-replacement of `futile.logger`:

```{r}
library(logger)
log_formatter(formatter_sprintf)
flog.trace <- log_trace
flog.debug <- log_debug
flog.info <- log_info
flog.warn <- log_warn
flog.error <- log_error

flog.info("Hello from logger in a futile.logger theme ...")
flog.warn("... where the default log message formatter is %s", "sprintf")
```

## logging

The [`logging`](https://cran.r-project.org/package=logging) package behaves very similarly to the Python logging module and so thus being pretty Pythonic, while `logger` tries to accommodate native R users' expectations -- so there are some minor nuances between the usage of the two packages.

### Initialize

In `logging`, you have to initialize a logger first via `addHandler` or simply by calling `basicConfig`, which is not required in `logger` as it already comes with a default log config:

<div class="comparison"><package>logging</package>
```{r}
library(logging)
basicConfig()
```
</div>
<div class="comparison"><package>logger</package>
```{r}
library(logger)
```
</div>

### Logging functions

After initializing the logging engine, actual logging works similarly in the two packages -- with a bit different function names:

* although `logging` uses mostly camelCase function names (eg `basicConfig`), but the logging functions are all lowercase without any separator, such as `loginfo` or `logwarn`
* `logger` uses snake_case for the function names, such as `log_info` and `log_warn`

<div class="comparison"><package>logging</package>
```{r}
loginfo("hi there")
logwarn("watch out")
```
</div>
<div class="comparison"><package>logger</package>
```{r}
log_info("hi there")
log_warn("watch out")
```
</div>

As you can see above, the default layout of the log messages is somewhat different:

* `logging` starts with the timestamp that is followed by the log level, optional namespace and the message separated by colons
* `logger` starts with the log level, followed by the timestamp between brackets and then the message

### Log levels

For the available log levels in `logging`, check `?loglevels`, and `?log_levels` for the same in `logger`:

<div class="comparison"><package>logging</package>
```{r}
str(as.list(loglevels))
```
</div>
<div class="comparison"><package>logger</package>
```{r}
levels <- mget(rev(logger:::log_levels_supported), envir = asNamespace("logger"))
str(levels, give.attr = FALSE)
```
</div>

### Performance

Performance-wise, there's no big difference between the two packages, but for more details, check the [Simple Benchmarks on Performance](https://daroczig.github.io/logger/articles/performance.html) vignette.

### Log record layout

Getting and setting the layout of the log record should happen up-front in both packages:

<div class="comparison"><package>logging</package>
```{r}
getLogger()[["handlers"]]$basic.stdout$formatter
```
</div>
<div class="comparison"><package>logger</package>
```{r}
log_layout()
```
</div>

`logger` provides multiple configurable layouts to fit the user's need, eg easily show the calling function of the lof request, the `pid` of the R process, name of the machine etc. or colorized outputs. See [Customizing the Format and the Destination of a Log Record](https://daroczig.github.io/logger/articles/customize_logger.html) vignette for more details.

### Log message formatting

If you want to pass dynamic log messages to the log engines, you can do that via the hard-coded `sprintf` in the `logging` package, while you can set that on a namespaces basis in `logger`, which is by default using `glue`:

<div class="comparison"><package>logging</package>
```{r}
loginfo("hi")
loginfo("hi %s", 84 / 2)
loginfo(paste("hi", 84 / 2))
loginfo(glue::glue("hi {84/2}"))
```
</div>
<div class="comparison"><package>logger</package>
```{r}
log_info("hi")
log_info("hi {84/2}")
log_formatter(formatter_sprintf)
log_info("hi %s", 84 / 2)
log_formatter(formatter_paste)
log_info("hi", 84 / 2)
```
</div>

For even better compatibility, there's also `?formatter_logging` that not only relies on `sprintf` when the first argument is a string, but will log the call and the result as well when the log object is an R expression:

```{r}
log_formatter(formatter_logging)
log_info("42")
log_info(42)
log_info(4 + 2)
log_info("foo %s", "bar")
log_info(12, 1 + 1, 2 * 2)
```

```{r echo=FALSE, results='hide'}
log_formatter(formatter_glue)
```

### Log record destination

Setting the destination of the log records works similarly in both packages, although he `logger` packages bundles a lot more options:

<div class="comparison"><package>logging</package>
```r
?addHandler
?writeToConsole
?writeToFile
```
</div>
<div class="comparison"><package>logger</package>
```r
?log_appender
?appender_console
?appender_file
?appender_tee
?appender_slack
?appender_pushbullet
```
</div>

### Hierarchical logging

Both packages support using different logging namespaces and stacking loggers within the same namespace.

### Using `logger` as a drop-in-replacement of `logging`

`logger` has no hard requirements, so it's an adequate alternative of `logging`. Although the function names are a bit different, and the message formatter also differs, but with some simple tweaks, `logger` can become an almost perfect drop-in-replacement of `logging` -- although not all log levels (eg \code{FINE} and \code{CRITICAL}) are supported:

```{r}
library(logger)
log_formatter(formatter_logging)
log_layout(layout_logging)
logdebug <- log_debug
loginfo <- log_info
logwarn <- log_warn
logerror <- log_error

loginfo("Hello from logger in a logging theme ...")
logwarn("... where the default log message formatter is %s", "sprintf", namespace = "foobar")
```

## log4r

The [`log4r`](https://cran.r-project.org/package=log4r) package provides an object-oriented approach for logging in R, so the logger object is to be passed to the log calls -- unlike in the `logger` package.

### Initialize

So thus it's important to create a logging object in `log4r` before being able to log messages, while that's automatically done in `logger:

<div class="comparison"><package>log4r</package>
```{r}
library(log4r)
logger <- create.logger(logfile = stdout(), level = "INFO")
```
</div>
<div class="comparison"><package>logger</package>
```{r}
library(logger)
```
</div>

Please note that in the background, `logger` does have a concept of logger objects, but that's behind the scene and the user does not have to specify / reference it. On the other hand, if you wish, you can do that via the `namespace` concept of `logger` -- more on that later.

### Logging functions

While `logger` has a `log_` prefix for all logging functions, `log4r` has lowercase functions names referring to the log level, which takes a logging object and the log message:

<div class="comparison"><package>log4r</package>
```{r}
info(logger, "hi there")
warn(logger, "watch out")
```
</div>
<div class="comparison"><package>logger</package>
```{r}
log_info("hi there")
log_warn("watch out")
```
</div>

As you can see the default layout of the messages is a bit different in the two packages.

### Log levels

Both packages are based on `log4j`, and `log4r` provides `DEBUG`, `INFO`, `WARN`, `ERROR` and `FATAL`, while `logger` also adds `TRACE` and `SUCCESS` on the top of these.

To change the log level threshold, use the `level` function on the logging object in `log4r`, while it's `log_level` in `logger`.

### Log record layout and formatter functions

The `log4r` provides a `logformat` argument in `create.logger` that can be used to override the default formatting, while `logger` provides formatter and layout functions for a flexible log record design.

### Log record destination

By default, `log4r` logs to a file that can be set to `stoud` to write to the console, while `logger` writes to the console by default, but logging to files via the `appender_file` functions is also possible -- besides a number of other log record destinations as well.

### Hierarchical logging and performance

Creating objects is the `log4r` way of handling multiple log environments, while `logger` handles that via `namespace`s.

## loggit and loggr

Both [`loggit`](https://cran.r-project.org/package=loggit) and [`loggr`](https://github.com/smbache/loggr) provides an easy way of capturing `message`, `warning` and `stop` function messages. Similar functionality has been added to `logger` in the form of `log_message`, `log_warning` and `log_error` functions.

```{r cleanup, include = FALSE}
logger:::namespaces_reset()
detach("package:logger", unload = TRUE)
detach("package:futile.logger", unload = TRUE)
detach("package:logging", unload = TRUE)
```

---
title: "Writing Custom Logger Extensions"
vignette: >
  %\VignetteIndexEntry{Writing Custom Logger Extensions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

In this vignette I suppose that you are already familiar with [Customizing the format and destination of log records](https://daroczig.github.io/logger/articles/customize_logger.html) vignette.

## Custom log message formatter functions

The log message formatter function should be able to take any number of R objects and convert those into a character vector that is safe to pass to the layout function.

This transformer function can be as simple as calling `paste`, `glue` or `sprintf` or something complex as well, eg looking up user attributes of a user id mentioned in a log record etc.

When writing a custom formatter function, it should also accept the original logging function call as `.logcall`, the parent call as `.topcall` and the log request's environment as `.topenv`, which can be used to find the relevant variables in your formatter.

If you are writing such a function or a generator function returning a log message formatter function, please keep the actual call resulting in the formatter function (eg `match.call()` in the generator function or the quoted function call) recorded in the `generator` attribute of the function so that `?log_formatter` can pretty-print that instead of the unnamed function body. See the [`formatters.R`](https://github.com/daroczig/logger/blob/master/R/formatters.R) file for more examples.

## Custom log layout rendering functions

The layout functions return the log record and take at least two arguments:

- the log level and
- a message already formatted as a string by a log message formatter function
- the namespace (as `namespace`), calling function (as `.topcall`) and its environment (as `.topenv`) of the log request, and the actual log call (as `.logcall`) automatically collected by `?log_level`

Such layout rendering function can be as simple as the default `?layout_simple`:

```r
layout_simple <- function(level, msg, ...) {
    paste0(attr(level, 'level'), ' [', format(Sys.time(), "%Y-%m-%d %H:%M:%S"), '] ', msg)
}
```

Or much more complex, eg looking up the hostname of the machine, public IP address etc and logging all these automatically with the message of the log request.

Your easiest option to set up a custom layout is calling `?layout_glue_generator` that comes with a nice API being able to access a bunch of meta-information on the log request via `?get_logger_meta_variables`. For such example, see the [Customizing the format and destination of log records](https://daroczig.github.io/logger/articles/customize_logger.html) vignette.

If you are writing such a function or a generator function returning a log message formatter function, please keep the actual call resulting in the formatter function (eg `match.call()` in the generator function or the quoted function call) recorded in the `generator` attribute of the function so that `?log_layout` can pretty-print that instead of the unnamed function body. See the [`layouts.R`](https://github.com/daroczig/logger/blob/master/R/layouts.R) file for more examples.

## Custom log record appenders

The appender functions take log records and delivers those to the desired destination. This can be as simple as writing to the console (`?appender_console`) or to a local file (`?appender_file`), or delivering the log record via an API request to a remote service, streaming somewhere or sending a Slack message (`?appender_slack`).

If you are writing such a function or a generator function returning a log message formatter function, please keep the actual call resulting in the formatter function (eg `match.call()` in the generator function or the quoted function call) recorded in the `generator` attribute of the function so that `?log_appender` can pretty-print that instead of the unnamed function body. See the [`appenders.R`](https://github.com/daroczig/logger/blob/master/R/appenders.R) file for more examples.

An example for a custom appender delivering log messages to a database table:

```r
## the dbr package provides and easy and secure way of connecting to databased from R
## although if you want to minimize the dependencies, feel free to stick with DBI etc.
library(dbr)
## init a persistent connection to the database using a yaml config in the background thanks to dbr
## NOTE that this is optional and a temporarily connection could be also used
##      for higher reliability but lower throughput
con <- db_connect('mydb')
## define custom function writing the log message to a table
log_appender(function(lines) {
    db_append(
        df = data.frame(timestamp = Sys.time(), message = lines),
        table = 'logs', db = con)
})
```

```{r cleanup, include = FALSE}
logger:::namespaces_reset()
```

---
title: "Logging from R Packages"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Logging from R Packages}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r pkgchecks, echo = FALSE}
## check if other logger packages are available and exit if not
for (pkg in c("devtools")) {
  if (!requireNamespace(pkg, quietly = TRUE)) {
    warning(paste(pkg, "package not available, so cannot build this vignette"))
    knitr::knit_exit()
  }
}
```

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

In this vignette I suppose that you are already familiar with [Customizing the format and destination of log records](https://daroczig.github.io/logger/articles/customize_logger.html) vignette, especially with the [Log namespaces](https://daroczig.github.io/logger/articles/customize_logger.html#log-namespaces) section.

So that your R package's users can suppress (or render with custom layout) the log messages triggered by your R package, it's wise to record all those log messages in a custom namespace. By default, if you are calling the `?log_level` function family from an R package after importing from the `logger` package, then `logger` will try to auto-guess the calling R package name and use that as the default namespace, see eg:

```{r}
library(logger)
devtools::load_all(system.file("demo-packages/logger-tester-package", package = "logger"))
logger_tester_function(INFO, "hi from tester package")
```

But if auto-guessing is not your style, then feel free to set your custom namespace (eg the name of your package) in all `?log_info` etc function calls and let your users know about how to suppress / reformat / redirect your log messages via `?log_threshold`, `?log_layout`, `?log_appender`.

Please note that setting the formatter function via `?log_formatter` should not be left to the R package end-users, as the log message formatter is specific to your logging calls, so that should be decided by the R package author. Feel free to pick any formatter function (eg `glue`, `sprintf`, `paste` or something else), and set that via `?log_formatter` when your R package is loaded. All other parameters of your `logger` will inherit from the `global` namespace -- set by your R package's end user.

```{r cleanup, include = FALSE}
logger:::namespaces_reset()
```

---
title: "The Anatomy of a Log Request"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{The Anatomy of a Log Request}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)
library(logger)
## backup settings
oldconf <- list(
    threshold = log_threshold(),
    layout = log_layout(),
    formatter = log_formatter(),
    appender = log_appender())
## knitr not picking up stderr
log_appender(appender_stdout)
```

```{r loggerStructureImage, echo=FALSE, out.extra='style="width: 100%;" title="The structure of a logger and the flow of a log record request" alt="The structure of a logger and the flow of a log record request"'}
knitr::include_graphics("logger_structure.svg")
```

To make a successful log record, `logger` requires the below components:

- a **log request**, eg

    ```r
    log_error('Oops')
    ```

    - including the log level (importance) of the record, which will be later used to decide if the log record is to be delivered or not: `ERROR` in this case
    - R objects to be logged: a simple string in this case, although it could be a character vector or any R object(s) that can be converted into a character vector by the `formatter` function

- the **environment** and meta-information of the log request, eg actual timestamp, hostname of the computer, the name of the user running the R script, the pid of the R process, calling function and the actual call etc.


    ```{r}
    f <- function() get_logger_meta_variables(log_level = INFO)
    f()
    ```

- a **logger definition** to process the log request, including

    - log level `threshold`, eg `INFO`, which defines the minimum log level required for actual logging -- all log requests with lower log level will be thrown away

        ```{r}
        log_threshold()
        ERROR <= INFO
        log_error('Oops')
        ```

    - `formatter` function, which takes R objects and converts those into actual log message(s) to be then passed to the `layout` function for the log record rendering -- such as `paste`, `sprintf`, `glue` or eg the below custom example:


        ```{r}
        formatter <- function(...) paste(..., collapse = ' ', sep = ' ')
        formatter(1:3, c('foo', 'bar'))
        ```

    - `layout` function, which takes log message(s) and further information on the log request (such as timestamp, hostname, username, calling function etc) to render the actual log records eg human-readable text, JSON etc


        ```r
        library(jsonlite)
        layout <- function(level, msg) toJSON(level = level, timestamp = time, hostname = node, message = msg)
        layout(INFO, 'Happy Thursday!')
        #> {'level': 'INFO', 'timestamp': '1970-01-01 00:00:00', 'hostname': 'foobar', 'message': 'Happy Thursday!'}
        ```

    - `appender` function, which takes fully-rendered log record(s) and delivers to somewhere, eg `stdout`, a file or a streaming service, eg

        ```{r}
        appender <- function(line) cat(line, '\n')
        appender('INFO [now] I am a log message')
        ```

Putting all these together (by explicitly setting the default config in the `global` namespace):

```{r}
log_threshold(INFO)
log_formatter(formatter_glue)
log_layout(layout_simple)
log_appender(appender_console)
log_debug('I am a low level log message that will not be printed with a high log level threshold')
log_warn('I am a higher level log message that is very likely to be printed')
```

Note, that all `logger` definitions and requests are tied to a logging namespace, and one log request might trigger multiple `logger` definitions as well (stacking). Find more information on these in the [Customizing the format and destination of log records](https://daroczig.github.io/logger/articles/customize_logger.html) vignette.

```{r cleanup, include = FALSE}
## restore settings
log_threshold(oldconf$threshold)
log_layout(oldconf$layout)
log_formatter(oldconf$formatter)
log_appender(oldconf$appender)
```

---
title: "Introduction to logger"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to logger}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(logger)
## backup settings
oldconf <- list(
    threshold = log_threshold(),
    layout = log_layout(),
    formatter = log_formatter(),
    appender = log_appender())
## knitr not picking up stderr
log_appender(appender_stdout)
```

If you are not only using R in the interactive console for ad-hoc data analysis, but running eg batch jobs (for ETL, reporting, modeling, forecasting etc) as well, then logging the status(changes) of your script is a must so that later on you can review / debug what have happened.

For most cases, it's enough to load the package and use the functions with the `log` prefix to log important and not so important messages, for example:

```{r}
library(logger)
log_info('Loading data')
data(mtcars)
log_info('The dataset includes {nrow(mtcars)} rows')
if (max(mtcars$hp) < 1000) {
    log_warn('Oh, no! There are no cars with more than 1K horsepower in the dataset :/')
    log_debug('The most powerful car is {rownames(mtcars)[which.max(mtcars$hp)]} with {max(mtcars$hp)} hp')
}
```

Interestingly, the most powerful car was not being logged -- because by default the `logger` prints messages with at least the `INFO` log level:

```{r}
log_threshold()
```

To change that, specify the new log level threshold, eg `TRACE` to log everything:

```{r}
log_threshold(TRACE)
```

The rerunning the above code chunk:

```{r}
log_info('Loading data')
data(mtcars)
log_info('The dataset includes {nrow(mtcars)} rows')
if (max(mtcars$hp) < 1000) {
    log_warn('Oh, no! There are no cars with more than 1K horsepower in the dataset :/')
    log_debug('The most powerful car is {rownames(mtcars)[which.max(mtcars$hp)]} with {max(mtcars$hp)} hp')
}
```

You may also find the `?log_eval` function useful to log both an R expression and its result in the same log record:

```{r}
f <- sqrt
g <- mean
x <- 1:31
log_eval(y <- f(g(x)), level = INFO)
str(y)
```

For more details, check the [function reference in the manual](https://daroczig.github.io/logger/reference/index.html), or start with the [The Anatomy of a Log Request](https://daroczig.github.io/logger/articles/anatomy.html) and [Customizing the Format and the Destination of a Log Record](https://daroczig.github.io/logger/articles/customize_logger.html) vignettes.

```{r cleanup, include = FALSE}
## restore settings
log_threshold(oldconf$threshold)
log_layout(oldconf$layout)
log_formatter(oldconf$formatter)
log_appender(oldconf$appender)
```

# JSON layout warns if you include msg

    Code
      layout <- layout_json(fields = "msg")
    Condition
      Warning in `layout_json()`:
      'msg' is always automatically included

# must throw errors

    Code
      layout_simple(FOOBAR)
    Condition
      Error:
      ! object 'FOOBAR' not found
    Code
      layout_simple(42)
    Condition
      Error in `layout_simple()`:
      ! argument "msg" is missing, with no default
    Code
      layout_simple(msg = "foobar")
    Condition
      Error in `layout_simple()`:
      ! argument "level" is missing, with no default

---

    Code
      layout_glue(FOOBAR)
    Condition
      Error:
      ! object 'FOOBAR' not found
    Code
      layout_glue(42)
    Condition
      Error in `layout_glue()`:
      ! Invalid log level, see ?log_levels
    Code
      layout_glue(msg = "foobar")
    Condition
      Error in `layout_glue()`:
      ! argument "level" is missing, with no default
    Code
      layout_glue(level = 53, msg = "foobar")
    Condition
      Error in `layout_glue()`:
      ! Invalid log level, see ?log_levels

# log_info() captures local info

    Code
      log_info("foobar")
    Output
      logger / global / logger / eval / eval(expr, envir)
    Code
      f()
    Output
      logger / global / logger / f / f()
    Code
      g()
    Output
      logger / global / logger / f / f()

# log_info() captures package info

    Code
      logger_tester_function(INFO, "x = ")
    Output
      logger.tester INFO x =  0.0807501375675201
    Code
      logger_info_tester_function("everything = ")
    Output
      logger.tester INFO everything =  42

# fail_on_missing_package

    Code
      fail_on_missing_package("logger", "9.9.9", call = quote(f()))
    Condition
      Error:
      ! Please install min. 9.9.9 version of logger to use f
    Code
      fail_on_missing_package("an.R.package-that-doesNotExists", call = quote(f()))
    Condition
      Error:
      ! Please install the 'an.R.package-that-doesNotExists' package to use f

# validate_log_level

    Code
      validate_log_level("FOOBAR")
    Condition
      Error in `validate_log_level()`:
      ! Invalid log level

# setters check inputs

    Code
      log_appender(1)
    Condition
      Error in `log_appender()`:
      ! `appender` must be a function
    Code
      log_formatter(1)
    Condition
      Error in `log_formatter()`:
      ! `formatter` must be a function
    Code
      log_layout(1)
    Condition
      Error in `log_layout()`:
      ! `layout` must be a function
    Code
      log_threshold("x")
    Condition
      Error in `validate_log_level()`:
      ! Invalid log level

# glue works

    Code
      formatter_glue("malformed {")
    Condition
      Error in `h()`:
      ! `glue` failed in `formatter_glue` on:
      
       chr "malformed {"
      
      Raw error message:
      
      Expecting '}'
      
      Please consider using another `log_formatter` or `skip_formatter` on strings with curly braces.

# glue gives informative error if message contains curlies

    Code
      log_info("hi{")
    Condition
      Error in `h()`:
      ! `glue` failed in `formatter_glue` on:
      
       chr "hi{"
      
      Raw error message:
      
      Expecting '}'
      
      Please consider using another `log_formatter` or `skip_formatter` on strings with curly braces.

# glue_safe works

    Code
      formatter_glue_safe("Hi {42}")
    Condition
      Error in `value[[3L]]()`:
      ! `glue_safe` failed in `formatter_glue_safe` on:
      
        chr "Hi {42}" 
      
      Raw error message:
      
       object '42' not found 
      
      Please consider using another `log_formatter` or `skip_formatter` on strings with curly braces.
    Code
      formatter_glue_safe("malformed {")
    Condition
      Error in `value[[3L]]()`:
      ! `glue_safe` failed in `formatter_glue_safe` on:
      
        chr "malformed {" 
      
      Raw error message:
      
       Expecting '}' 
      
      Please consider using another `log_formatter` or `skip_formatter` on strings with curly braces.

# sprintf works

    Code
      formatter_sprintf("%s and %i", 1)
    Condition
      Error in `sprintf()`:
      ! too few arguments

# skip formatter

    Code
      log_info(skip_formatter("hi {x}", x = 4))
    Condition
      Error in `skip_formatter()`:
      ! Cannot skip the formatter function if further arguments are passed besides the actual log message(s)

# %except% logs errors and returns default value

    Code
      out <- f()
    Output
      except / global / logger / f / f()
      WARN Running '1' as 'FunDoesNotExist(1:10)' failed: 'could not find function "FunDoesNotExist"'

# log_messages

    Code
      writeLines(eval_outside("log_messages()", "message(42)"))
    Output
      INFO 42

# log_warnings

    Code
      writeLines(eval_outside("log_warnings(TRUE)", "warning(42)", "log(-1)"))
    Output
      WARN 42
      WARN NaNs produced

# log_errors

    Code
      writeLines(eval_outside("log_errors()", "stop(42)"))
    Output
      ERROR 42
    Code
      writeLines(eval_outside("log_errors()", "foobar"))
    Output
      ERROR object 'foobar' not found
    Code
      writeLines(eval_outside("log_errors()", "f<-function(x) {42 * \"foobar\"}; f()"))
    Output
      ERROR non-numeric argument to binary operator
    Code
      writeLines(eval_outside("log_errors(traceback = TRUE)",
        "source(\"helper.R\", keep.source = TRUE)", "function_that_fails()"))
    Output
      ERROR I'm failing
      ERROR Traceback:
      ERROR 2: stop("I'm failing") at helper.R #46
      ERROR 1: function_that_fails()

# shiny input initialization is detected

    Code
      writeLines(obs)
    Output
      INFO mock-session Default Shiny inputs initialized: {}

# shiny input initialization is detected with different log-level

    Code
      writeLines(obs)
    Output
      ERROR mock-session Default Shiny inputs initialized: {}

# shiny input change is detected

    Code
      writeLines(obs)
    Output
      INFO mock-session Default Shiny inputs initialized: {}
      INFO mock-session Shiny input change detected in a: NULL -> 2

# shiny input change is logged with different level

    Code
      writeLines(obs)
    Output
      ERROR mock-session Default Shiny inputs initialized: {}
      ERROR mock-session Shiny input change detected in a: NULL -> 2

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatters.R
\name{formatter_glue_safe}
\alias{formatter_glue_safe}
\title{Apply \code{\link[glue:glue_safe]{glue::glue_safe()}} to convert R objects into a character vector}
\usage{
formatter_glue_safe(
  ...,
  .logcall = sys.call(),
  .topcall = sys.call(-1),
  .topenv = parent.frame()
)
}
\arguments{
\item{...}{passed to \code{\link[glue:glue_safe]{glue::glue_safe()}} for the text interpolation}

\item{.logcall}{the logging call being evaluated (useful in
formatters and layouts when you want to have access to the raw,
unevaluated R expression)}

\item{.topcall}{R expression from which the logging function was
called (useful in formatters and layouts to extract the calling
function's name or arguments)}

\item{.topenv}{original frame of the \code{.topcall} calling function
where the formatter function will be evaluated and that is used
to look up the \code{namespace} as well via \code{logger:::top_env_name}}
}
\value{
character vector
}
\description{
Apply \code{\link[glue:glue_safe]{glue::glue_safe()}} to convert R objects into a character vector
}
\seealso{
Other log_formatters: 
\code{\link{formatter_cli}()},
\code{\link{formatter_glue}()},
\code{\link{formatter_glue_or_sprintf}()},
\code{\link{formatter_json}()},
\code{\link{formatter_logging}()},
\code{\link{formatter_pander}()},
\code{\link{formatter_paste}()},
\code{\link{formatter_sprintf}()}
}
\concept{log_formatters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{log_elapsed}
\alias{log_elapsed}
\alias{log_elapsed_start}
\title{Log cumulative running time}
\usage{
log_elapsed(..., level = INFO, namespace = NA_character_)

log_elapsed_start(level = INFO, namespace = NA_character_, quiet = FALSE)
}
\arguments{
\item{...}{passed to \code{\link[=log_level]{log_level()}}}

\item{level}{see \code{\link[=log_levels]{log_levels()}}}

\item{namespace}{x}

\item{quiet}{Should starting the time emit a log message}
}
\description{
This function is working like \code{\link[=log_tictoc]{log_tictoc()}} but differs in that it continues
to count up rather than resetting the timer at every call. You can set the
start time using \code{log_elapsed_start()}, but if that hasn't been called it
will show the time since the R session started.
}
\examples{
log_elapsed_start()
Sys.sleep(0.4)
log_elapsed("Tast 1")
Sys.sleep(0.2)
log_elapsed("Task 2")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layouts.R
\name{layout_glue_colors}
\alias{layout_glue_colors}
\title{Format a log message with \code{glue} and ANSI escape codes to add colors}
\usage{
layout_glue_colors(
  level,
  msg,
  namespace = NA_character_,
  .logcall = sys.call(),
  .topcall = sys.call(-1),
  .topenv = parent.frame()
)
}
\arguments{
\item{level}{log level, see \code{\link{log_levels}} for more details}

\item{msg}{string message}

\item{namespace}{string referring to the \code{logger} environment / config to be used to override the target of the message record to be used instead of the default namespace, which is defined by the R package name from which the logger was called, and falls back to a common, global namespace.}

\item{.logcall}{the logging call being evaluated (useful in formatters and layouts when you want to have access to the raw, unevaluated R expression)}

\item{.topcall}{R expression from which the logging function was called (useful in formatters and layouts to extract the calling function's name or arguments)}

\item{.topenv}{original frame of the \code{.topcall} calling function where the formatter function will be evaluated and that is used to look up the \code{namespace} as well via \code{logger:::top_env_name}}
}
\value{
character vector
}
\description{
Format a log message with \code{glue} and ANSI escape codes to add colors
}
\note{
This functionality depends on the \pkg{crayon} package.
}
\examples{
\dontrun{
log_layout(layout_glue_colors)
log_threshold(TRACE)
log_info('Starting the script...')
log_debug('This is the second line')
log_trace('That is being placed right after the first one.')
log_warn('Some errors might come!')
log_error('This is a problem')
log_debug('Getting an error is usually bad')
log_error('This is another problem')
log_fatal('The last problem.')
}
}
\seealso{
This is a \code{\link{log_layout}}, for alternatives, see \code{\link{layout_blank}}, \code{\link{layout_simple}}, \code{\link{layout_glue}}, \code{\link{layout_json}}, \code{\link{layout_json_parser}}, or generator functions such as \code{\link{layout_glue_generator}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layouts.R
\name{layout_glue_generator}
\alias{layout_glue_generator}
\title{Generate log layout function using common variables available via glue syntax}
\usage{
layout_glue_generator(
  format = "{level} [{format(time, \\"\%Y-\%m-\%d \%H:\%M:\%S\\")}] {msg}"
)
}
\arguments{
\item{format}{\code{\link[glue:glue]{glue::glue()}}-flavored layout of the message using the above
variables}
}
\value{
function taking \code{level} and \code{msg} arguments - keeping the
original call creating the generator in the \code{generator} attribute
that is returned when calling \code{\link[=log_layout]{log_layout()}} for the currently
used layout
}
\description{
\code{format} is passed to \code{\link[glue:glue]{glue::glue()}} with access to the below variables:
\itemize{ \item msg: the actual log message \item further variables
set by \code{\link[=get_logger_meta_variables]{get_logger_meta_variables()}} }
}
\examples{
\dontshow{old <- logger:::namespaces_set()}
example_layout <- layout_glue_generator(
  format = "{node}/{pid}/{ns}/{ans}/{topenv}/{fn} {time} {level}: {msg}"
)
example_layout(INFO, "try {runif(1)}")

log_layout(example_layout)
log_info("try {runif(1)}")
\dontshow{logger:::namespaces_set(old)}
}
\seealso{
See example calls from \code{\link[=layout_glue]{layout_glue()}} and \code{\link[=layout_glue_colors]{layout_glue_colors()}}.

Other log_layouts: 
\code{\link{get_logger_meta_variables}()},
\code{\link{layout_blank}()},
\code{\link{layout_gha}()},
\code{\link{layout_glue}()},
\code{\link{layout_glue_colors}()},
\code{\link{layout_json}()},
\code{\link{layout_json_parser}()},
\code{\link{layout_logging}()},
\code{\link{layout_simple}()}
}
\concept{log_layouts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatters.R
\name{formatter_sprintf}
\alias{formatter_sprintf}
\title{Apply \code{sprintf} to convert R objects into a character vector}
\usage{
formatter_sprintf(
  fmt,
  ...,
  .logcall = sys.call(),
  .topcall = sys.call(-1),
  .topenv = parent.frame()
)
}
\arguments{
\item{fmt}{passed to \code{sprintf}}

\item{...}{passed to \code{sprintf}}

\item{.logcall}{the logging call being evaluated (useful in formatters and layouts when you want to have access to the raw, unevaluated R expression)}

\item{.topcall}{R expression from which the logging function was called (useful in formatters and layouts to extract the calling function's name or arguments)}

\item{.topenv}{original frame of the \code{.topcall} calling function where the formatter function will be evaluated and that is used to look up the \code{namespace} as well via \code{logger:::top_env_name}}
}
\value{
character vector
}
\description{
Apply \code{sprintf} to convert R objects into a character vector
}
\seealso{
This is a \code{\link{log_formatter}}, for alternatives, see \code{\link{formatter_paste}}, \code{\link{formatter_glue}}, \code{\link{formatter_glue_or_sprintf}}, \code{\link{formatter_logging}}, \code{\link{formatter_json}}, \code{\link{formatter_pander}} and \code{\link{skip_formatter}} for marking a string not to apply the formatter on it.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/levels.R
\docType{data}
\name{log_levels}
\alias{log_levels}
\alias{OFF}
\alias{FATAL}
\alias{ERROR}
\alias{WARN}
\alias{SUCCESS}
\alias{INFO}
\alias{DEBUG}
\alias{TRACE}
\title{Log levels}
\usage{
OFF

FATAL

ERROR

WARN

SUCCESS

INFO

DEBUG

TRACE
}
\description{
The standard Apache logj4 log levels plus a custom level for
\code{SUCCESS}. For the full list of these log levels and suggested
usage, check the below Details.
}
\details{
List of supported log levels:
\itemize{
\item \code{OFF} No events will be logged
\item \code{FATAL} Severe error that will prevent the application from continuing
\item \code{ERROR} An error in the application, possibly recoverable
\item \code{WARN} An event that might possible lead to an error
\item \code{SUCCESS} An explicit success event above the INFO level that you want to log
\item \code{INFO} An event for informational purposes
\item \code{DEBUG} A general debugging event
\item \code{TRACE} A fine-grained debug message, typically capturing the flow through the application.
}
}
\references{
\url{https://logging.apache.org/log4j/2.x/javadoc/log4j-api/org/apache/logging/log4j/Level.html},
\url{https://logging.apache.org/log4j/2.x/manual/customloglevels.html}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/appenders.R
\name{appender_syslognet}
\alias{appender_syslognet}
\title{Send log messages to a network syslog server}
\usage{
appender_syslognet(identifier, server, port = 601L)
}
\arguments{
\item{identifier}{program/function identification (string).}

\item{server}{machine where syslog daemon runs (string).}

\item{port}{port where syslog daemon listens (integer).}
}
\value{
A function taking a \code{lines} argument.
}
\description{
Send log messages to a network syslog server
}
\note{
This functionality depends on the \pkg{syslognet} package.
}
\examples{
\dontrun{
if (requireNamespace("syslognet", quietly = TRUE)) {
  log_appender(appender_syslognet("test_app", 'remoteserver'))
  log_info("Test message.")
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tester.R
\name{logger_info_tester_function}
\alias{logger_info_tester_function}
\title{Testing logging INFO from package}
\usage{
logger_info_tester_function(msg)
}
\arguments{
\item{msg}{bar}
}
\description{
Testing logging INFO from package
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layouts.R
\name{layout_json}
\alias{layout_json}
\title{Generate log layout function rendering JSON}
\usage{
layout_json(fields = default_fields())
}
\arguments{
\item{fields}{character vector of field names to be included in the
JSON}
}
\value{
character vector
}
\description{
Generate log layout function rendering JSON
}
\note{
This functionality depends on the \pkg{jsonlite} package.
}
\examples{
\dontshow{old <- logger:::namespaces_set()}
log_layout(layout_json())
log_info(42)
log_info("ok {1:3} + {1:3} = {2*(1:3)}")
\dontshow{logger:::namespaces_set(old)}
}
\seealso{
Other log_layouts: 
\code{\link{get_logger_meta_variables}()},
\code{\link{layout_blank}()},
\code{\link{layout_gha}()},
\code{\link{layout_glue}()},
\code{\link{layout_glue_colors}()},
\code{\link{layout_glue_generator}()},
\code{\link{layout_json_parser}()},
\code{\link{layout_logging}()},
\code{\link{layout_simple}()}
}
\concept{log_layouts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/appenders.R
\name{appender_void}
\alias{appender_void}
\title{Dummy appender not delivering the log record to anywhere}
\usage{
appender_void(lines)
}
\arguments{
\item{lines}{character vector}
}
\description{
Dummy appender not delivering the log record to anywhere
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{log_separator}
\alias{log_separator}
\title{Logs a long line to stand out from the console}
\usage{
log_separator(
  level = INFO,
  namespace = NA_character_,
  separator = "=",
  width = 80,
  .logcall = sys.call(),
  .topcall = sys.call(-1),
  .topenv = parent.frame(),
  .timestamp = Sys.time()
)
}
\arguments{
\item{level}{log level, see \code{\link[=log_levels]{log_levels()}} for more details}

\item{namespace}{string referring to the \code{logger} environment /
config to be used to override the target of the message record to
be used instead of the default namespace, which is defined by the
R package name from which the logger was called, and falls back
to a common, global namespace.}

\item{separator}{character to be used as a separator}

\item{width}{max width of message -- longer text will be wrapped into multiple lines}

\item{.logcall}{the logging call being evaluated (useful in
formatters and layouts when you want to have access to the raw,
unevaluated R expression)}

\item{.topcall}{R expression from which the logging function was
called (useful in formatters and layouts to extract the calling
function's name or arguments)}

\item{.topenv}{original frame of the \code{.topcall} calling function
where the formatter function will be evaluated and that is used
to look up the \code{namespace} as well via \code{logger:::top_env_name}}

\item{.timestamp}{The time the logging occured. Defaults to the current time
but may be overwritten if the logging is delayed from the time it happend}
}
\description{
Logs a long line to stand out from the console
}
\examples{
\dontshow{old <- logger:::namespaces_set()}
log_separator()
log_separator(ERROR, separator = "!", width = 60)
log_separator(ERROR, separator = "!", width = 100)
logger <- layout_glue_generator(format = "{node}/{pid}/{namespace}/{fn} {time} {level}: {msg}")
log_layout(logger)
log_separator(ERROR, separator = "!", width = 100)
log_layout(layout_blank)
log_separator(ERROR, separator = "!", width = 80)
\dontshow{logger:::namespaces_set(old)}
}
\seealso{
\code{\link[=log_with_separator]{log_with_separator()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hooks.R
\name{warn_if_globalCallingHandlers_is_not_available}
\alias{warn_if_globalCallingHandlers_is_not_available}
\title{Warn to update R to 4+}
\usage{
warn_if_globalCallingHandlers_is_not_available()
}
\description{
Warn to update R to 4+
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logger.R
\name{log_layout}
\alias{log_layout}
\title{Get or set log record layout}
\usage{
log_layout(layout = NULL, namespace = "global", index = 1)
}
\arguments{
\item{layout}{function defining the structure of a log record, eg
\code{\link[=layout_simple]{layout_simple()}}, \code{\link[=layout_glue]{layout_glue()}} or \code{\link[=layout_glue_colors]{layout_glue_colors()}},
\code{\link[=layout_json]{layout_json()}}, or generator functions such as
\code{\link[=layout_glue_generator]{layout_glue_generator()}}, default NULL}

\item{namespace}{logger namespace}

\item{index}{index of the logger within the namespace}
}
\description{
Get or set log record layout
}
\examples{
\dontshow{old <- logger:::namespaces_set()}
log_layout(layout_json())
log_info(42)
\dontshow{logger:::namespaces_set(old)}
}
\seealso{
Other log configutation functions: 
\code{\link{log_appender}()},
\code{\link{log_formatter}()},
\code{\link{log_threshold}()}
}
\concept{log configutation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logger.R
\name{log_formatter}
\alias{log_formatter}
\title{Get or set log message formatter}
\usage{
log_formatter(formatter = NULL, namespace = "global", index = 1)
}
\arguments{
\item{formatter}{function defining how R objects are converted into
a single string, eg \code{\link[=formatter_paste]{formatter_paste()}}, \code{\link[=formatter_sprintf]{formatter_sprintf()}},
\code{\link[=formatter_glue]{formatter_glue()}}, \code{\link[=formatter_glue_or_sprintf]{formatter_glue_or_sprintf()}},
\code{\link[=formatter_logging]{formatter_logging()}}, default NULL}

\item{namespace}{logger namespace}

\item{index}{index of the logger within the namespace}
}
\description{
Get or set log message formatter
}
\seealso{
Other log configutation functions: 
\code{\link{log_appender}()},
\code{\link{log_layout}()},
\code{\link{log_threshold}()}
}
\concept{log configutation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hooks.R
\name{log_errors}
\alias{log_errors}
\title{Injects a logger call to standard errors}
\usage{
log_errors(
  muffle = getOption("logger_muffle_errors", FALSE),
  traceback = FALSE
)
}
\arguments{
\item{muffle}{if TRUE, the error is not thrown after being logged}

\item{traceback}{if TRUE the error traceback is logged along with the error
message}
}
\description{
This function sets a hook to trigger \code{\link[=log_error]{log_error()}} when \code{\link[=stop]{stop()}} is called
to log the error messages with the global \code{logger} layout and appender.
}
\examples{
\dontrun{
log_errors()
stop("foobar")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logger.R
\name{logger}
\alias{logger}
\title{Generate logging utility}
\usage{
logger(threshold, formatter, layout, appender)
}
\arguments{
\item{threshold}{omit log messages below this \code{\link[=log_levels]{log_levels()}}}

\item{formatter}{function pre-processing the message of the log
record when it's not wrapped in a \code{\link[=skip_formatter]{skip_formatter()}} call}

\item{layout}{function rendering the layout of the actual log
record}

\item{appender}{function writing the log record}
}
\value{
A function taking the log \code{level} to compare with the set
threshold, all the \code{...} arguments passed to the formatter
function, besides the standard \code{namespace}, \code{.logcall},
\code{.topcall} and \code{.topenv} arguments (see \code{\link[=log_level]{log_level()}} for more
details). The function invisibly returns a list including the
original \code{level}, \code{namespace}, all \code{...} transformed to a list as
\code{params}, the log \code{message} (after calling the \code{formatter}
function) and the log \code{record} (after calling the \code{layout}
function), and a list of \code{handlers} with the \code{formatter},
\code{layout} and \code{appender} functions.
}
\description{
A logger consists of a log level \code{threshold}, a log message
\code{formatter} function, a log record \code{layout} formatting function and
the \code{appender} function deciding on the destination of the log
record. For more details, see the package \code{README.md}.
}
\details{
By default, a general logger definition is created when loading the \code{logger} package, that uses
\itemize{
\item \code{\link[=INFO]{INFO()}} (or as per the \code{LOGGER_LOG_LEVEL} environment variable override) as the log level threshold
\item \code{\link[=layout_simple]{layout_simple()}} as the layout function showing the log level, timestamp and log message
\item \code{\link[=formatter_glue]{formatter_glue()}} (or \code{\link[=formatter_sprintf]{formatter_sprintf()}} if \pkg{glue} is not installed) as the
default formatter function transforming the R objects to be logged to a character vector
\item \code{\link[=appender_console]{appender_console()}} as the default log record destination
}
}
\note{
It's quite unlikely that you need to call this function
directly, but instead set the logger parameters and functions at
\code{\link[=log_threshold]{log_threshold()}}, \code{\link[=log_formatter]{log_formatter()}}, \code{\link[=log_layout]{log_layout()}} and
\code{\link[=log_appender]{log_appender()}} and then call \code{\link[=log_levels]{log_levels()}} and its
derivatives, such as \code{\link[=log_info]{log_info()}} directly.
}
\examples{
\dontrun{
do.call(logger, logger:::namespaces$global[[1]])(INFO, 42)
do.call(logger, logger:::namespaces$global[[1]])(INFO, "{pi}")
x <- 42
do.call(logger, logger:::namespaces$global[[1]])(INFO, "{x}^2 = {x^2}")
}
}
\references{
For more details, see the Anatomy of a Log Request
vignette at
\url{https://daroczig.github.io/logger/articles/anatomy.html}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/color.R
\name{colorize_by_log_level}
\alias{colorize_by_log_level}
\alias{grayscale_by_log_level}
\title{Color string by the related log level}
\usage{
colorize_by_log_level(msg, level)

grayscale_by_log_level(msg, level)
}
\arguments{
\item{msg}{String to color.}

\item{level}{see \code{\link[=log_levels]{log_levels()}}}
}
\value{
A string with ANSI escape codes.
}
\description{
Color log messages according to their severity with either a rainbow
or grayscale color scheme. The greyscale theme assumes a dark background on
the terminal.
}
\examples{
\dontshow{if (requireNamespace("crayon")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
cat(colorize_by_log_level("foobar", FATAL), "\n")
cat(colorize_by_log_level("foobar", ERROR), "\n")
cat(colorize_by_log_level("foobar", WARN), "\n")
cat(colorize_by_log_level("foobar", SUCCESS), "\n")
cat(colorize_by_log_level("foobar", INFO), "\n")
cat(colorize_by_log_level("foobar", DEBUG), "\n")
cat(colorize_by_log_level("foobar", TRACE), "\n")

cat(grayscale_by_log_level("foobar", FATAL), "\n")
cat(grayscale_by_log_level("foobar", ERROR), "\n")
cat(grayscale_by_log_level("foobar", WARN), "\n")
cat(grayscale_by_log_level("foobar", SUCCESS), "\n")
cat(grayscale_by_log_level("foobar", INFO), "\n")
cat(grayscale_by_log_level("foobar", DEBUG), "\n")
cat(grayscale_by_log_level("foobar", TRACE), "\n")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logger.R
\name{get_logger_definitions}
\alias{get_logger_definitions}
\title{Find the logger definition(s) specified for the current namespace with a fallback to the global namespace}
\usage{
get_logger_definitions(namespace = NA_character_, .topenv = parent.frame())
}
\arguments{
\item{namespace}{override the default / auto-picked namespace with a custom string}
}
\value{
list of function(s)
}
\description{
Find the logger definition(s) specified for the current namespace with a fallback to the global namespace
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/appenders.R
\name{appender_telegram}
\alias{appender_telegram}
\title{Send log messages to a Telegram chat}
\usage{
appender_telegram(
  chat_id = Sys.getenv("TELEGRAM_CHAT_ID"),
  bot_token = Sys.getenv("TELEGRAM_BOT_TOKEN"),
  parse_mode = NULL
)
}
\arguments{
\item{chat_id}{Unique identifier for the target chat or username of
the target channel (in the format @channelusername)}

\item{bot_token}{Telegram Authorization token}

\item{parse_mode}{Message parse mode. Allowed values: Markdown or
HTML}
}
\value{
function taking \code{lines} argument
}
\description{
Send log messages to a Telegram chat
}
\note{
This functionality depends on the \pkg{telegram} package.
}
\seealso{
Other log_appenders: 
\code{\link{appender_async}()},
\code{\link{appender_console}()},
\code{\link{appender_file}()},
\code{\link{appender_kinesis}()},
\code{\link{appender_ntfy}()},
\code{\link{appender_pushbullet}()},
\code{\link{appender_slack}()},
\code{\link{appender_stdout}()},
\code{\link{appender_syslog}()},
\code{\link{appender_tee}()}
}
\concept{log_appenders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatters.R
\name{formatter_pander}
\alias{formatter_pander}
\title{Formats R objects with pander}
\usage{
formatter_pander(
  x,
  ...,
  .logcall = sys.call(),
  .topcall = sys.call(-1),
  .topenv = parent.frame()
)
}
\arguments{
\item{x}{object to be logged}

\item{...}{optional parameters passed to \code{pander}}

\item{.logcall}{the logging call being evaluated (useful in formatters and layouts when you want to have access to the raw, unevaluated R expression)}

\item{.topcall}{R expression from which the logging function was called (useful in formatters and layouts to extract the calling function's name or arguments)}

\item{.topenv}{original frame of the \code{.topcall} calling function where the formatter function will be evaluated and that is used to look up the \code{namespace} as well via \code{logger:::top_env_name}}
}
\value{
character vector
}
\description{
Formats R objects with pander
}
\note{
This functionality depends on the \pkg{pander} package.
}
\examples{
\dontrun{
log_formatter(formatter_pander)
log_info('42')
log_info(42)
log_info(4+2)
log_info(head(iris))
log_info(head(iris), style = 'simple')
log_info(lm(hp ~ wt, mtcars))
}
}
\seealso{
This is a \code{\link{log_formatter}}, for alternatives, see \code{\link{formatter_paste}}, \code{\link{formatter_sprintf}}, \code{\link{formatter_glue}}, \code{\link{formatter_glue_or_sprintf}}, \code{\link{formatter_logging}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logger.R
\name{validate_log_level}
\alias{validate_log_level}
\title{Assure valid log level}
\usage{
validate_log_level(level)
}
\arguments{
\item{level}{\code{\link{log_levels}} object or string representation}
}
\value{
\code{\link{log_levels}} object
}
\description{
Assure valid log level
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{fail_on_missing_package}
\alias{fail_on_missing_package}
\title{Check if R package can be loaded and fails loudly otherwise}
\usage{
fail_on_missing_package(pkg, min_version, call = NULL)
}
\arguments{
\item{pkg}{string}

\item{min_version}{optional minimum version needed}

\item{call}{Call to include in error message.}
}
\description{
Check if R package can be loaded and fails loudly otherwise
}
\examples{
f <- function() fail_on_missing_package("foobar")
try(f())
g <- function() fail_on_missing_package("stats")
g()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logger.R
\name{log_indices}
\alias{log_indices}
\title{Returns number of currently active indices}
\usage{
log_indices(namespace = "global")
}
\arguments{
\item{namespace}{override the default / auto-picked namespace with
a custom string}
}
\value{
number of indices
}
\description{
Returns number of currently active indices
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/appenders.R
\name{appender_async}
\alias{appender_async}
\title{Delays executing the actual appender function to the future in a
background process to avoid blocking the main R session}
\usage{
appender_async(
  appender,
  namespace = "async_logger",
  init = function() log_info("Background process started")
)
}
\arguments{
\item{appender}{a \code{\link[=log_appender]{log_appender()}} function with a \code{generator}
attribute (TODO note not required, all fn will be passed if
not)}

\item{namespace}{\code{logger} namespace to use for logging messages on
starting up the background process}

\item{init}{optional function to run in the background process that
is useful to set up the environment required for logging, eg if
the \code{appender} function requires some extra packages to be
loaded or some environment variables to be set etc}
}
\value{
function taking \code{lines} argument
}
\description{
Delays executing the actual appender function to the future in a
background process to avoid blocking the main R session
}
\note{
This functionality depends on the \pkg{mirai} package.
}
\examples{
\dontrun{
appender_file_slow <- function(file) {
  force(file)
  function(lines) {
    Sys.sleep(1)
    cat(lines, sep = "\n", file = file, append = TRUE)
  }
}

## log what's happening in the background
log_threshold(TRACE, namespace = "async_logger")
log_appender(appender_console, namespace = "async_logger")

## start async appender
t <- tempfile()
log_info("Logging in the background to {t}")

## use async appender
log_appender(appender_async(appender_file_slow(file = t)))
log_info("Was this slow?")
system.time(for (i in 1:25) log_info(i))

readLines(t)
Sys.sleep(10)
readLines(t)

}
}
\seealso{
Other log_appenders: 
\code{\link{appender_console}()},
\code{\link{appender_file}()},
\code{\link{appender_kinesis}()},
\code{\link{appender_ntfy}()},
\code{\link{appender_pushbullet}()},
\code{\link{appender_slack}()},
\code{\link{appender_stdout}()},
\code{\link{appender_syslog}()},
\code{\link{appender_tee}()},
\code{\link{appender_telegram}()}
}
\concept{log_appenders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logger-package.R
\docType{package}
\name{logger-package}
\alias{logger-package}
\title{logger: A Lightweight, Modern and Flexible Logging Utility}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Inspired by the the 'futile.logger' R package and 'logging' Python module, this utility provides a flexible and extensible way of formatting and delivering log messages with low overhead.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://daroczig.github.io/logger/}
  \item Report bugs at \url{https://github.com/daroczig/logger/issues}
}

}
\author{
\strong{Maintainer}: Gergely Daróczi \email{daroczig@rapporter.net} (\href{https://orcid.org/0000-0003-3149-8537}{ORCID})

Authors:
\itemize{
  \item Hadley Wickham \email{hadley@posit.co} (\href{https://orcid.org/0000-0003-4757-117X}{ORCID})
}

Other contributors:
\itemize{
  \item Jonathan Carroll \email{rpkg@jcarroll.com.au} (\href{https://orcid.org/0000-0002-1404-5264}{ORCID}) [contributor]
  \item Spare Cores [funder]
  \item System1 [funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{deparse_to_one_line}
\alias{deparse_to_one_line}
\title{Deparse and join all lines into a single line}
\usage{
deparse_to_one_line(x)
}
\arguments{
\item{x}{object to \code{deparse}}
}
\value{
string
}
\description{
Calling \code{deparse} and joining all the returned lines into a
single line, separated by whitespace, and then cleaning up all the
duplicated whitespace (except for excessive whitespace in strings
between single or double quotes).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{log_with_separator}
\alias{log_with_separator}
\title{Logs a message in a very visible way}
\usage{
log_with_separator(
  ...,
  level = INFO,
  namespace = NA_character_,
  separator = "=",
  width = 80
)
}
\arguments{
\item{...}{R objects that can be converted to a character vector
via the active message formatter function}

\item{level}{log level, see \code{\link[=log_levels]{log_levels()}} for more details}

\item{namespace}{string referring to the \code{logger} environment /
config to be used to override the target of the message record to
be used instead of the default namespace, which is defined by the
R package name from which the logger was called, and falls back
to a common, global namespace.}

\item{separator}{character to be used as a separator}

\item{width}{max width of message -- longer text will be wrapped into multiple lines}
}
\description{
Logs a message in a very visible way
}
\examples{
\dontshow{old <- logger:::namespaces_set()}
log_with_separator("An important message")
log_with_separator("Some critical KPI down!!!", separator = "$")
log_with_separator("This message is worth a {1e3} words")
log_with_separator(paste(
  "A very important message with a bunch of extra words that will",
  "eventually wrap into a multi-line message for our quite nice demo :wow:"
))
log_with_separator(
  paste(
    "A very important message with a bunch of extra words that will",
    "eventually wrap into a multi-line message for our quite nice demo :wow:"
  ),
  width = 60
)
log_with_separator("Boo!", level = FATAL)
log_layout(layout_blank)
log_with_separator("Boo!", level = FATAL)
logger <- layout_glue_generator(format = "{node}/{pid}/{namespace}/{fn} {time} {level}: {msg}")
log_layout(logger)
log_with_separator("Boo!", level = FATAL, width = 120)
\dontshow{logger:::namespaces_set(old)}
}
\seealso{
\code{\link[=log_separator]{log_separator()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layouts.R
\name{layout_glue}
\alias{layout_glue}
\title{Format a log message with \code{\link[glue:glue]{glue::glue()}}}
\usage{
layout_glue(
  level,
  msg,
  namespace = NA_character_,
  .logcall = sys.call(),
  .topcall = sys.call(-1),
  .topenv = parent.frame(),
  .timestamp = Sys.time()
)
}
\arguments{
\item{level}{log level, see \code{\link[=log_levels]{log_levels()}} for more details}

\item{msg}{string message}

\item{namespace}{string referring to the \code{logger} environment /
config to be used to override the target of the message record to
be used instead of the default namespace, which is defined by the
R package name from which the logger was called, and falls back
to a common, global namespace.}

\item{.logcall}{the logging call being evaluated (useful in
formatters and layouts when you want to have access to the raw,
unevaluated R expression)}

\item{.topcall}{R expression from which the logging function was
called (useful in formatters and layouts to extract the calling
function's name or arguments)}

\item{.topenv}{original frame of the \code{.topcall} calling function
where the formatter function will be evaluated and that is used
to look up the \code{namespace} as well via \code{logger:::top_env_name}}

\item{.timestamp}{The time the logging occured. Defaults to the current time
but may be overwritten if the logging is delayed from the time it happend}
}
\value{
character vector
}
\description{
By default, this layout includes the log level of the log record as
per \code{\link[=log_levels]{log_levels()}}, the current timestamp and the actual log
message -- that you can override via calling
\code{\link[=layout_glue_generator]{layout_glue_generator()}} directly. For colorized output, see
\code{\link[=layout_glue_colors]{layout_glue_colors()}}.
}
\seealso{
Other log_layouts: 
\code{\link{get_logger_meta_variables}()},
\code{\link{layout_blank}()},
\code{\link{layout_gha}()},
\code{\link{layout_glue_colors}()},
\code{\link{layout_glue_generator}()},
\code{\link{layout_json}()},
\code{\link{layout_json_parser}()},
\code{\link{layout_logging}()},
\code{\link{layout_simple}()}
}
\concept{log_layouts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logger.R
\name{log_namespaces}
\alias{log_namespaces}
\title{Looks up logger namespaces}
\usage{
log_namespaces()
}
\value{
character vector of namespace names
}
\description{
Looks up logger namespaces
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{log_chunk_time}
\alias{log_chunk_time}
\title{Automatically log execution time of knitr chunks}
\usage{
log_chunk_time(..., level = INFO, namespace = NA_character_)
}
\arguments{
\item{...}{passed to \code{\link[=log_level]{log_level()}}}

\item{level}{see \code{\link[=log_levels]{log_levels()}}}

\item{namespace}{x}
}
\description{
Calling this function in the first chunk of a document will instruct knitr
to automatically log the execution time of each chunk. If using
\code{\link[=formatter_glue]{formatter_glue()}} or \code{\link[=formatter_cli]{formatter_cli()}} then the \code{options} variable will be
available, providing the chunk options such as chunk label etc.
}
\examples{
# To be put in the first chunk of a document
log_chunk_time("chunk {options$label}")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{log_tictoc}
\alias{log_tictoc}
\title{Tic-toc logging}
\usage{
log_tictoc(..., level = INFO, namespace = NA_character_)
}
\arguments{
\item{...}{passed to \code{log_level}}

\item{level}{x}

\item{namespace}{x}
}
\description{
Tic-toc logging
}
\examples{
\dontrun{
log_tictoc('warming up')
Sys.sleep(0.1)
log_tictoc('running')
Sys.sleep(0.1)
log_tictoc('running')
Sys.sleep(runif(1))
log_tictoc('and running')
}
}
\author{
Thanks to Neal Fultz for the idea and original implementation!
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/appenders.R
\name{appender_tee}
\alias{appender_tee}
\title{Append log messages to a file and stdout as well}
\usage{
appender_tee(
  file,
  append = TRUE,
  max_lines = Inf,
  max_bytes = Inf,
  max_files = 1L
)
}
\arguments{
\item{file}{path}

\item{append}{boolean passed to \code{cat} defining if the file should
be overwritten with the most recent log message instead of
appending}

\item{max_lines}{numeric specifying the maximum number of lines
allowed in a file before rotating}

\item{max_bytes}{numeric specifying the maximum number of bytes
allowed in a file before rotating}

\item{max_files}{integer specifying the maximum number of files to
be used in rotation}
}
\value{
function taking \code{lines} argument
}
\description{
This appends log messages to both console and a file. The same
rotation options are available as in \code{\link[=appender_file]{appender_file()}}.
}
\seealso{
Other log_appenders: 
\code{\link{appender_async}()},
\code{\link{appender_console}()},
\code{\link{appender_file}()},
\code{\link{appender_kinesis}()},
\code{\link{appender_ntfy}()},
\code{\link{appender_pushbullet}()},
\code{\link{appender_slack}()},
\code{\link{appender_stdout}()},
\code{\link{appender_syslog}()},
\code{\link{appender_telegram}()}
}
\concept{log_appenders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatters.R
\name{skip_formatter}
\alias{skip_formatter}
\title{Skip the formatter function}
\usage{
skip_formatter(message, ...)
}
\arguments{
\item{message}{character vector directly passed to the appender
function in \code{\link[=logger]{logger()}}}

\item{...}{should be never set}
}
\value{
character vector with \code{skip_formatter} attribute set to
\code{TRUE}
}
\description{
Adds the \code{skip_formatter} attribute to an object so that logger
will skip calling the formatter function(s). This is useful if you
want to preprocess the log message with a custom function instead
of the active formatter function(s). Note that the \code{message} should
be a string, and \code{skip_formatter} should be the only input for the
logging function to make this work.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/appenders.R
\name{appender_kinesis}
\alias{appender_kinesis}
\title{Send log messages to a Amazon Kinesis stream}
\usage{
appender_kinesis(stream)
}
\arguments{
\item{stream}{name of the Kinesis stream}
}
\value{
function taking \code{lines} and optional \code{partition_key}
argument
}
\description{
Send log messages to a Amazon Kinesis stream
}
\note{
This functionality depends on the \pkg{botor} package.
}
\seealso{
Other log_appenders: 
\code{\link{appender_async}()},
\code{\link{appender_console}()},
\code{\link{appender_file}()},
\code{\link{appender_ntfy}()},
\code{\link{appender_pushbullet}()},
\code{\link{appender_slack}()},
\code{\link{appender_stdout}()},
\code{\link{appender_syslog}()},
\code{\link{appender_tee}()},
\code{\link{appender_telegram}()}
}
\concept{log_appenders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layouts.R
\name{layout_simple}
\alias{layout_simple}
\title{Format a log record by concatenating the log level, timestamp and message}
\usage{
layout_simple(
  level,
  msg,
  namespace = NA_character_,
  .logcall = sys.call(),
  .topcall = sys.call(-1),
  .topenv = parent.frame()
)
}
\arguments{
\item{level}{log level, see \code{\link{log_levels}} for more details}

\item{msg}{string message}

\item{namespace}{string referring to the \code{logger} environment / config to be used to override the target of the message record to be used instead of the default namespace, which is defined by the R package name from which the logger was called, and falls back to a common, global namespace.}

\item{.logcall}{the logging call being evaluated (useful in formatters and layouts when you want to have access to the raw, unevaluated R expression)}

\item{.topcall}{R expression from which the logging function was called (useful in formatters and layouts to extract the calling function's name or arguments)}

\item{.topenv}{original frame of the \code{.topcall} calling function where the formatter function will be evaluated and that is used to look up the \code{namespace} as well via \code{logger:::top_env_name}}
}
\value{
character vector
}
\description{
Format a log record by concatenating the log level, timestamp and message
}
\seealso{
This is a \code{\link{log_layout}}, for alternatives, see \code{\link{layout_blank}}, \code{\link{layout_glue}}, \code{\link{layout_glue_colors}}, \code{\link{layout_json}}, \code{\link{layout_json_parser}}, or generator functions such as \code{\link{layout_glue_generator}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/appenders.R
\name{appender_file}
\alias{appender_file}
\title{Append log messages to a file}
\usage{
appender_file(
  file,
  append = TRUE,
  max_lines = Inf,
  max_bytes = Inf,
  max_files = 1L
)
}
\arguments{
\item{file}{path}

\item{append}{boolean passed to \code{cat} defining if the file should
be overwritten with the most recent log message instead of
appending}

\item{max_lines}{numeric specifying the maximum number of lines
allowed in a file before rotating}

\item{max_bytes}{numeric specifying the maximum number of bytes
allowed in a file before rotating}

\item{max_files}{integer specifying the maximum number of files to
be used in rotation}
}
\value{
function taking \code{lines} argument
}
\description{
Log messages are written to a file with basic log rotation: when
max number of lines or bytes is defined to be other than \code{Inf},
then the log file is renamed with a \code{.1} suffix and a new log file
is created. The renaming happens recursively (eg \code{logfile.1}
renamed to \code{logfile.2}) until the specified \code{max_files}, then the
oldest file (\code{logfile.{max_files-1}}) is deleted.
}
\examples{
\dontshow{old <- logger:::namespaces_set()}
## ##########################################################################
## simple example logging to a file
t <- tempfile()
log_appender(appender_file(t))
for (i in 1:25) log_info(i)
readLines(t)

## ##########################################################################
## more complex example of logging to file
## rotated after every 3rd line up to max 5 files

## create a folder storing the log files
t <- tempfile()
dir.create(t)
f <- file.path(t, "log")

## define the file logger with log rotation enabled
log_appender(appender_file(f, max_lines = 3, max_files = 5L))

## enable internal logging to see what's actually happening in the logrotate steps
log_threshold(TRACE, namespace = ".logger")
## log 25 messages
for (i in 1:25) log_info(i)

## see what was logged
lapply(list.files(t, full.names = TRUE), function(t) {
  cat("\n##", t, "\n")
  cat(readLines(t), sep = "\n")
})

\dontshow{logger:::namespaces_set(old)}
}
\seealso{
Other log_appenders: 
\code{\link{appender_async}()},
\code{\link{appender_console}()},
\code{\link{appender_kinesis}()},
\code{\link{appender_ntfy}()},
\code{\link{appender_pushbullet}()},
\code{\link{appender_slack}()},
\code{\link{appender_stdout}()},
\code{\link{appender_syslog}()},
\code{\link{appender_tee}()},
\code{\link{appender_telegram}()}
}
\concept{log_appenders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logger.R
\name{fallback_namespace}
\alias{fallback_namespace}
\title{Checks if provided namespace exists and falls back to global if not}
\usage{
fallback_namespace(namespace)
}
\arguments{
\item{namespace}{string}
}
\value{
string
}
\description{
Checks if provided namespace exists and falls back to global if not
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatters.R
\name{formatter_json}
\alias{formatter_json}
\title{Transforms all passed R objects into a JSON list}
\usage{
formatter_json(
  ...,
  .logcall = sys.call(),
  .topcall = sys.call(-1),
  .topenv = parent.frame()
)
}
\arguments{
\item{...}{passed to \code{toJSON} wrapped into a \code{list}}

\item{.logcall}{the logging call being evaluated (useful in
formatters and layouts when you want to have access to the raw,
unevaluated R expression)}

\item{.topcall}{R expression from which the logging function was
called (useful in formatters and layouts to extract the calling
function's name or arguments)}

\item{.topenv}{original frame of the \code{.topcall} calling function
where the formatter function will be evaluated and that is used
to look up the \code{namespace} as well via \code{logger:::top_env_name}}
}
\value{
character vector
}
\description{
Transforms all passed R objects into a JSON list
}
\note{
This functionality depends on the \pkg{jsonlite} package.
}
\examples{
\dontshow{old <- logger:::namespaces_set()}
log_formatter(formatter_json)
log_layout(layout_json_parser())
log_info(everything = 42)
log_info(mtcars = mtcars, species = iris$Species)
\dontshow{logger:::namespaces_set(old)}
}
\seealso{
Other log_formatters: 
\code{\link{formatter_glue}()},
\code{\link{formatter_glue_or_sprintf}()},
\code{\link{formatter_glue_safe}()},
\code{\link{formatter_logging}()},
\code{\link{formatter_pander}()},
\code{\link{formatter_paste}()},
\code{\link{formatter_sprintf}()}
}
\concept{log_formatters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/color.R
\name{grayscale_by_log_level}
\alias{grayscale_by_log_level}
\title{Render a string with light/dark gray based on the related log level}
\usage{
grayscale_by_log_level(msg, level)
}
\arguments{
\item{msg}{string}

\item{level}{see \code{\link{log_levels}}}
}
\value{
string with ANSI escape code
}
\description{
Adding color to a string to be used in terminal output. Supports ANSI standard colors 8 or 256.
}
\examples{
\dontrun{
cat(grayscale_by_log_level(FATAL, 'foobar'), '\n')
cat(grayscale_by_log_level(ERROR, 'foobar'), '\n')
cat(grayscale_by_log_level(WARN, 'foobar'), '\n')
cat(grayscale_by_log_level(SUCCESS, 'foobar'), '\n')
cat(grayscale_by_log_level(INFO, 'foobar'), '\n')
cat(grayscale_by_log_level(DEBUG, 'foobar'), '\n')
cat(grayscale_by_log_level(TRACE, 'foobar'), '\n')
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/levels.R
\name{as.loglevel}
\alias{as.loglevel}
\title{Convert R object into a logger log-level}
\usage{
as.loglevel(x)
}
\arguments{
\item{x}{string or integer}
}
\value{
pander log-level, e.g. \code{INFO}
}
\description{
Convert R object into a logger log-level
}
\examples{
as.loglevel(INFO)
as.loglevel(400L)
as.loglevel(400)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hooks.R
\name{log_shiny_input_changes}
\alias{log_shiny_input_changes}
\title{Auto logging input changes in Shiny app}
\usage{
log_shiny_input_changes(
  input,
  level = INFO,
  namespace = NA_character_,
  excluded_inputs = character()
)
}
\arguments{
\item{input}{passed from Shiny's \code{server}}

\item{level}{log level}

\item{namespace}{the name of the namespace}

\item{excluded_inputs}{character vector of input names to exclude from logging}
}
\description{
This is to be called in the \code{server} section of the Shiny app.
}
\examples{
\dontrun{
library(shiny)

ui <- bootstrapPage(
  numericInput("mean", "mean", 0),
  numericInput("sd", "sd", 1),
  textInput("title", "title", "title"),
  textInput("foo", "This is not used at all, still gets logged", "foo"),
  passwordInput("password", "Password not to be logged", "secret"),
  plotOutput("plot")
)

server <- function(input, output) {
  logger::log_shiny_input_changes(input, excluded_inputs = "password")

  output$plot <- renderPlot({
    hist(rnorm(1e3, input$mean, input$sd), main = input$title)
  })
}

shinyApp(ui = ui, server = server)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/appenders.R
\name{appender_syslog}
\alias{appender_syslog}
\title{Send log messages to the POSIX system log}
\usage{
appender_syslog(identifier, ...)
}
\arguments{
\item{identifier}{A string identifying the process.}

\item{...}{Further arguments passed on to \code{\link[rsyslog:syslog]{rsyslog::open_syslog()}}.}
}
\value{
function taking \code{lines} argument
}
\description{
Send log messages to the POSIX system log
}
\note{
This functionality depends on the \pkg{rsyslog} package.
}
\examples{
\dontrun{
if (requireNamespace("rsyslog", quietly = TRUE)) {
  log_appender(appender_syslog("test"))
  log_info("Test message.")
}
}
}
\seealso{
Other log_appenders: 
\code{\link{appender_async}()},
\code{\link{appender_console}()},
\code{\link{appender_file}()},
\code{\link{appender_kinesis}()},
\code{\link{appender_ntfy}()},
\code{\link{appender_pushbullet}()},
\code{\link{appender_slack}()},
\code{\link{appender_stdout}()},
\code{\link{appender_tee}()},
\code{\link{appender_telegram}()}
}
\concept{log_appenders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/appenders.R
\name{appender_stdout}
\alias{appender_stdout}
\title{Append log record to stdout}
\usage{
appender_stdout(lines)
}
\arguments{
\item{lines}{character vector}
}
\description{
Append log record to stdout
}
\seealso{
Other log_appenders: 
\code{\link{appender_async}()},
\code{\link{appender_console}()},
\code{\link{appender_file}()},
\code{\link{appender_kinesis}()},
\code{\link{appender_ntfy}()},
\code{\link{appender_pushbullet}()},
\code{\link{appender_slack}()},
\code{\link{appender_syslog}()},
\code{\link{appender_tee}()},
\code{\link{appender_telegram}()}
}
\concept{log_appenders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{log_eval}
\alias{log_eval}
\title{Evaluate an expression and log results}
\usage{
log_eval(expr, level = TRACE, multiline = FALSE)
}
\arguments{
\item{expr}{R expression to be evaluated while logging the expression itself along with the result}

\item{level}{\code{\link{log_levels}}}

\item{multiline}{setting to \code{FALSE} will print both the expression (enforced to be on one line by removing line-breaks if any) and its result on a single line separated by \code{=>}, while setting to \code{TRUE} will log the expression and the result in separate sections reserving line-breaks and rendering the printed results}
}
\description{
Evaluate an expression and log results
}
\examples{
\dontrun{
log_eval(pi * 2, level = INFO)

## lowering the log level threshold so that we don't have to set a higher level in log_eval
log_threshold(TRACE)
log_eval(x <- 4)
log_eval(sqrt(x))

## log_eval can be called in-line as well as returning the return value of the expression
x <- log_eval(mean(runif(1e3)))
x

## https://twitter.com/krlmlr/status/1067864829547999232
f <- sqrt
g <- mean
x <- 1:31
log_eval(f(g(x)), level = INFO)
log_eval(y <- f(g(x)), level = INFO)

## returning a function
log_eval(f <- sqrt)
log_eval(f)

## evaluating something returning a wall of "text"
log_eval(f <- log_eval)
log_eval(f <- log_eval, multiline = TRUE)

## doing something computationally intensive
log_eval(system.time(for(i in 1:100) mad(runif(1000))), multiline = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layouts.R
\name{layout_json_parser}
\alias{layout_json_parser}
\title{Generate log layout function rendering JSON after merging meta
fields with parsed list from JSON message}
\usage{
layout_json_parser(fields = default_fields())
}
\arguments{
\item{fields}{character vector of field names to be included in the
JSON. If named, the names will be used as field names in the JSON.}
}
\description{
Generate log layout function rendering JSON after merging meta
fields with parsed list from JSON message
}
\note{
This functionality depends on the \pkg{jsonlite} package.
}
\examples{
\dontshow{old <- logger:::namespaces_set()}
log_formatter(formatter_json)
log_info(everything = 42)

log_layout(layout_json_parser())
log_info(everything = 42)

log_layout(layout_json_parser(fields = c("time", "node")))
log_info(cars = row.names(mtcars), species = unique(iris$Species))

log_layout(layout_json_parser(fields = c(timestamp = "time", "node")))
log_info(
  message = paste(
    "Compared to the previous example,
    the 'time' field is renamed to 'timestamp'"
  )
)
\dontshow{logger:::namespaces_set(old)}
}
\seealso{
Other log_layouts: 
\code{\link{get_logger_meta_variables}()},
\code{\link{layout_blank}()},
\code{\link{layout_gha}()},
\code{\link{layout_glue}()},
\code{\link{layout_glue_colors}()},
\code{\link{layout_glue_generator}()},
\code{\link{layout_json}()},
\code{\link{layout_logging}()},
\code{\link{layout_simple}()}
}
\concept{log_layouts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/try.R
\name{\%except\%}
\alias{\%except\%}
\title{Try to evaluate an expressions and evaluate another expression on
exception}
\usage{
try \%except\% except
}
\arguments{
\item{try}{R expression}

\item{except}{fallback R expression to be evaluated if \code{try} fails}
}
\description{
Try to evaluate an expressions and evaluate another expression on
exception
}
\note{
Suppress log messages in the \code{except} namespace if you don't
want to throw a \code{WARN} log message on the exception branch.
}
\examples{
everything \%except\% 42
everything <- "640kb"
everything \%except\% 42

FunDoesNotExist(1:10) \%except\% sum(1:10) / length(1:10)
FunDoesNotExist(1:10) \%except\% (sum(1:10) / length(1:10))
FunDoesNotExist(1:10) \%except\% MEAN(1:10) \%except\% mean(1:10)
FunDoesNotExist(1:10) \%except\% (MEAN(1:10) \%except\% mean(1:10))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{top_env_name}
\alias{top_env_name}
\title{Returns the name of the top level environment from which the logger was called}
\usage{
top_env_name(.topenv = parent.frame())
}
\arguments{
\item{.topenv}{call environment}
}
\value{
string
}
\description{
Returns the name of the top level environment from which the logger was called
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/appenders.R
\name{appender_slack}
\alias{appender_slack}
\title{Send log messages to a Slack channel}
\usage{
appender_slack(
  channel = Sys.getenv("SLACK_CHANNEL"),
  username = Sys.getenv("SLACK_USERNAME"),
  icon_emoji = Sys.getenv("SLACK_ICON_EMOJI"),
  api_token = Sys.getenv("SLACK_API_TOKEN"),
  preformatted = TRUE
)
}
\arguments{
\item{channel}{Slack channel name with a hashtag prefix for public channel and no prefix for private channels}

\item{username}{Slack (bot) username}

\item{icon_emoji}{optional override for the bot icon}

\item{api_token}{Slack API token}

\item{preformatted}{use code tags around the message?}
}
\value{
function taking \code{lines} argument
}
\description{
Send log messages to a Slack channel
}
\note{
This functionality depends on the \pkg{slackr} package.
}
\seealso{
This is generator function for \code{\link{log_appender}}, for alternatives, see eg \code{\link{appender_console}}, \code{\link{appender_file}}, \code{\link{appender_tee}}, \code{\link{appender_pushbullet}}, \code{\link{appender_telegram}}, \code{\link{appender_syslog}}, \code{\link{appender_kinesis}} and \code{\link{appender_async}} for evaluate any \code{\link{log_appender}} function in a background process.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/appenders.R
\name{appender_pushbullet}
\alias{appender_pushbullet}
\title{Send log messages to Pushbullet}
\usage{
appender_pushbullet(...)
}
\arguments{
\item{...}{parameters passed to \link[RPushbullet:pbPost]{RPushbullet::pbPost}, such as \code{recipients}
or \code{apikey}, although it's probably much better to set all these
in the \verb{~/.rpushbullet.json} as per package docs at
\url{http://dirk.eddelbuettel.com/code/rpushbullet.html}}
}
\description{
Send log messages to Pushbullet
}
\note{
This functionality depends on the \pkg{RPushbullet} package.
}
\seealso{
Other log_appenders: 
\code{\link{appender_async}()},
\code{\link{appender_console}()},
\code{\link{appender_file}()},
\code{\link{appender_kinesis}()},
\code{\link{appender_ntfy}()},
\code{\link{appender_slack}()},
\code{\link{appender_stdout}()},
\code{\link{appender_syslog}()},
\code{\link{appender_tee}()},
\code{\link{appender_telegram}()}
}
\concept{log_appenders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/appenders.R
\name{appender_console}
\alias{appender_console}
\alias{appender_stderr}
\title{Append log record to stderr}
\usage{
appender_console(lines)

appender_stderr(lines)
}
\arguments{
\item{lines}{character vector}
}
\description{
Append log record to stderr
}
\seealso{
Other log_appenders: 
\code{\link{appender_async}()},
\code{\link{appender_file}()},
\code{\link{appender_kinesis}()},
\code{\link{appender_pushbullet}()},
\code{\link{appender_slack}()},
\code{\link{appender_stdout}()},
\code{\link{appender_syslog}()},
\code{\link{appender_tee}()},
\code{\link{appender_telegram}()}
}
\concept{log_appenders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hooks.R
\name{log_warnings}
\alias{log_warnings}
\title{Injects a logger call to standard warnings}
\usage{
log_warnings(muffle = getOption("logger_muffle_warnings", FALSE))
}
\arguments{
\item{muffle}{if TRUE, the warning is not shown after being logged}
}
\description{
This function sets a hook to trigger \code{\link[=log_warn]{log_warn()}} when \code{warning} is called
to log the warning messages with the global \code{logger} layout and appender.
}
\examples{
\dontrun{
log_warnings()
for (i in 1:5) {
  Sys.sleep(runif(1))
  warning(i)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logger.R
\name{log_threshold}
\alias{log_threshold}
\title{Get or set log level threshold}
\usage{
log_threshold(level = NULL, namespace = "global", index = 1)
}
\arguments{
\item{level}{see \code{\link[=log_levels]{log_levels()}}}

\item{namespace}{logger namespace}

\item{index}{index of the logger within the namespace}
}
\value{
currently set log level threshold
}
\description{
Get or set log level threshold
}
\examples{
\dontshow{old <- logger:::namespaces_set()}
## check the currently set log level threshold
log_threshold()

## change the log level threshold to WARN
log_threshold(WARN)
log_info(1)
log_warn(2)

## add another logger with a lower log level threshold and check the number of logged messages
log_threshold(INFO, index = 2)
log_info(1)
log_warn(2)

## set the log level threshold in all namespaces to ERROR
log_threshold(ERROR, namespace = log_namespaces())
\dontshow{logger:::namespaces_set(old)}
}
\seealso{
Other log configutation functions: 
\code{\link{log_appender}()},
\code{\link{log_formatter}()},
\code{\link{log_layout}()}
}
\concept{log configutation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hooks.R
\name{log_messages}
\alias{log_messages}
\title{Injects a logger call to standard messages}
\usage{
log_messages()
}
\description{
This function uses \code{trace} to add a \code{log_info} function call when
\code{message} is called to log the informative messages with the
\code{logger} layout and appender.
}
\examples{
\dontrun{
log_messages()
message("hi there")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layouts.R
\name{layout_logging}
\alias{layout_logging}
\title{Format a log record as the logging package does by default}
\usage{
layout_logging(
  level,
  msg,
  namespace = NA_character_,
  .logcall = sys.call(),
  .topcall = sys.call(-1),
  .topenv = parent.frame()
)
}
\arguments{
\item{level}{log level, see \code{\link{log_levels}} for more details}

\item{msg}{string message}

\item{namespace}{string referring to the \code{logger} environment / config to be used to override the target of the message record to be used instead of the default namespace, which is defined by the R package name from which the logger was called, and falls back to a common, global namespace.}

\item{.logcall}{the logging call being evaluated (useful in formatters and layouts when you want to have access to the raw, unevaluated R expression)}

\item{.topcall}{R expression from which the logging function was called (useful in formatters and layouts to extract the calling function's name or arguments)}

\item{.topenv}{original frame of the \code{.topcall} calling function where the formatter function will be evaluated and that is used to look up the \code{namespace} as well via \code{logger:::top_env_name}}
}
\value{
character vector
}
\description{
Format a log record as the logging package does by default
}
\examples{
\dontrun{
log_layout(layout_logging)
log_info(42)
log_info(42, namespace = 'everything')

devtools::load_all(system.file('demo-packages/logger-tester-package', package = 'logger'))
logger_tester_function(INFO, 42)
}
}
\seealso{
This is a \code{\link{log_layout}}, for alternatives, see \code{\link{layout_blank}}, \code{\link{layout_glue}}, \code{\link{layout_glue_colors}}, \code{\link{layout_json}}, \code{\link{layout_json_parser}}, or generator functions such as \code{\link{layout_glue_generator}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logger.R
\name{log_level}
\alias{log_level}
\alias{log_fatal}
\alias{log_error}
\alias{log_warn}
\alias{log_success}
\alias{log_info}
\alias{log_debug}
\alias{log_trace}
\title{Log a message with given log level}
\usage{
log_level(
  level,
  ...,
  namespace = NA_character_,
  .logcall = sys.call(),
  .topcall = sys.call(-1),
  .topenv = parent.frame(),
  .timestamp = Sys.time()
)

log_fatal(
  ...,
  namespace = NA_character_,
  .logcall = sys.call(),
  .topcall = sys.call(-1),
  .topenv = parent.frame(),
  .timestamp = Sys.time()
)

log_error(
  ...,
  namespace = NA_character_,
  .logcall = sys.call(),
  .topcall = sys.call(-1),
  .topenv = parent.frame(),
  .timestamp = Sys.time()
)

log_warn(
  ...,
  namespace = NA_character_,
  .logcall = sys.call(),
  .topcall = sys.call(-1),
  .topenv = parent.frame(),
  .timestamp = Sys.time()
)

log_success(
  ...,
  namespace = NA_character_,
  .logcall = sys.call(),
  .topcall = sys.call(-1),
  .topenv = parent.frame(),
  .timestamp = Sys.time()
)

log_info(
  ...,
  namespace = NA_character_,
  .logcall = sys.call(),
  .topcall = sys.call(-1),
  .topenv = parent.frame(),
  .timestamp = Sys.time()
)

log_debug(
  ...,
  namespace = NA_character_,
  .logcall = sys.call(),
  .topcall = sys.call(-1),
  .topenv = parent.frame(),
  .timestamp = Sys.time()
)

log_trace(
  ...,
  namespace = NA_character_,
  .logcall = sys.call(),
  .topcall = sys.call(-1),
  .topenv = parent.frame(),
  .timestamp = Sys.time()
)
}
\arguments{
\item{level}{log level, see \code{\link[=log_levels]{log_levels()}} for more details}

\item{...}{R objects that can be converted to a character vector
via the active message formatter function}

\item{namespace}{string referring to the \code{logger} environment /
config to be used to override the target of the message record to
be used instead of the default namespace, which is defined by the
R package name from which the logger was called, and falls back
to a common, global namespace.}

\item{.logcall}{the logging call being evaluated (useful in
formatters and layouts when you want to have access to the raw,
unevaluated R expression)}

\item{.topcall}{R expression from which the logging function was
called (useful in formatters and layouts to extract the calling
function's name or arguments)}

\item{.topenv}{original frame of the \code{.topcall} calling function
where the formatter function will be evaluated and that is used
to look up the \code{namespace} as well via \code{logger:::top_env_name}}

\item{.timestamp}{The time the logging occured. Defaults to the current time
but may be overwritten if the logging is delayed from the time it happend}
}
\value{
Invisible \code{list} of \code{logger} objects. See \code{\link[=logger]{logger()}} for more details on the format.
}
\description{
Log a message with given log level
}
\examples{
\dontshow{old <- logger:::namespaces_set()}
log_level(INFO, "hi there")
log_info("hi there")

## output omitted
log_debug("hi there")

## lower threshold and retry
log_threshold(TRACE)
log_debug("hi there")

## multiple lines
log_info("ok {1:3} + {1:3} = {2*(1:3)}")

## use json layout
log_layout(layout_json(c("time", "level")))
log_info("ok {1:3} + {1:3} = {2*(1:3)}")
\dontshow{logger:::namespaces_set(old)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatters.R
\name{formatter_logging}
\alias{formatter_logging}
\title{Mimic the default formatter used in the \pkg{logging} package}
\usage{
formatter_logging(
  ...,
  .logcall = sys.call(),
  .topcall = sys.call(-1),
  .topenv = parent.frame()
)
}
\arguments{
\item{...}{string and further params passed to \code{\link[=sprintf]{sprintf()}} or R
expressions to be evaluated}

\item{.logcall}{the logging call being evaluated (useful in
formatters and layouts when you want to have access to the raw,
unevaluated R expression)}

\item{.topcall}{R expression from which the logging function was
called (useful in formatters and layouts to extract the calling
function's name or arguments)}

\item{.topenv}{original frame of the \code{.topcall} calling function
where the formatter function will be evaluated and that is used
to look up the \code{namespace} as well via \code{logger:::top_env_name}}
}
\value{
character vector
}
\description{
The \pkg{logging} package uses a formatter that behaves differently
when the input is a string or other R object. If the first argument
is a string, then \code{\link[=sprintf]{sprintf()}} is being called -- otherwise it does
something like \code{\link[=log_eval]{log_eval()}} and logs the R expression(s) and the
result(s) as well.
}
\examples{
\dontshow{old <- logger:::namespaces_set()}
log_formatter(formatter_logging)
log_info("42")
log_info(42)
log_info(4 + 2)
log_info("foo \%s", "bar")
log_info("vector \%s", 1:3)
log_info(12, 1 + 1, 2 * 2)
\dontshow{logger:::namespaces_set(old)}
}
\seealso{
Other log_formatters: 
\code{\link{formatter_cli}()},
\code{\link{formatter_glue}()},
\code{\link{formatter_glue_or_sprintf}()},
\code{\link{formatter_glue_safe}()},
\code{\link{formatter_json}()},
\code{\link{formatter_pander}()},
\code{\link{formatter_paste}()},
\code{\link{formatter_sprintf}()}
}
\concept{log_formatters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logger.R
\name{delete_logger_index}
\alias{delete_logger_index}
\title{Delete an index from a logger namespace}
\usage{
delete_logger_index(namespace = "global", index)
}
\arguments{
\item{namespace}{logger namespace}

\item{index}{index of the logger within the namespace}
}
\description{
Delete an index from a logger namespace
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/appenders.R
\name{appender_ntfy}
\alias{appender_ntfy}
\title{Send log messages to ntfy}
\usage{
appender_ntfy(title = "{logger}", tags = c("memo"), ...)
}
\arguments{
\item{title}{notification title}

\item{tags}{emoji (or general tag) for notification. See \link[ntfy:emoji]{ntfy::emoji}}

\item{...}{extra parameters passed to \link[ntfy:ntfy_send]{ntfy::ntfy_send} such as
\code{priority}, \code{topic}, etc.}
}
\description{
Send log messages to ntfy
}
\details{
Configure server and topic via environment variables. See
\code{\link[ntfy:ntfy_topic]{ntfy::ntfy_topic()}} for details
}
\note{
This functionality depends on the \pkg{ntfy} package.
}
\seealso{
Other log_appenders: 
\code{\link{appender_async}()},
\code{\link{appender_console}()},
\code{\link{appender_file}()},
\code{\link{appender_kinesis}()},
\code{\link{appender_pushbullet}()},
\code{\link{appender_slack}()},
\code{\link{appender_stdout}()},
\code{\link{appender_syslog}()},
\code{\link{appender_tee}()},
\code{\link{appender_telegram}()}
}
\concept{log_appenders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatters.R
\name{formatter_glue_or_sprintf}
\alias{formatter_glue_or_sprintf}
\title{Apply \code{\link[glue:glue]{glue::glue()}} and \code{\link[=sprintf]{sprintf()}}}
\usage{
formatter_glue_or_sprintf(
  msg,
  ...,
  .logcall = sys.call(),
  .topcall = sys.call(-1),
  .topenv = parent.frame()
)
}
\arguments{
\item{msg}{passed to \code{\link[=sprintf]{sprintf()}} as \code{fmt} or handled as part of \code{...}
in \code{\link[glue:glue]{glue::glue()}}}

\item{...}{passed to \code{\link[glue:glue]{glue::glue()}} for the text interpolation}

\item{.logcall}{the logging call being evaluated (useful in
formatters and layouts when you want to have access to the raw,
unevaluated R expression)}

\item{.topcall}{R expression from which the logging function was
called (useful in formatters and layouts to extract the calling
function's name or arguments)}

\item{.topenv}{original frame of the \code{.topcall} calling function
where the formatter function will be evaluated and that is used
to look up the \code{namespace} as well via \code{logger:::top_env_name}}
}
\value{
character vector
}
\description{
The best of both words: using both formatter functions in your log
messages, which can be useful eg if you are migrating from
\code{\link[=sprintf]{sprintf()}} formatted log messages to \code{\link[glue:glue]{glue::glue()}} or similar.
}
\details{
Note that this function tries to be smart when passing arguments to
\code{\link[glue:glue]{glue::glue()}} and \code{\link[=sprintf]{sprintf()}}, but might fail with some edge cases, and
returns an unformatted string.
}
\examples{
formatter_glue_or_sprintf("{a} + {b} = \%s", a = 2, b = 3, 5)
formatter_glue_or_sprintf("{pi} * {2} = \%s", pi * 2)
formatter_glue_or_sprintf("{pi} * {2} = {pi*2}")

formatter_glue_or_sprintf("Hi ", "{c('foo', 'bar')}, did you know that 2*4={2*4}")
formatter_glue_or_sprintf("Hi {c('foo', 'bar')}, did you know that 2*4={2*4}")
formatter_glue_or_sprintf("Hi {c('foo', 'bar')}, did you know that 2*4=\%s", 2 * 4)
formatter_glue_or_sprintf("Hi \%s, did you know that 2*4={2*4}", c("foo", "bar"))
formatter_glue_or_sprintf("Hi \%s, did you know that 2*4=\%s", c("foo", "bar"), 2 * 4)
}
\seealso{
Other log_formatters: 
\code{\link{formatter_cli}()},
\code{\link{formatter_glue}()},
\code{\link{formatter_glue_safe}()},
\code{\link{formatter_json}()},
\code{\link{formatter_logging}()},
\code{\link{formatter_pander}()},
\code{\link{formatter_paste}()},
\code{\link{formatter_sprintf}()}
}
\concept{log_formatters}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logger.R
\name{log_appender}
\alias{log_appender}
\title{Get or set log record appender function}
\usage{
log_appender(appender = NULL, namespace = "global", index = 1)
}
\arguments{
\item{appender}{function delivering a log record to the
destination, eg \code{\link[=appender_console]{appender_console()}}, \code{\link[=appender_file]{appender_file()}} or
\code{\link[=appender_tee]{appender_tee()}}, default NULL}

\item{namespace}{logger namespace}

\item{index}{index of the logger within the namespace}
}
\description{
Get or set log record appender function
}
\examples{
\dontshow{old <- logger:::namespaces_set()}
## change appender to "tee" that writes to the console and a file as well
t <- tempfile()
log_appender(appender_tee(t))
log_info(42)
log_info(43)
log_info(44)
readLines(t)

## poor man's tee by stacking loggers in the namespace
t <- tempfile()
log_appender(appender_stdout)
log_appender(appender_file(t), index = 2)
log_info(42)
readLines(t)
\dontshow{logger:::namespaces_set(old)}
}
\seealso{
Other log configutation functions: 
\code{\link{log_formatter}()},
\code{\link{log_layout}()},
\code{\link{log_threshold}()}
}
\concept{log configutation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logger.R
\name{with_log_threshold}
\alias{with_log_threshold}
\title{Evaluate R expression with a temporarily updated log level threshold}
\usage{
with_log_threshold(
  expression,
  threshold = ERROR,
  namespace = "global",
  index = 1
)
}
\arguments{
\item{expression}{R command}

\item{threshold}{\code{\link[=log_levels]{log_levels()}}}

\item{namespace}{logger namespace}

\item{index}{index of the logger within the namespace}
}
\description{
Evaluate R expression with a temporarily updated log level threshold
}
\examples{
\dontshow{old <- logger:::namespaces_set()}
log_threshold(TRACE)
log_trace("Logging everything!")
x <- with_log_threshold(
  {
    log_info("Now we are temporarily suppressing eg INFO messages")
    log_warn("WARN")
    log_debug("Debug messages are suppressed as well")
    log_error("ERROR")
    invisible(42)
  },
  threshold = WARN
)
x
log_trace("DONE")
\dontshow{logger:::namespaces_set(old)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layouts.R
\name{layout_syslognet}
\alias{layout_syslognet}
\title{Format a log record for syslognet}
\usage{
layout_syslognet(
  level,
  msg,
  namespace = NA_character_,
  .logcall = sys.call(),
  .topcall = sys.call(-1),
  .topenv = parent.frame(),
  .timestamp = Sys.time()
)
}
\arguments{
\item{level}{log level, see \code{\link[=log_levels]{log_levels()}} for more details}

\item{msg}{string message}

\item{namespace}{string referring to the \code{logger} environment /
config to be used to override the target of the message record to
be used instead of the default namespace, which is defined by the
R package name from which the logger was called, and falls back
to a common, global namespace.}

\item{.logcall}{the logging call being evaluated (useful in
formatters and layouts when you want to have access to the raw,
unevaluated R expression)}

\item{.topcall}{R expression from which the logging function was
called (useful in formatters and layouts to extract the calling
function's name or arguments)}

\item{.topenv}{original frame of the \code{.topcall} calling function
where the formatter function will be evaluated and that is used
to look up the \code{namespace} as well via \code{logger:::top_env_name}}

\item{.timestamp}{The time the logging occured. Defaults to the current time
but may be overwritten if the logging is delayed from the time it happend}
}
\value{
A character vector with a severity attribute.
}
\description{
Format a log record for syslognet.
This function converts the logger log level to a
log severity level according to RFC 5424 "The Syslog Protocol".
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layouts.R
\name{get_logger_meta_variables}
\alias{get_logger_meta_variables}
\title{Collect useful information about the logging environment to be used in log messages}
\usage{
get_logger_meta_variables(
  log_level = NULL,
  namespace = NA_character_,
  .logcall = sys.call(),
  .topcall = sys.call(-1),
  .topenv = parent.frame(),
  .timestamp = Sys.time()
)
}
\arguments{
\item{log_level}{log level as per \code{\link[=log_levels]{log_levels()}}}

\item{namespace}{string referring to the \code{logger} environment /
config to be used to override the target of the message record to
be used instead of the default namespace, which is defined by the
R package name from which the logger was called, and falls back
to a common, global namespace.}

\item{.logcall}{the logging call being evaluated (useful in
formatters and layouts when you want to have access to the raw,
unevaluated R expression)}

\item{.topcall}{R expression from which the logging function was
called (useful in formatters and layouts to extract the calling
function's name or arguments)}

\item{.topenv}{original frame of the \code{.topcall} calling function
where the formatter function will be evaluated and that is used
to look up the \code{namespace} as well via \code{logger:::top_env_name}}

\item{.timestamp}{The time the logging occured. Defaults to the current time
but may be overwritten if the logging is delayed from the time it happend}
}
\value{
list
}
\description{
Available variables to be used in the log formatter functions, eg in \code{\link[=layout_glue_generator]{layout_glue_generator()}}:
}
\details{
\itemize{
\item \code{levelr}: log level as an R object, eg \code{\link[=INFO]{INFO()}}
\item \code{level}: log level as a string, eg \code{\link[=INFO]{INFO()}}
\item \code{time}: current time as \code{POSIXct}
\item \code{node}: name by which the machine is known on the network as reported by \code{Sys.info}
\item \code{arch}: machine type, typically the CPU architecture
\item \code{os_name}: Operating System's name
\item \code{os_release}: Operating System's release
\item \code{os_version}: Operating System's version
\item \code{user}: name of the real user id as reported by \code{Sys.info}
\item \code{pid}: the process identification number of the R session
\item \code{node}: name by which the machine is known on the network as reported by \code{Sys.info}
\item \code{r_version}: R's major and minor version as a string
\item \code{ns}: namespace usually defaults to \code{global} or the name of the holding R package
of the calling the logging function
\item \code{ns_pkg_version}: the version of \code{ns} when it's a package
\item \code{ans}: same as \code{ns} if there's a defined \code{\link[=logger]{logger()}} for the namespace,
otherwise a fallback namespace (eg usually \code{global})
\item \code{topenv}: the name of the top environment from which the parent call was called
(eg R package name or \code{GlobalEnv})
\item \code{call}: parent call (if any) calling the logging function
\item \code{location}: A list with element \code{path} and \code{line} giving the location of the
log call
\item \code{fn}: function's (if any) name calling the logging function
}
}
\seealso{
\code{\link[=layout_glue_generator]{layout_glue_generator()}}

Other log_layouts: 
\code{\link{layout_blank}()},
\code{\link{layout_gha}()},
\code{\link{layout_glue}()},
\code{\link{layout_glue_colors}()},
\code{\link{layout_glue_generator}()},
\code{\link{layout_json}()},
\code{\link{layout_json_parser}()},
\code{\link{layout_logging}()},
\code{\link{layout_simple}()}
}
\concept{log_layouts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{log_failure}
\alias{log_failure}
\title{Logs the error message to console before failing}
\usage{
log_failure(expression)
}
\arguments{
\item{expression}{call}
}
\description{
Logs the error message to console before failing
}
\examples{
log_failure("foobar")
try(log_failure(foobar))
}
